;;  Mips.md	     Machine Description for MIPS based processors
;;  Contributed by   A. Lichnewsky, lich@inria.inria.fr
;;  Changes by       Michael Meissner, meissner@osf.org
;;  Copyright (C) 1989-1991 Free Software Foundation, Inc.

;; This file is part of GNU CC.

;; GNU CC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU CC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU CC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.



;; ....................
;;
;;	Attributes
;;
;; ....................

;; Classification of each insn.
;; branch	conditional branch
;; jump		unconditional jump
;; call		unconditional call
;; load		load instruction(s)
;; store	store instruction(s)
;; move		data movement within same register set
;; xfer		transfer to/from coprocessor
;; arith	integer arithmetic instruction
;; imul		integer multiply
;; idiv		integer divide
;; icmp		integer compare
;; fadd		floating point add/subtract
;; fmul		floating point multiply
;; fdiv		floating point divide
;; fabs		floating point absolute value
;; fneg		floating point negation
;; fcmp		floating point compare
;; fcvt		floating point convert
;; multi	multiword sequence (or user asm statements)
;; nop		no operation

(define_attr "type"
  "unknown,branch,jump,call,load,store,move,xfer,arith,imul,idiv,icmp,fadd,fmul,fdiv,fabs,fneg,fcmp,fcvt,multi,nop"
  (const_string "unknown"))

;; Main data type used by the insn
(define_attr "mode" "unknown,none,QI,HI,SI,DI,SF,DF" (const_string "unknown"))

;; # instructions (4 bytes each)
(define_attr "length" "" (const_int 1))

;; whether or not an instruction has a mandiatory delay slot
(define_attr "dslot" "no,yes"
  (if_then_else (eq_attr "type" "branch,jump,call,load,xfer,fcmp")
		(const_string "yes")
		(const_string "no")))

;; Describe a user's asm statement.
(define_asm_attributes
  [(set_attr "type" "multi")])



;; .........................
;;
;;	Delay slots, can't describe load/fcmp/xfer delay slots here
;;
;; .........................

(define_delay (eq_attr "type" "call,jump,branch")
  [(and (eq_attr "dslot" "no")
	(eq_attr "length" "1"))
   (nil) (nil)])



;; .........................
;;
;;	Functional units
;;
;; .........................

; (define_function_unit NAME MULTIPLICITY SIMULTANEITY TEST READY-DELAY BUSY-DELAY [CONFLICT-LIST])

(define_function_unit "memory"  1 0 (eq_attr "type" "load")	 2 0)
(define_function_unit "memory"  1 0 (eq_attr "type" "store")	 1 0)

(define_function_unit "imuldiv" 1 1 (eq_attr "type" "imul")	12 24)	; 17 for R6000
(define_function_unit "imuldiv" 1 1 (eq_attr "type" "idiv")	35 70)	; 36/38 for R6000

(define_function_unit "adder"	1 1 (eq_attr "type" "fadd")	 2  4)
(define_function_unit "fast"	1 1 (eq_attr "type" "fabs,fneg") 1  2)

(define_function_unit "mult"	1 1 (ior (eq_attr "type" "fmul") (eq_attr "mode" "SF"))	  4  4)
(define_function_unit "mult"	1 1 (ior (eq_attr "type" "fmul") (eq_attr "mode" "DF"))	  5  6)
(define_function_unit "divide"	1 1 (ior (eq_attr "type" "fdiv") (eq_attr "mode" "SF"))	 12 16)
(define_function_unit "divide"	1 1 (ior (eq_attr "type" "fdiv") (eq_attr "mode" "DF"))	 19 30)


;;
;;  ....................
;;
;;	ADDITION
;;
;;  ....................
;;

(define_insn "adddf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(plus:DF (match_operand:DF 1 "register_operand" "f")
		 (match_operand:DF 2 "register_operand" "f")))]
  ""
  "add.d\\t%0,%1,%2"
  [(set_attr "type"	"fadd")
   (set_attr "mode"	"DF")
   (set_attr "length"	"1")])

(define_insn "addsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(plus:SF (match_operand:SF 1 "register_operand" "f")
		 (match_operand:SF 2 "register_operand" "f")))]
  ""
  "add.s\\t%0,%1,%2"
  [(set_attr "type"	"fadd")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])

(define_insn "addsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI (match_operand:SI 1 "arith_operand" "%r")
		 (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  return (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 0)
    ? \"sub%:\\t%0,%1,%n2\"
    : \"add%:\\t%0,%1,%2\";
}"
  [(set_attr "type"	"arith")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	SUBTRACTION
;;
;;  ....................
;;

(define_insn "subdf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(minus:DF (match_operand:DF 1 "register_operand" "f")
		  (match_operand:DF 2 "register_operand" "f")))]
  ""
  "sub.d\\t%0,%1,%2"
  [(set_attr "type"	"fadd")
   (set_attr "mode"	"DF")
   (set_attr "length"	"1")])

(define_insn "subsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(minus:SF (match_operand:SF 1 "register_operand" "f")
		  (match_operand:SF 2 "register_operand" "f")))]
  ""
  "sub.s\\t%0,%1,%2"
  [(set_attr "type"	"fadd")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])

(define_insn "subsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(minus:SI (match_operand:SI 1 "reg_or_0_operand" "rJ")
		  (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  return (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 0)
    ? \"add%:\\t%0,%z1,%n2\"
    : \"sub%:\\t%0,%z1,%2\";
}"
  [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	MULTIPLICATION
;;
;;  ....................
;;

(define_insn "muldf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(mult:DF (match_operand:DF 1 "register_operand" "f")
		 (match_operand:DF 2 "register_operand" "f")))]
  ""
  "mul.d\\t%0,%1,%2"
  [(set_attr "type"	"fmul")
   (set_attr "mode"	"DF")
   (set_attr "length"	"1")])

(define_insn "mulsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(mult:SF (match_operand:SF 1 "register_operand" "f")
		 (match_operand:SF 2 "register_operand" "f")))]
  ""
  "mul.s\\t%0,%1,%2"
  [(set_attr "type"	"fmul")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])

(define_insn "mulsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(mult:SI (match_operand:SI 1 "register_operand" "%r")
		 (match_operand:SI 2 "arith_operand" "rI")))
   (clobber (reg:SI 64))
   (clobber (reg:SI 65))]
  ""
  "mul\\t%0,%1,%2"
  [(set_attr "type"	"imul")
   (set_attr "mode"	"SI")
   (set_attr "length"	"3")])		;; mult + mflo + delay

(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(mult:SI (match_operand:SI 1 "register_operand" "")
		 (match_operand:SI 2 "register_operand" "")))
   (clobber (reg:SI 64))
   (clobber (reg:SI 65))]
  ""
  [(parallel [(set (reg:SI 65)		;; low register
		   (mult:SI (match_dup 1)
			    (match_dup 2)))
	      (clobber (reg:SI 64))])
   (set (match_dup 0)
	(reg:SI 65))]
  "")

(define_insn "mulsi3_internal"
  [(set (reg:SI 65)		;; low register
	(mult:SI (match_operand:SI 0 "register_operand" "r")
		 (match_operand:SI 1 "register_operand" "r")))
   (clobber (reg:SI 64))]
  ""
  "mult\\t%0,%1"
  [(set_attr "type"	"imul")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	DIVISION and REMAINDER
;;
;;  ....................
;;

(define_insn "divdf3"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(div:DF (match_operand:DF 1 "register_operand" "f")
		(match_operand:DF 2 "register_operand" "f")))]
  ""
  "div.d\\t%0,%1,%2"
  [(set_attr "type"	"fdiv")
   (set_attr "mode"	"DF")
   (set_attr "length"	"1")])

(define_insn "divsf3"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(div:SF (match_operand:SF 1 "register_operand" "f")
		(match_operand:SF 2 "register_operand" "f")))]
  ""
  "div.s\\t%0,%1,%2"
  [(set_attr "type"	"fdiv")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])

;; If optimizing, prefer the divmod functions over separate div and
;; mod functions, since this will allow using one instruction for both
;; the quotient and remainder.  At present, the divmod is not moved out
;; of loops if it is constant within the loop, so allow -mdebugc to
;; use the old method of doing things.

;; 64 is the multiply/divide hi register
;; 65 is the multiply/divide lo register

(define_insn "divmodsi4"
  [(parallel [(set (match_operand:SI 0 "register_operand" "=r")
		   (div:SI (match_operand:SI 1 "register_operand" "r")
			   (match_operand:SI 2 "register_operand" "r")))
	      (set (match_operand:SI 3 "register_operand" "=r")
		   (mod:SI (match_dup 1)
			   (match_dup 2)))
	      (clobber (reg:SI 64))
	      (clobber (reg:SI 65))])]
  "optimize && !TARGET_DEBUG_C_MODE"
  "*
{
  if (find_reg_note (insn, REG_UNUSED, operands[3]))
    return \"div\\t%0,%1,%2\";

  if (find_reg_note (insn, REG_UNUSED, operands[0]))
    return \"rem\\t%3,%1,%2\";

  return \"div\\t%0,%1,%2\;mfhi\\t%3\";
}"
  [(set_attr "type"	"idiv")
   (set_attr "mode"	"SI")
   (set_attr "length"	"13")])		;; various tests for dividing by 0 and such

(define_insn "udivmodsi4"
  [(parallel [(set (match_operand:SI 0 "register_operand" "=r")
		   (udiv:SI (match_operand:SI 1 "register_operand" "r")
			    (match_operand:SI 2 "register_operand" "r")))
	      (set (match_operand:SI 3 "register_operand" "=r")
		   (umod:SI (match_dup 1)
			    (match_dup 2)))
	      (clobber (reg:SI 64))
	      (clobber (reg:SI 65))])]
  "optimize && !TARGET_DEBUG_C_MODE"
  "*
{
  if (find_reg_note (insn, REG_UNUSED, operands[3]))
    return \"divu\\t%0,%1,%2\";

  if (find_reg_note (insn, REG_UNUSED, operands[0]))
    return \"remu\\t%3,%1,%2\";

  return \"divu\\t%0,%1,%2\;mfhi\\t%3\";
}"
  [(set_attr "type"	"idiv")
   (set_attr "mode"	"SI")
   (set_attr "length"	"13")])		;; various tests for dividing by 0 and such

(define_insn "divsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(div:SI (match_operand:SI 1 "register_operand" "r")
		(match_operand:SI 2 "register_operand" "r")))
   (clobber (reg:SI 64))
   (clobber (reg:SI 65))]
  "!optimize || TARGET_DEBUG_C_MODE"
  "div\\t%0,%1,%2"
  [(set_attr "type"	"idiv")
   (set_attr "mode"	"SI")
   (set_attr "length"	"13")])		;; various tests for dividing by 0 and such

(define_insn "modsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(mod:SI (match_operand:SI 1 "register_operand" "r")
		(match_operand:SI 2 "register_operand" "r")))
   (clobber (reg:SI 64))
   (clobber (reg:SI 65))]
  "!optimize || TARGET_DEBUG_C_MODE"
  "rem\\t%0,%1,%2"
  [(set_attr "type"	"idiv")
   (set_attr "mode"	"SI")
   (set_attr "length"	"14")])		;; various tests for dividing by 0 and such

(define_insn "udivsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(udiv:SI (match_operand:SI 1 "register_operand" "r")
		 (match_operand:SI 2 "register_operand" "r")))
   (clobber (reg:SI 64))
   (clobber (reg:SI 65))]
  "!optimize || TARGET_DEBUG_C_MODE"
  "divu\\t%0,%1,%2"
  [(set_attr "type"	"idiv")
   (set_attr "mode"	"SI")
   (set_attr "length"	"14")])		;; various tests for dividing by 0 and such

(define_insn "umodsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(umod:SI (match_operand:SI 1 "register_operand" "r")
		 (match_operand:SI 2 "register_operand" "r")))
   (clobber (reg:SI 64))
   (clobber (reg:SI 65))]
  "!optimize || TARGET_DEBUG_C_MODE"
  "remu\\t%0,%1,%2"
  [(set_attr "type"	"idiv")
   (set_attr "mode"	"SI")
   (set_attr "length"	"14")])		;; various tests for dividing by 0 and such


;;
;;  ....................
;;
;;	ABSOLUTE VALUE
;;
;;  ....................

;; Don't use the integer abs macro, since that signals an exception on
;; -2147483648 (sigh).

(define_insn "abssf2"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(abs:SF (match_operand:SF 1 "register_operand" "f")))]
  ""
  "abs.s\\t%0,%1"
  [(set_attr "type"	"fabs")
   (set_attr "mode"	"DF")
   (set_attr "length"	"1")])

(define_insn "absdf2"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(abs:DF (match_operand:DF 1 "register_operand" "f")))]
  ""
  "abs.d\\t%0,%1"
  [(set_attr "type"	"fabs")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])

;;
;;  ....................
;;
;;	LOGICAL
;;
;;  ....................
;;

(define_insn "andsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(and:SI (match_operand:SI 1 "uns_arith_operand" "%r,r")
		(match_operand:SI 2 "uns_arith_operand" "r,K")))]
  ""
  "@
   and\\t%0,%1,%2
   andi\\t%0,%1,%x2"
  [(set_attr "type"	"arith,arith")
   (set_attr "mode"	"SI,SI")
   (set_attr "length"	"1,1")])

;; Simple hack to recognize the "nor" instruction on the MIPS
;; This must appear before the normal or patterns, so that the
;; combiner will correctly fold things.

(define_insn "norsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(not:SI (ior:SI (match_operand:SI 1 "reg_or_0_operand" "rJ")
			(match_operand:SI 2 "reg_or_0_operand" "rJ"))))]
  ""
  "nor\\t%0,%z1,%z2"
  [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_insn "iorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(ior:SI (match_operand:SI 1 "uns_arith_operand" "%r,r")
		(match_operand:SI 2 "uns_arith_operand" "r,K")))]
  ""
  "@
   or\\t%0,%1,%2
   ori\\t%0,%1,%x2"
  [(set_attr "type"	"arith,arith")
   (set_attr "mode"	"SI,SI")
   (set_attr "length"	"1,1")])

(define_insn "xorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(xor:SI (match_operand:SI 1 "uns_arith_operand" "%r,r")
		(match_operand:SI 2 "uns_arith_operand" "r,K")))]
  ""
  "@
   xor\\t%0,%1,%2
   xori\\t%0,%1,%x2"
  [(set_attr "type"	"arith,arith")
   (set_attr "mode"	"SI,SI")
   (set_attr "length"	"1,1")])


;;
;;  ....................
;;
;;	TRUNCATION
;;
;;  ....................

(define_insn "truncdfsf2"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(float_truncate:SF (match_operand:DF 1 "register_operand" "f")))]
  ""
  "cvt.s.d\\t%0,%1"
  [(set_attr "type"	"fcvt")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	ZERO EXTENSION
;;
;;  ....................

;; Extension insns.
;; Those for integer source operand
;; are ordered widest source type first.

(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(zero_extend:SI (match_operand:HI 1 "general_operand" "r,R,m")))]
  ""
  "@
   andi\\t%0,%1,0xffff
   lhu\\t%0,%1
   lhu\\t%0,%1"
  [(set_attr "type"	"arith,load,load")
   (set_attr "mode"	"SI,SI,SI")
   (set_attr "length"	"1,1,2")])

(define_insn "zero_extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(zero_extend:HI (match_operand:QI 1 "general_operand" "r,R,m")))]
  ""
  "@
   andi\\t%0,%1,0x00ff
   lbu\\t%0,%1
   lbu\\t%0,%1"
  [(set_attr "type"	"arith,load,load")
   (set_attr "mode"	"HI,HI,HI")
   (set_attr "length"	"1,1,2")])

(define_insn "zero_extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(zero_extend:SI (match_operand:QI 1 "general_operand" "r,R,m")))]
  ""
  "@
   andi\\t%0,%1,0x00ff
   lbu\\t%0,%1
   lbu\\t%0,%1"
  [(set_attr "type"	"arith,load,load")
   (set_attr "mode"	"SI,SI,SI")
   (set_attr "length"	"1,1,2")])


;;
;;  ....................
;;
;;	SIGN EXTENSION
;;
;;  ....................

;; Extension insns.
;; Those for integer source operand
;; are ordered widest source type first.

(define_insn "extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(sign_extend:SI (match_operand:HI 1 "general_operand" "r,R,m")))]
  ""
  "@
   sll\\t%0,%1,16\;sra\\t%0,%0,16
   lh\\t%0,%1
   lh\\t%0,%1"
  [(set_attr "type"	"multi,load,load")
   (set_attr "mode"	"SI,SI,SI")
   (set_attr "length"	"2,1,2")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(sign_extend:SI (match_operand:HI 1 "register_operand" "")))]
  ""
  [(set (match_dup 0)
     	(ashift:SI (match_dup 1)
     		   (const_int 16)))
   (set (match_dup 0)
     	(ashiftrt:SI (match_dup 0)
     		     (const_int 16)))]
  "operands[1] = gen_lowpart (SImode, operands[1]);")

(define_insn "extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(sign_extend:HI (match_operand:QI 1 "general_operand" "r,R,m")))]
  ""
  "@
   sll\\t%0,%1,24\;sra\\t%0,%0,24
   lb\\t%0,%1
   lb\\t%0,%1"
  [(set_attr "type"	"multi,load,load")
   (set_attr "mode"	"HI,HI,HI")
   (set_attr "length"	"2,1,2")])

(define_split
  [(set (match_operand:HI 0 "register_operand" "")
	(sign_extend:HI (match_operand:QI 1 "register_operand" "")))]
  ""
  [(set (match_dup 0)
     	(ashift:HI (match_dup 1)
     		   (const_int 24)))
   (set (match_dup 0)
     	(ashiftrt:HI (match_dup 0)
     		     (const_int 24)))]
  " operands[1] = gen_lowpart (SImode, operands[1]);")

(define_insn "extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(sign_extend:SI (match_operand:QI 1 "general_operand" "r,R,m")))]
  ""
  "@
   sll\\t%0,%1,24\;sra\\t%0,%0,24
   lb\\t%0,%1
   lb\\t%0,%1"
  [(set_attr "type"	"multi,load,load")
   (set_attr "mode"	"SI,SI,SI")
   (set_attr "length"	"2,1,2")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(sign_extend:SI (match_operand:QI 1 "register_operand" "")))]
  ""
  [(set (match_dup 0)
     	(ashift:SI (match_dup 1)
     		   (const_int 24)))
   (set (match_dup 0)
     	(ashiftrt:SI (match_dup 0)
     		     (const_int 24)))]
  "operands[1] = gen_lowpart (SImode, operands[1]);")

(define_insn "extendsfdf2"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(float_extend:DF (match_operand:SF 1 "register_operand" "f")))]
  ""
  "cvt.d.s\\t%0,%1"
  [(set_attr "type"	"fcvt")
   (set_attr "mode"	"DF")
   (set_attr "length"	"1")])



;;
;;  ....................
;;
;;	CONVERSIONS
;;
;;  ....................


(define_insn "fix_truncdfsi2_internal"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(fix:DF (match_operand:DF 1 "register_operand" "f")))
   (clobber (match_operand:SI 2 "register_operand" "r"))]
  ""
  "trunc.w.d %0,%1,%2"
  [(set_attr "type"	"fcvt")
   (set_attr "mode"	"DF")
   (set_attr "length"	"12")])

(define_insn "fix_truncdfsi2_internal2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(fix:SI (fix:DF (match_operand:DF 1 "register_operand" "f"))))]
  ""
  "mfc1\\t%0,%1"
  [(set_attr "type"	"xfer")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_expand "fix_truncdfsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(fix:SI (fix:DF (match_operand:DF 1 "register_operand" "f"))))]
  ""
  "
{
  rtx reg1 = gen_reg_rtx (DFmode);	/* fp reg that gets trunc result */
  rtx reg2 = gen_reg_rtx (SImode);	/* gp reg that saves FP status bits */
  emit_insn (gen_fix_truncdfsi2_internal (reg1, operands[1], reg2));
  operands[1] = reg1;
  /* Fall through and generate default code */
}")

(define_insn "fix_truncsfsi2_internal"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(fix:SF (match_operand:SF 1 "register_operand" "f")))
   (clobber (match_operand:SI 2 "register_operand" "r"))]
  ""
  "trunc.w.s %0,%1,%2"
  [(set_attr "type"	"fcvt")
   (set_attr "mode"	"SF")
   (set_attr "length"	"12")])

(define_insn "fix_truncsfsi2_internal2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(fix:SI (fix:SF (match_operand:SF 1 "register_operand" "f"))))]
  ""
  "mfc1\\t%0,%1"
  [(set_attr "type"	"xfer")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_expand "fix_truncsfsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(fix:SI (fix:SF (match_operand:SF 1 "register_operand" "f"))))]
  ""
  "
{
  rtx reg1 = gen_reg_rtx (SFmode);	/* fp reg that gets trunc result */
  rtx reg2 = gen_reg_rtx (SImode);	/* gp reg that saves FP status bits */
  emit_insn (gen_fix_truncsfsi2_internal (reg1, operands[1], reg2));
  operands[1] = reg1;
  /* Fall through and generate default code */
}")

(define_insn "floatsidf2"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(float:DF (match_operand:SI 1 "register_operand" "r")))]
  ""
  "mtc1\\t%1,%0\;cvt.d.w\\t%0,%0"
  [(set_attr "type"	"fcvt")
   (set_attr "mode"	"DF")
   (set_attr "length"	"13")])

(define_insn "floatsisf2"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(float:SF (match_operand:SI 1 "register_operand" "r")))]
  ""
  "mtc1\\t%1,%0\;cvt.s.w\\t%0,%0"
  [(set_attr "type"	"fcvt")
   (set_attr "mode"	"DF")
   (set_attr "length"	"13")])

(define_expand "fixuns_truncdfsi2"
  [(set (match_operand:SI 0 "register_operand" "")
	(unsigned_fix:SI (match_operand:DF 1 "register_operand" "")))]
  ""
  "
{
  rtx reg1 = gen_reg_rtx (DFmode);
  rtx reg2 = gen_reg_rtx (DFmode);
  rtx reg3 = gen_reg_rtx (SImode);
  rtx label1 = gen_label_rtx ();
  rtx label2 = gen_label_rtx ();
  REAL_VALUE_TYPE offset = REAL_VALUE_LDEXP (1.0, 31);

  if (reg1)			/* turn off complaints about unreached code */
    {
      extern rtx gen_cmpdf ();
      emit_move_insn (reg1, immed_real_const_1 (offset, DFmode));
      do_pending_stack_adjust ();

      emit_insn (gen_cmpdf (operands[1], reg1));
      emit_jump_insn (gen_bge (label1));

      emit_insn (gen_fix_truncdfsi2 (operands[0], operands[1]));
      emit_jump_insn (gen_rtx (SET, VOIDmode, pc_rtx,
			       gen_rtx (LABEL_REF, VOIDmode, label2)));
      emit_barrier ();

      emit_label (label1);
      emit_move_insn (reg2, gen_rtx (MINUS, DFmode, operands[1], reg1));
      emit_move_insn (reg3, gen_rtx (CONST_INT, VOIDmode, 0x80000000));

      emit_insn (gen_fix_truncdfsi2 (operands[0], reg2));
      emit_insn (gen_iorsi3 (operands[0], operands[0], reg3));

      emit_label (label2);

      /* allow REG_NOTES to be set on last insn (labels don't have enough
	 fields, and can't be used for REG_NOTES anyway).  */
      emit_insn (gen_rtx (USE, VOIDmode, stack_pointer_rtx));
      DONE;
    }
}")

(define_expand "fixuns_truncsfsi2"
  [(set (match_operand:SI 0 "register_operand" "")
	(unsigned_fix:SI (match_operand:SF 1 "register_operand" "")))]
  ""
  "
{
  rtx reg1 = gen_reg_rtx (SFmode);
  rtx reg2 = gen_reg_rtx (SFmode);
  rtx reg3 = gen_reg_rtx (SImode);
  rtx label1 = gen_label_rtx ();
  rtx label2 = gen_label_rtx ();
  REAL_VALUE_TYPE offset = REAL_VALUE_LDEXP (1.0, 31);

  if (reg1)			/* turn off complaints about unreached code */
    {
      extern rtx gen_cmpsf ();
      emit_move_insn (reg1, immed_real_const_1 (offset, SFmode));
      do_pending_stack_adjust ();

      emit_insn (gen_cmpsf (operands[1], reg1));
      emit_jump_insn (gen_bge (label1));

      emit_insn (gen_fix_truncsfsi2 (operands[0], operands[1]));
      emit_jump_insn (gen_rtx (SET, VOIDmode, pc_rtx,
			       gen_rtx (LABEL_REF, VOIDmode, label2)));
      emit_barrier ();

      emit_label (label1);
      emit_move_insn (reg2, gen_rtx (MINUS, SFmode, operands[1], reg1));
      emit_move_insn (reg3, gen_rtx (CONST_INT, VOIDmode, 0x80000000));

      emit_insn (gen_fix_truncsfsi2 (operands[0], reg2));
      emit_insn (gen_iorsi3 (operands[0], operands[0], reg3));

      emit_label (label2);

      /* allow REG_NOTES to be set on last insn (labels don't have enough
	 fields, and can't be used for REG_NOTES anyway).  */
      emit_insn (gen_rtx (USE, VOIDmode, stack_pointer_rtx));
      DONE;
    }
}")


;; Wild things used when unions make double and int overlap.  This
;; must be supported since corresponding code gets generated.
;; The clobber in the first pattern is for catching double -> int
;; converts which have been optimized away.

(define_insn "movdf_xfer1"
  [(set (subreg:DF (match_operand:DI 0 "register_operand" "=r,*r,f,*f") 0)
	(match_operand:DF 1  "register_operand" "r,*f,f,*r"))
   (clobber (match_operand  2  "register_operand" "r,r,r,r"))]
  ""
  "* mips_move_2words (operands, insn); "
  [(set_attr "type"	"move,xfer,move,xfer")
   (set_attr "mode"	"DF,DF,DF,DF")
   (set_attr "length"	"2,2,1,2")])

(define_insn "movdf_xfer2"
  [(set (subreg:DF (match_operand:DI 0 "register_operand" "=r,*r,f,*f") 0)
	(match_operand:DF 1  "register_operand" "r,*f,f,*r"))]
  ""
  "* mips_move_2words (operands, insn); "
  [(set_attr "type"	"move,xfer,move,xfer")
   (set_attr "mode"	"DF,DF,DF,DF")
   (set_attr "length"	"2,2,1,2")])

(define_insn "movdf_xfer3"
  [(set (subreg:DF (match_operand:DI 0 "register_operand" "=r,r,*f,*f") 0)
	(match_operand:DF 1  "memory_operand" "R,o,*R,*o"))]
  ""
  "* mips_move_2words (operands, insn); "
  [(set_attr "type"	"load,load,load,load")
   (set_attr "mode"	"DF,DF,DF,DF")
   (set_attr "length"	"2,4,2,4")])


;;
;;  ....................
;;
;;	DATA MOVEMENT
;;
;;  ....................

;; unaligned word moves generated by the block moves.
;; We use (use (reg:SI 0)) to select this pattern rather than the
;; normal movsi.  Make these before the normal move patterns so they
;; match first.

(define_insn "movsi_ulw"
  [(set (match_operand:SI 0 "register_operand" "=&r,&r")
	(match_operand:SI 1 "memory_operand" "R,o"))
   (use (reg:SI 0))]
  ""
  "*
{
  if (TARGET_GAS && !simple_memory_operand (operands[1], SImode))
    {
      operands[2] = gen_rtx (REG, SImode, GP_REG_FIRST + 1);
      return \"%[la\\t%2,%1\;ulw\\t%0,0(%2)%]\";
    }

  return \"ulw\\t%0,%1\";
}"
  [(set_attr "type"	"load,load")
   (set_attr "mode"	"SI,SI")
   (set_attr "length"	"2,4")])

(define_insn "movsi_usw"
  [(set (match_operand:SI 0 "simple_memory_operand" "=&R,&o")
	(match_operand:SI 1 "register_operand" "r,r"))
   (use (reg:SI 0))]
  ""
  "*
{
  if (TARGET_GAS && !simple_memory_operand (operands[0], SImode))
    {
      operands[2] = gen_rtx (REG, SImode, GP_REG_FIRST + 1);
      return \"%[la\\t%2,%0\;usw\\t%1,0(%2)%]\";
    }

  return \"usw\\t%1,%0\";
}"
  [(set_attr "type"	"load,load")
   (set_attr "mode"	"SI,SI")
   (set_attr "length"	"2,4")])

;; 64-bit integer point moves

(define_expand "movdi"
  [(set (match_operand:DI 0 "general_operand" "")
	(match_operand:DI 1 "general_operand" ""))]
  ""
  "
{
  extern rtx force_reg ();

  if (GET_CODE (operands[0]) == MEM && !reg_or_0_operand (operands[1], DImode))
    operands[1] = force_reg (DImode, operands[1]);

  /* fall through and generate normal move code.  */
}")

;; Unlike most other insns, the move insns can't be split with
;; different predicates, because register spilling and other parts of
;; the compiler, have memoized the insn number already.

(define_insn "movdi_internal"
  [(set (match_operand:DI 0 "general_operand" "=r,r,r,r,R,o,*r,*f,*f,*f,*f,*R,*o")
	(match_operand:DI 1 "general_operand" " r,iF,R,o,r,r,*f,*r,*f,*R,*o,*f,*f"))]
  "(GET_CODE (operands[0]) == REG
    || (GET_CODE (operands[0]) == SUBREG && GET_CODE (SUBREG_REG (operands[0])) != MEM)
    || (GET_CODE (operands[0]) == MEM && reg_or_0_operand (operands[1], DImode)))"
  "* mips_move_2words (operands, insn); "
  [(set_attr "type"	"move,arith,load,load,store,store,xfer,xfer,move,load,load,store,store")
   (set_attr "mode"	"DI,DI,DI,DI,DI,DI,DI,DI,DI,DI,DI,DI,DI")
   (set_attr "length"	"2,4,2,4,2,4,2,2,1,2,4,2,4")])

;; 32-bit Integer moves

(define_expand "movsi"
  [(set (match_operand:SI 0 "general_operand" "")
	(match_operand:SI 1 "general_operand" ""))]
  ""
  "
{
  extern rtx force_reg ();

  if ((GET_CODE (operands[0]) == MEM && !reg_or_0_operand (operands[1], SImode))
      || (md_register_operand (operands[0], SImode) && GET_CODE (operands[1]) != REG)
      || (md_register_operand (operands[1], SImode) && GET_CODE (operands[0]) != REG))
    {
      operands[1] = force_reg (SImode, operands[1]);
    }

  /* fall through and generate normal move code.  */
}")


(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(match_operand:SI 1 "large_int" ""))]
  ""
  [(set (match_dup 0)
	(match_dup 2))
   (set (match_dup 0)
     	(ior:SI (match_dup 0)
		(match_dup 3)))]
  "
{
  operands[2] = gen_rtx (CONST_INT, VOIDmode, INTVAL (operands[1]) & 0xffff0000);
  operands[3] = gen_rtx (CONST_INT, VOIDmode, INTVAL (operands[1]) & 0x0000ffff);
}")

;; Unlike most other insns, the move insns can't be split with
;; different predicates, because register spilling and other parts of
;; the compiler, have memoized the insn number already.

(define_insn "movsi_internal"
  [(set (match_operand:SI 0 "general_operand" "=r,r,r,r,r,R,m,*r,*f,*f,*f,*f,*xs,*R,*m")
	(match_operand:SI 1 "general_operand"  "r,IKL,Mnis,R,m,rJ,rJ,*fxs,*r,*f,*R,*m,*r,*f,*f"))]
  "(GET_CODE (operands[0]) == REG
    || (GET_CODE (operands[0]) == SUBREG && GET_CODE (SUBREG_REG (operands[0])) != MEM)
    || (GET_CODE (operands[0]) == MEM && reg_or_0_operand (operands[1], SImode)))"
  "* return mips_move_1word (operands, insn);"
  [(set_attr "type"	"move,arith,arith,load,load,store,store,xfer,xfer,move,load,load,xfer,store,store")
   (set_attr "mode"	"SI,SI,SI,SI,SI,SI,SI,SI,SI,SI,SI,SI,SI,SI,SI")
   (set_attr "length"	"1,1,2,1,2,1,2,1,1,1,1,2,1,1,2")])

;; 16-bit Integer moves

(define_expand "movhi"
  [(set (match_operand:HI 0 "general_operand" "")
	(match_operand:HI 1 "general_operand" ""))]
  ""
  "
{
  extern rtx force_reg ();

  if (GET_CODE (operands[0]) == MEM && !reg_or_0_operand (operands[1], HImode))
    operands[1] = force_reg (HImode, operands[1]);

  /* fall through and generate normal move code.  */
}")

;; Unlike most other insns, the move insns can't be split with
;; different predicates, because register spilling and other parts of
;; the compiler, have memoized the insn number already.

(define_insn "movhi_internal"
  [(set (match_operand:HI 0 "general_operand" "=r,r,r,r,R,m,*r,*f,*f,*xs")
	(match_operand:HI 1 "general_operand" "r,IK,R,m,rJ,rJ,*fxs,*r,*f,*r"))]
  "(GET_CODE (operands[0]) == REG
    || (GET_CODE (operands[0]) == SUBREG && GET_CODE (SUBREG_REG (operands[0])) != MEM)
    || (GET_CODE (operands[0]) == MEM && reg_or_0_operand (operands[1], HImode)))"
  "* return mips_move_1word (operands, insn);"
  [(set_attr "type"	"move,arith,load,load,store,store,xfer,xfer,move,xfer")
   (set_attr "mode"	"HI,HI,HI,HI,HI,HI,HI,HI,HI,HI")
   (set_attr "length"	"1,1,1,2,1,2,1,1,1,1")])

;; 8-bit Integer moves

(define_expand "movqi"
  [(set (match_operand:QI 0 "general_operand" "")
	(match_operand:QI 1 "general_operand" ""))]
  ""
  "
{
  extern rtx force_reg ();

  if (GET_CODE (operands[0]) == MEM && !reg_or_0_operand (operands[1], QImode))
    operands[1] = force_reg (QImode, operands[1]);

  /* fall through and generate normal move code.  */
}")

;; Unlike most other insns, the move insns can't be split with
;; different predicates, because register spilling and other parts of
;; the compiler, have memoized the insn number already.

(define_insn "movqi_internal"
  [(set (match_operand:QI 0 "general_operand" "=r,r,r,r,R,m,*r,*f,*f,*xs")
	(match_operand:QI 1 "general_operand" " r,IK,R,m,rJ,rJ,*fxs,*r,*f,*r"))]
  "(GET_CODE (operands[0]) == REG
    || (GET_CODE (operands[0]) == SUBREG && GET_CODE (SUBREG_REG (operands[0])) != MEM)
    || (GET_CODE (operands[0]) == MEM && reg_or_0_operand (operands[1], QImode)))"
  "* return mips_move_1word (operands, insn);"
  [(set_attr "type"	"move,arith,load,load,store,store,xfer,xfer,move,xfer")
   (set_attr "mode"	"QI,QI,QI,QI,QI,QI,QI,QI,QI,QI")
   (set_attr "length"	"1,1,1,2,1,2,1,1,1,1")])


;; 32-bit floating point moves

(define_expand "movsf"
  [(set (match_operand:SF 0 "general_operand" "")
	(match_operand:SF 1 "general_operand" ""))]
  ""
  "
{
  extern rtx force_reg ();

  if (GET_CODE (operands[0]) == MEM && !reg_or_0_operand (operands[1], SFmode))
    operands[1] = force_reg (SFmode, operands[1]);

  /* fall through and generate normal move code.  */
}")

(define_insn "movsf_internal"
  [(set (match_operand:SF 0 "general_operand" "=f,f,f,f,R,m,*f,*r,*r,*r,*r,*R,*m")
	(match_operand:SF 1 "general_operand" "f,G,R,Em,fG,fG,*r,*f,*Gr,*R,*Em,*r,*r"))]
  "(GET_CODE (operands[0]) == REG
    || (GET_CODE (operands[0]) == SUBREG && GET_CODE (SUBREG_REG (operands[0])) != MEM)
    || (GET_CODE (operands[0]) == MEM && reg_or_0_operand (operands[1], SFmode)))"
  "* return mips_move_1word (operands, insn);"
  [(set_attr "type"	"move,xfer,load,load,store,store,xfer,xfer,move,load,load,store,store")
   (set_attr "mode"	"SF,SF,SF,SF,SF,SF,SF,SF,SF,SF,SF,SF,SF")
   (set_attr "length"	"1,1,1,2,1,2,1,1,1,1,2,1,2")])

;; 64-bit floating point moves

(define_expand "movdf"
  [(set (match_operand:DF 0 "general_operand" "")
	(match_operand:DF 1 "general_operand" ""))]
  ""
  "
{
  extern rtx force_reg ();

  if (GET_CODE (operands[0]) == MEM && !reg_or_0_operand (operands[1], DFmode))
    operands[1] = force_reg (DFmode, operands[1]);

  /* fall through and generate normal move code.  */
}")

(define_insn "movdf_internal"
  [(set (match_operand:DF 0 "general_operand" "=f,f,f,R,o,f,*f,*r,*r,*r,*r,*R,*o")
	(match_operand:DF 1 "general_operand"  "f,R,o,fG,fG,E,*r,*f,*rG,*R,*oE,*r,*r"))]
  "(GET_CODE (operands[0]) == REG
    || (GET_CODE (operands[0]) == SUBREG && GET_CODE (SUBREG_REG (operands[0])) != MEM)
    || (GET_CODE (operands[0]) == MEM && reg_or_0_operand (operands[1], DFmode)))"
  "* mips_move_2words (operands, insn); "
  [(set_attr "type"	"move,load,load,store,store,load,xfer,xfer,move,load,load,store,store")
   (set_attr "mode"	"DF,DF,DF,DF,DF,DF,DF,DF,DF,DF,DF,DF,DF")
   (set_attr "length"	"1,2,4,2,4,4,2,2,2,2,4,2,4")])


;; Block moves, see out-mips.c for more details.
;; Argument 0 is the destination
;; Argument 1 is the source
;; Argument 2 is the length
;; Argument 3 is the alignment

(define_expand "movstrsi"
  [(parallel [(set (mem:BLK (match_operand:BLK 0 "general_operand" ""))
		   (mem:BLK (match_operand:BLK 1 "general_operand" "")))
	      (use (match_operand:SI 2 "arith32_operand" ""))
	      (use (match_operand:SI 3 "immediate_operand" ""))])]
  ""
  "
{
  if (operands[0])		/* avoid unused code messages */
    {
      expand_block_move (operands);
      DONE;
    }
}")


;;
;;  ....................
;;
;;	SHIFTS
;;
;;  ....................

(define_insn "ashlsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashift:SI (match_operand:SI 1 "register_operand" "r")
		   (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  if (GET_CODE (operands[2]) == CONST_INT)
    operands[2] = gen_rtx (CONST_INT, VOIDmode, (XINT (operands[2], 0))& 0x1f);

  return \"sll\\t%0,%1,%2\";
}"
  [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_insn "ashrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI (match_operand:SI 1 "register_operand" "r")
		     (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  if (GET_CODE (operands[2]) == CONST_INT)
    operands[2] = gen_rtx (CONST_INT, VOIDmode, (XINT (operands[2], 0))& 0x1f);

  return \"sra\\t%0,%1,%2\";
}"
  [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_insn "lshrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lshiftrt:SI (match_operand:SI 1 "register_operand" "r")
		     (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "*
{
  if (GET_CODE (operands[2]) == CONST_INT)
    operands[2] = gen_rtx (CONST_INT, VOIDmode, (XINT (operands[2], 0))& 0x1f);

  return \"srl\\t%0,%1,%2\";
}"
  [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	NEGATION and ONE'S COMPLEMENT
;;
;;  ....................

(define_expand "negsi2"
  [(set (match_operand:SI 0 "register_operand" "")
	(minus:SI (const_int 0)
		  (match_operand:SI 1 "arith_operand" "")))]
  ""
  "")

(define_insn "negdf2"
  [(set (match_operand:DF 0 "register_operand" "=f")
	(neg:DF (match_operand:DF 1 "register_operand" "f")))]
  ""
  "neg.d\\t%0,%1"
  [(set_attr "type"	"fneg")
   (set_attr "mode"	"DF")
   (set_attr "length"	"1")])

(define_insn "negsf2"
  [(set (match_operand:SF 0 "register_operand" "=f")
	(neg:SF (match_operand:SF 1 "register_operand" "f")))]
  ""
  "neg.s\\t%0,%1"
  [(set_attr "type"	"fneg")
   (set_attr "mode"	"SF")
   (set_attr "length"	"1")])


(define_expand "one_cmplsi2"
  [(set (match_operand:SI 0 "register_operand" "")
	(not:SI (ior:SI (match_operand:SI 1 "register_operand" "")
			(const_int 0))))]
  ""
  "")


;;
;;  ....................
;;
;;	COMPARISONS
;;
;;  ....................

;; Flow here is rather complex:
;;
;;  1)	The cmp{si,sf,df} routine is called.  It deposits the
;;	arguments into the branch_cmp array, and the type into
;;	branch_type.  No RTL is generated.
;;
;;  2)	The appropriate branch define_expand is called, which then
;;	creates the appropriate RTL for the comparison and branch.
;;	Different CC modes are used, based on what type of branch is
;;	done, so that we can constrain things appropriately.  There
;;	are assumptions in the rest of GCC that break if we fold the
;;	operands into the branchs for integer operations, and use cc0
;;	for floating point.
;;
;;  3)	The compare define_insns then once again set branch_cmp and
;;	branch_type, and the branch define_insns use them.
;;
;;  4)	If a set condition code is done instead of a branch, then the
;;	operands are folded into the RTL, and a separate set of cc0 is
;;	not done.  This allows slt's to be put into delay slots.

(define_expand "cmpsi"
  [(set (cc0)
	(compare:CC (match_operand:SI 0 "reg_or_0_operand" "")
		    (match_operand:SI 1 "arith_operand" "")))]
  ""
  "
{
  if (operands[0])		/* avoid unused code message */
    {
      branch_cmp[0] = operands[0];
      branch_cmp[1] = operands[1];
      branch_type = CMP_SI;
      DONE;
    }
}")

(define_expand "tstsi"
  [(set (cc0)
	(match_operand:SI 0 "register_operand" ""))]
  ""
  "
{
  if (operands[0])		/* avoid unused code message */
    {
      branch_cmp[0] = operands[0];
      branch_cmp[1] = const0_rtx;
      branch_type = CMP_SI;
      DONE;
    }
}")

(define_insn "cmpsi_eqne"
  [(set (cc0)
	(compare:CC_EQ (match_operand:SI 0 "reg_or_0_operand" "rJ")
		       (match_operand:SI 1 "reg_or_0_operand" "rJ")))]
  ""
  "*
{
  branch_cmp[0] = operands[0];
  branch_cmp[1] = operands[1];
  branch_type = CMP_SI;
  return \"\";
}"
  [(set_attr "type"	"icmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"0")])

(define_insn "cmpsi_zero"
  [(set (cc0)
	(match_operand:SI 0 "reg_or_0_operand" "rJ"))]
  ""
  "*
{
  branch_cmp[0] = operands[0];
  branch_cmp[1] = const0_rtx;
  branch_type = CMP_SI;
  return \"\";
}"
  [(set_attr "type"	"icmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"0")])

(define_insn "cmpsi_relational"
  [(set (cc0)
	(compare:CC (match_operand:SI 0 "reg_or_0_operand" "rJ")
		    (match_operand:SI 1 "arith_operand" "rI")))]
  ""
  "*
{
  branch_cmp[0] = operands[0];
  branch_cmp[1] = operands[1];
  branch_type = CMP_SI;
  return \"\";
}"
  [(set_attr "type"	"icmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"0")])

(define_expand "cmpdf"
  [(set (cc0)
	(compare:CC_FP (match_operand:DF 0 "register_operand" "")
		       (match_operand:DF 1 "register_operand" "")))]
  ""
  "
{
  if (operands[0])		/* avoid unused code message */
    {
      branch_cmp[0] = operands[0];
      branch_cmp[1] = operands[1];
      branch_type = CMP_DF;
      DONE;
    }
}")

(define_insn "cmpdf_eq"
  [(parallel [(set (cc0)
		   (eq:CC_FP (match_operand:DF 0 "register_operand" "f")
			     (match_operand:DF 1 "register_operand" "f")))
	      (clobber (reg:SI 66))])]
  ""
  "c.eq.d\\t%0,%1%#"
  [(set_attr "type"	"fcmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "cmpdf_lt"
  [(parallel [(set (cc0)
		   (lt:CC_FP (match_operand:DF 0 "register_operand" "f")
			     (match_operand:DF 1 "register_operand" "f")))
	      (clobber (reg:SI 66))])]
  ""
  "c.lt.d\\t%0,%1%#"
  [(set_attr "type"	"fcmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "cmpdf_le"
  [(parallel [(set (cc0)
		   (le:CC_FP (match_operand:DF 0 "register_operand" "f")
			     (match_operand:DF 1 "register_operand" "f")))
	      (clobber (reg:SI 66))])]
  ""
  "c.le.d\\t%0,%1%#"
  [(set_attr "type"	"fcmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_expand "cmpsf"
  [(set (cc0)
	(compare:CC_FP (match_operand:SF 0 "register_operand" "")
		       (match_operand:SF 1 "register_operand" "")))]
  ""
  "
{
  if (operands[0])		/* avoid unused code message */
    {
      branch_cmp[0] = operands[0];
      branch_cmp[1] = operands[1];
      branch_type = CMP_SF;
      DONE;
    }
}")

(define_insn "cmpsf_eq"
  [(parallel [(set (cc0)
		   (eq:CC_FP (match_operand:SF 0 "register_operand" "f")
			     (match_operand:SF 1 "register_operand" "f")))
	      (clobber (reg:SI 66))])]
  ""
  "c.eq.s\\t%0,%1%#"
  [(set_attr "type"	"fcmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "cmpsf_lt"
  [(parallel [(set (cc0)
		   (lt:CC_FP (match_operand:SF 0 "register_operand" "f")
			     (match_operand:SF 1 "register_operand" "f")))
	      (clobber (reg:SI 66))])]
  ""
  "c.lt.s\\t%0,%1%#"
  [(set_attr "type"	"fcmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "cmpsf_le"
  [(parallel [(set (cc0)
		   (le:CC_FP (match_operand:SF 0 "register_operand" "f")
			     (match_operand:SF 1 "register_operand" "f")))
	      (clobber (reg:SI 66))])]
  ""
  "c.le.s\\t%0,%1%#"
  [(set_attr "type"	"fcmp")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	CONDITIONAL BRANCHES
;;
;;  ....................


;; We really can't note that integer branches clobber $at, and FP
;; branches clobber $fcr31 because if we use a parallel operation, a
;; normal insn is used to hold the value instead of jump_insn.  See
;; above for cmpxx saving the operands in branch_cmp and branch_type.

(define_insn "branch_fp_true1"
  [(set (pc)
	(if_then_else (eq:CC_FP (cc0)
				(const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "%*bc1t\\t%0"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_fp_false1"
  [(set (pc)
	(if_then_else (eq:CC_FP (cc0)
				(const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "%*bc1f\\t%0"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_fp_true2"
  [(set (pc)
	(if_then_else (ne:CC_FP (cc0)
				(const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "%*bc1t\\t%0"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_fp_false2"
  [(set (pc)
	(if_then_else (ne:CC_FP (cc0)
				(const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "%*bc1f\\t%0"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_eqne_true"
  [(set (pc)
	(if_then_else (match_operator:CC_EQ 0 "equality_op" [(cc0) (const_int 0)])
		      (label_ref (match_operand 1 "" ""))
		      (pc)))]
  ""
  "*
{
  operands[2] = branch_cmp[0];
  operands[3] = branch_cmp[1];
  return \"%*b%C0\\t%z2,%z3,%1\";
}"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_eqne_false"
  [(set (pc)
	(if_then_else (match_operator:CC_EQ 0 "equality_op" [(cc0) (const_int 0)])
		      (pc)
		      (label_ref (match_operand 1 "" ""))))]
  ""
  "*
{
  operands[2] = branch_cmp[0];
  operands[3] = branch_cmp[1];
  return \"%*b%N0\\t%z2,%z3,%1\";
}"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_zero_true"
  [(set (pc)
	(if_then_else (match_operator:CC_0 0 "cmp_op" [(cc0) (const_int 0)])
		      (label_ref (match_operand 1 "" ""))
		      (pc)))]
  ""
  "*
{
  operands[2] = branch_cmp[0];
  operands[3] = const0_rtx;
  switch (GET_CODE (operands[0]))
    {
    case EQ:  return \"%*beq\\t%z2,%z3,%1\";
    case NE:  return \"%*bne\\t%z2,%z3,%1\";
    case GTU: return \"%*bne\\t%z2,%z3,%1\";
    case GEU: return \"%*j\\t%1\";
    case LTU: return \"#bltuz\";
    case LEU: return \"%*beq\\t%z2,%z3,%1\";
    }

  return \"%*b%C0z\\t%z2,%1\";
}"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_zero_false"
  [(set (pc)
	(if_then_else (match_operator:CC_EQ 0 "cmp_op" [(cc0) (const_int 0)])
		      (pc)
		      (label_ref (match_operand 1 "" ""))))]
  ""
  "*
{
  operands[2] = branch_cmp[0];
  operands[3] = const0_rtx;
  switch (GET_CODE (operands[0]))
    {
    case EQ:  return \"%*bne\\t%z2,%z3,%1\";
    case NE:  return \"%*beq\\t%z2,%z3,%1\";
    case GTU: return \"%*beq\\t%z2,%z3,%1\";
    case GEU: return \"#bltuz\";
    case LTU: return \"%*j\\t%1\";
    case LEU: return \"%*bne\\t%z2,%z3,%1\";
    }

  return \"%*b%N0z\\t%z2,%1\";
}"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "branch_relop_true"
  [(set (pc)
	(if_then_else (match_operator:CC 0 "cmp2_op" [(cc0) (const_int 0)])
		      (label_ref (match_operand 1 "" ""))
		      (pc)))]
  ""
  "*
{
  operands[2] = branch_cmp[0];
  operands[3] = branch_cmp[1];

  return \"%&b%C0\\t%z2,%z3,%1%!\";
}"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"2")])

(define_insn "branch_relop_false"
  [(set (pc)
	(if_then_else (match_operator:CC 0 "cmp2_op" [(cc0) (const_int 0)])
		      (pc)
		      (label_ref (match_operand 1 "" ""))))]
  ""
  "*
{
  operands[2] = branch_cmp[0];
  operands[3] = branch_cmp[1];

  return \"%&b%N0\\t%z2,%z3,%1%!\";
}"
  [(set_attr "type"	"branch")
   (set_attr "mode"	"none")
   (set_attr "length"	"2")])

(define_expand "beq"
  [(set (pc)
	(if_then_else (eq:CC_EQ (cc0)
				(const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, EQ);
      DONE;
    }
}")

(define_expand "bne"
  [(set (pc)
	(if_then_else (ne:CC_EQ (cc0)
				(const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, NE);
      DONE;
    }
}")

(define_expand "bgt"
  [(set (pc)
	(if_then_else (gt:CC (cc0)
			     (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, GT);
      DONE;
    }
}")

(define_expand "bge"
  [(set (pc)
	(if_then_else (ge:CC (cc0)
			     (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, GE);
      DONE;
    }
}")

(define_expand "blt"
  [(set (pc)
	(if_then_else (lt:CC (cc0)
			     (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, LT);
      DONE;
    }
}")

(define_expand "ble"
  [(set (pc)
	(if_then_else (le:CC (cc0)
			     (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, LE);
      DONE;
    }
}")

(define_expand "bgtu"
  [(set (pc)
	(if_then_else (gtu:CC (cc0)
			      (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, GTU);
      DONE;
    }
}")

(define_expand "bgeu"
  [(set (pc)
	(if_then_else (geu:CC (cc0)
			      (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, GEU);
      DONE;
    }
}")


(define_expand "bltu"
  [(set (pc)
	(if_then_else (ltu:CC (cc0)
			      (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, LTU);
      DONE;
    }
}")

(define_expand "bleu"
  [(set (pc)
	(if_then_else (leu:CC (cc0)
			      (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  if (operands[0])		/* avoid unused code warning */
    {
      gen_conditional_branch (operands, LEU);
      DONE;
    }
}")


;;
;;  ....................
;;
;;	SETTING A REGISTER FROM A COMPARISON
;;
;;  ....................

(define_expand "seq"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(eq:CC_EQ (match_dup 1)
		  (match_dup 2)))]
  ""
  "
{
  extern rtx force_reg ();

  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 0)
    operands[2] = force_reg (SImode, operands[2]);

  /* fall through and generate default code */
}")

(define_insn "seq_si"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(eq:CC_EQ (match_operand:SI 1 "reg_or_0_operand" "%rJ,rJ,rJ")
		  (match_operand:SI 2 "uns_arith_operand" "J,r,K")))]
  ""
  "@
   sltu\\t%0,%1,1
   xor\\t%0,%1,%2\;sltu\\t%0,%0,1
   xori\\t%0,%1,%x2\;sltu\\t%0,%0,1"
 [(set_attr "type"	"arith,arith,arith")
   (set_attr "mode"	"SI,SI,SI")
   (set_attr "length"	"1,2,2")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "=r")
	(eq:CC_EQ (match_operand:SI 1 "reg_or_0_operand" "rJ")
		  (match_operand:SI 2 "uns_arith_operand" "rK")))]
  "GET_CODE (operands[2]) != CONST_INT || INTVAL (operands[2]) != 0"
  [(set (match_dup 0)
	(xor:SI (match_dup 1)
		(match_dup 2)))
   (set (match_dup 0)
	(ltu:CC (match_dup 0)
		(const_int 1)))]
  "")

(define_expand "sne"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ne:CC_EQ (match_dup 1)
		  (match_dup 2)))]
  ""
  "
{
  extern rtx force_reg ();

  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 0)
    operands[2] = force_reg (SImode, operands[2]);

  /* fall through and generate default code */
}")

(define_insn "sne_si"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(ne:CC_EQ (match_operand:SI 1 "reg_or_0_operand" "%rJ,rJ,rJ")
		  (match_operand:SI 2 "uns_arith_operand" "J,r,K")))]
  ""
  "*
{
  operands[3] = const0_rtx;
  if (GET_CODE (operands[2]) != CONST_INT)
    return \"xor\\t%0,%1,%2\;sltu\\t%0,%z3,%0\";

  if (INTVAL (operands[2]) == 0)
    return \"sltu\\t%0,%z3,%1\";

  return \"xori\\t%0,%1,%x2\;sltu\\t%0,%z3,%0\";
}"
 [(set_attr "type"	"arith,arith,arith")
   (set_attr "mode"	"SI,SI,SI")
   (set_attr "length"	"1,2,2")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ne:CC_EQ (match_operand:SI 1 "reg_or_0_operand" "rJ")
		  (match_operand:SI 2 "uns_arith_operand" "rK")))]
  "GET_CODE (operands[2]) != CONST_INT || INTVAL (operands[2]) != 0"
  [(set (match_dup 0)
	(xor:SI (match_dup 1)
		(match_dup 2)))
   (set (match_dup 0)
	(ltu:CC (const_int 0)
		(match_dup 0)))]
  "")

(define_expand "sgt"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(gt:CC (match_dup 1)
	       (match_dup 2)))]
  ""
  "
{
  extern rtx force_reg ();

  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) != 0)
    operands[2] = force_reg (SImode, operands[2]);

  /* fall through and generate default code */
}")

(define_insn "sgt_si"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(gt:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
	       (match_operand:SI 2 "reg_or_0_operand" "rJ")))]
  ""
  "slt\\t%0,%z2,%z1"
 [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_expand "sge"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ge:CC (match_dup 1)
	       (match_dup 2)))]
  ""
  "
{
  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  /* fall through and generate default code */
}")

(define_insn "sge_si"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ge:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
	       (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "slt\\t%0,%z1,%2\;xori\\t%0,%0,0x0001"
 [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"2")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ge:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
	       (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  [(set (match_dup 0)
	(lt:CC (match_dup 1)
	       (match_dup 2)))
   (set (match_dup 0)
	(xor:SI (match_dup 0)
		(const_int 1)))]
  "")

(define_expand "slt"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lt:CC (match_dup 1)
	       (match_dup 2)))]
  ""
  "
{
  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  /* fall through and generate default code */
}")

(define_insn "slt_si"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lt:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
	       (match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "slt\\t%0,%z1,%2"
 [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_expand "sle"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(le:CC (match_dup 1)
	       (match_dup 2)))]
  ""
  "
{
  extern rtx force_reg ();

  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) >= 32767)
    operands[2] = force_reg (SImode, operands[2]);

  /* fall through and generate default code */
}")

(define_insn "sle_si"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(le:CC (match_operand:SI 1 "reg_or_0_operand" "rJ,rJ")
	       (match_operand:SI 2 "arith_operand" "r,I")))]
  "GET_CODE (operands[2]) != CONST_INT || INTVAL (operands[2]) < 32767"
  "@
   slt\\t%0,%z2,%z1\;xori\\t%0,%0,0x0001
   slt\\t%0,%z1,(%2+1)"
 [(set_attr "type"	"arith,arith")
   (set_attr "mode"	"SI,SI")
   (set_attr "length"	"2,1")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "=r")
	(le:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
	       (match_operand:SI 2 "register_operand" "r")))]
  ""
  [(set (match_dup 0)
	(lt:CC (match_dup 2)
	       (match_dup 1)))
   (set (match_dup 0)
	(xor:SI (match_dup 0)
		(const_int 1)))]
  "")

(define_expand "sgtu"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(gtu:CC (match_dup 1)
		(match_dup 2)))]
  ""
  "
{
  extern rtx force_reg ();

  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) != 0)
    operands[2] = force_reg (SImode, operands[2]);

  /* fall through and generate default code */
}")

(define_insn "sgtu_si"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(gtu:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
		(match_operand:SI 2 "reg_or_0_operand" "rJ")))]
  ""
  "sltu\\t%0,%z2,%z1"
 [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_expand "sgeu"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (geu:CC (match_dup 1)
                (match_dup 2)))]
  ""
  "
{
  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  /* fall through and generate default code */
}")

(define_insn "sgeu_si"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(geu:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
		(match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "sltu\\t%0,%z1,%2\;xori\\t%0,%0,0x0001"
 [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"2")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "=r")
	(geu:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
		(match_operand:SI 2 "arith_operand" "rI")))]
  ""
  [(set (match_dup 0)
	(ltu:CC (match_dup 1)
		(match_dup 2)))
   (set (match_dup 0)
	(xor:SI (match_dup 0)
		(const_int 1)))]
  "")

(define_expand "sltu"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ltu:CC (match_dup 1)
		(match_dup 2)))]
  ""
  "
{
  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  /* fall through and generate default code */
}")

(define_insn "sltu_si"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ltu:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
		(match_operand:SI 2 "arith_operand" "rI")))]
  ""
  "sltu\\t%0,%z1,%2"
 [(set_attr "type"	"arith")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

(define_expand "sleu"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(leu:CC (match_dup 1)
		(match_dup 2)))]
  ""
  "
{
  extern rtx force_reg ();

  if (branch_type != CMP_SI)
    FAIL;

  /* set up operands from compare.  */
  operands[1] = branch_cmp[0];
  operands[2] = branch_cmp[1];

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) >= 32767)
    operands[2] = force_reg (SImode, operands[2]);

  /* fall through and generate default code */
}")

(define_insn "sleu_si"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(leu:CC (match_operand:SI 1 "reg_or_0_operand" "rJ,rJ")
		(match_operand:SI 2 "arith_operand" "r,I")))]
  "GET_CODE (operands[2]) != CONST_INT || INTVAL (operands[2]) < 32767"
  "@
   sltu\\t%0,%z2,%z1\;xori\\t%0,%0,0x0001
   sltu\\t%0,%z1,(%2+1)"
 [(set_attr "type"	"arith,arith")
   (set_attr "mode"	"SI,SI")
   (set_attr "length"	"2,1")])

(define_split
  [(set (match_operand:SI 0 "register_operand" "=r")
	(leu:CC (match_operand:SI 1 "reg_or_0_operand" "rJ")
		(match_operand:SI 2 "register_operand" "r")))]
  ""
  [(set (match_dup 0)
	(ltu:CC (match_dup 2)
		(match_dup 1)))
   (set (match_dup 0)
	(xor:SI (match_dup 0)
		(const_int 1)))]
  "")


;;
;;  ....................
;;
;;	UNCONDITIONAL BRANCHES
;;
;;  ....................

;; Unconditional branches.

(define_insn "jump"
  [(set (pc)
	(label_ref (match_operand 0 "" "")))]
  ""
  "*
{
  if (GET_CODE (operands[0]) == REG)
    return \"%*j\\t%0\";
  else
    return \"%*j\\t%l0\";
}"
  [(set_attr "type"	"jump")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "indirect_jump"
  [(set (pc) (match_operand:SI 0 "register_operand" "r"))]
  ""
  "%*j\\t%0"
  [(set_attr "type"	"jump")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "tablejump"
  [(set (pc)
	(match_operand:SI 0 "register_operand" "r"))
   (use (label_ref (match_operand 1 "" "")))]
  ""
  "%*j\\t%0"
  [(set_attr "type"	"jump")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

;; Function return, only allow after optimization, so that we can
;; eliminate jumps to jumps if no stack space is used.

(define_insn "return"
  [(return)]
  "null_epilogue ()"
  "*
{
  operands[0] = gen_rtx (REG, SImode, GP_REG_FIRST + 31);
  return \"%*j\\t%0\";
}"
  [(set_attr "type"	"jump")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	FUNCTION CALLS
;;
;;  ....................

(define_insn "call"
  [(call (match_operand 0 "memory_operand" "m")
	 (match_operand 1 "" "i"))
   (clobber (reg:SI 31))]
  ""
  "*
{
  register rtx target = XEXP (operands[0], 0);

  if (GET_CODE (target) == SYMBOL_REF)
    return \"%*jal\\t%0\";

  else
    {
      operands[0] = target;
      operands[1] = gen_rtx (REG, SImode, GP_REG_FIRST + 31);
      return \"%*jal\\t%1,%0\";
    }
}"
  [(set_attr "type"	"call")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "call_value"
  [(set (match_operand 0 "" "=rf")
        (call (match_operand 1 "memory_operand" "m")
              (match_operand 2 "" "i")))
   (clobber (reg:SI 31))]
  ""
  "*
{
  register rtx target = XEXP (operands[1], 0);

  if (GET_CODE (target) == SYMBOL_REF)
    return \"%*jal\\t%1\";

  else
    {
      operands[1] = target;
      operands[2] = gen_rtx (REG, SImode, GP_REG_FIRST + 31);
      return \"%*jal\\t%2,%1\";
    }
}"
  [(set_attr "type"	"call")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])


;;
;;  ....................
;;
;;	MISC.
;;
;;  ....................
;;

(define_insn "nop"
  [(const_int 0)]
  ""
  "%(nop%)"
  [(set_attr "type"	"nop")
   (set_attr "mode"	"none")
   (set_attr "length"	"1")])

(define_insn "probe"
  [(mem:SI (reg:SI 29))]
  ""
  "*
{
  operands[0] = stack_pointer_rtx;
  return \"%[lw\\t%@,0(%1)\\t\\t# stack probe%]\";
}"
  [(set_attr "type"	"load")
   (set_attr "mode"	"SI")
   (set_attr "length"	"1")])

;;
;; Local variables:
;; mode:emacs-lisp
;; comment-start: ";; "
;; eval: (set-syntax-table (copy-sequence (syntax-table)))
;; eval: (modify-syntax-entry ?[ "(]")
;; eval: (modify-syntax-entry ?] ")[")
;; eval: (modify-syntax-entry ?{ "(}")
;; eval: (modify-syntax-entry ?} "){")
;; End:
