/* Functions related to building classes and their related objects.
   Copyright (C) 1987 Free Software Foundation, Inc.
   Contributed by Michael Tiemann (tiemann@cygnus.com)

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


/* High-level class interface. */

#include "config.h"
#include "tree.h"
#include "cplus-tree.h"
#include "flags.h"
#include "assert.h"
#include <stdio.h>
#include "cplus-class.h"

#include "obstack.h"
#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free

extern int xmalloc ();
extern void free ();

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

/* in decl.c.  */
extern tree lookup_tag_current_binding_level ();

/* in method.c.  */
extern void do_inline_function_hair ();

/* Way of stacking class types.  */
static tree *current_class_base, *current_class_stack;
static int current_class_stacksize;
int current_class_depth;

struct class_level
{
  /* The previous class level.  */
  struct class_level *level_chain;

  /* The class instance variable, as a PARM_DECL.  */
  tree decl;
  /* The class instance variable, as an object.  */
  tree object;
  /* The virtual function table pointer
     for the class instance variable.  */
  tree vtable_decl;

  /* Name of the current class.  */
  tree name;
  /* Type of the current class.  */
  tree type;

  /* Flags for this class level.  */
  int this_is_variable;
  int memoized_lookups;
  int save_memoized;
  int unused;
};

tree current_class_decl, C_C_D;	/* PARM_DECL: the class instance variable */
tree current_vtable_decl;

/* The following two can be derived from the previous one */
tree current_class_name;	/* IDENTIFIER_NODE: name of current class */
tree current_class_type;	/* _TYPE: the type of the current class */
static tree prev_class_type;	/* _TYPE: the previous type that was a class */

static tree get_vtable_name (), get_vfield_name ();
tree the_null_vtable_entry;

/* Way of stacking langauge names.  */
static tree *current_lang_base, *current_lang_stack;
static int current_lang_stacksize;

/* Names of languages we recognize.  */
tree lang_name_c, lang_name_cplusplus, lang_name_objc;
tree current_lang_name;

tree minus_one_node;

/* When layout out an aggregate type, the size of the
   basetypes (virtual and non-virtual) is passed to layout_record
   via this node.  */
static tree base_layout_decl;

#if 0
/* Make sure that the tag NAME is defined *in the current binding level*
   at least as a forward reference.
   CODE says which kind of tag NAME ought to be.

   Not used for C++.  Not maintained.  */

tree
start_struct (code, name)
     enum tree_code code;
     tree name;
{
  /* If there is already a tag defined at this binding level
     (as a forward reference), just return it.  */
  register tree ref = 0;

  if (name != 0)
    ref = lookup_tag (code, name, current_binding_level, 1);
  if (ref && TREE_CODE (ref) == code)
    {
      if (TYPE_FIELDS (ref))
	error ((code == UNION_TYPE ? "redefinition of `union %s'"
		: "redefinition of `struct %s'"),
	       IDENTIFIER_POINTER (name));

      return ref;
    }

  /* Otherwise create a forward-reference just so the tag is in scope.  */

  ref = make_lang_type (code);
  /* Must re-synch this with xref_tag if you are going to use it.  */
  assert (0);
  pushtag (name, ref);
  return ref;
}
#endif

/* Virtual baseclass things.  */
tree
build_vbase_pointer (exp, type)
     tree exp, type;
{
  char *name;

  name = (char *) alloca (TYPE_NAME_LENGTH (type) + sizeof (VBASE_NAME) + 1);
  sprintf (name, VBASE_NAME_FORMAT, TYPE_NAME_STRING (type));
  return build_component_ref (exp, get_identifier (name), 0, 0);
}

/* Build multi-level access to EXPR using hierarchy path PATH.
   CODE is PLUS_EXPR if we are going with the grain,
   and MINUS_EXPR if we are not (in which case, we cannot traverse
   virtual baseclass links).

   TYPE is the type we want this path to have on exit.

   ALIAS_THIS is non-zero if EXPR in an expression involving `this'.  */
tree
build_vbase_path (code, type, expr, path, alias_this)
     enum tree_code code;
     tree type;
     tree expr;
     tree path;
     int alias_this;
{
  register int changed = 0;
  tree last = NULL_TREE, last_virtual = NULL_TREE;
  int nonnull = 0;
  int fixed_type_p = resolves_to_fixed_type_p (expr, &nonnull);
  tree null_expr = 0, nonnull_expr = expr;
  tree basetype;
  tree offset = integer_zero_node;

  if (TREE_CHAIN (path))
    path = nreverse (copy_list (path));

  basetype = TREE_VALUE (path);
  while (path)
    {
      if (TREE_VIA_VIRTUAL (path))
	{
	  last_virtual = TYPE_MAIN_VARIANT (TREE_VALUE (path));
	  if (code == PLUS_EXPR)
	    {
	      changed = ! fixed_type_p;

	      if (changed)
		{
		  extern tree flag_assume_nonnull_objects;
		  tree ind;

		  if (last)
		    nonnull_expr = convert_pointer_to (TREE_VALUE (last), nonnull_expr);
		  ind = build_indirect_ref (nonnull_expr, 0);
		  nonnull_expr = build_vbase_pointer (ind, last_virtual);
		  if (nonnull == 0 && !flag_assume_nonnull_objects
		      && null_expr == NULL_TREE)
		    {
		      null_expr = build1 (NOP_EXPR, TYPE_POINTER_TO (last_virtual), integer_zero_node);
		      expr = build (COND_EXPR, TYPE_POINTER_TO (last_virtual),
				    build (EQ_EXPR, integer_type_node, expr,
					   integer_zero_node),
				    null_expr, nonnull_expr);
		    }
		}
	      /* else we'll figure out the offset below.  */

	      /* Happens in the case of parse errors.  */
	      if (nonnull_expr == error_mark_node)
		return error_mark_node;
	    }
	  else
	    {
	      error_with_aggr_type (last_virtual, "cannot cast up from virtual baseclass `%s'");
	      return error_mark_node;
	    }
	}
      last = path;
      path = TREE_CHAIN (path);
    }
  /* LAST is now the last basetype on the path.  */
  last = TREE_VALUE (last);

  /* A pointer to a virtual base member of a non-null object
     is non-null.  Therefore, we only need to test for zeroness once.
     Make EXPR the cannonical expression to deal with here.  */
  if (null_expr)
    {
      TREE_OPERAND (expr, 2) = nonnull_expr;
      TREE_TYPE (TREE_OPERAND (expr, 1)) = TREE_TYPE (nonnull_expr);
    }
  else
    expr = nonnull_expr;

  /* If we go through any virtual base pointers, make sure that
     casts to BASETYPE from the last virtual base class use
     the right value for BASETYPE.  */
  if (changed)
    {
      tree intype = TREE_TYPE (TREE_TYPE (expr));
      if (TYPE_MAIN_VARIANT (intype) == TYPE_MAIN_VARIANT (last))
	basetype = intype;
      else
	{
	  basetype = get_base_type (last, TYPE_MAIN_VARIANT (intype), 0);
	  offset = CLASSTYPE_OFFSET (basetype);
	}
    }
  else
    {
      if (last_virtual)
	offset = ASSOC_OFFSET (value_member (last_virtual,
					     CLASSTYPE_VBASECLASSES (basetype)));
      else
	offset = CLASSTYPE_OFFSET (last);

      code = PLUS_EXPR;
    }

  if (TREE_INT_CST_LOW (offset))
    {
      /* For multiple inheritance: if `this' can be set by any
	 function, then it could be 0 on entry to any function.
	 Preserve such zeroness here.  Otherwise, only in the
	 case of constructors need we worry, and in those cases,
	 it will be zero, or initialized to some legal value to
	 which we may add.  */
      if (nonnull == 0 && (alias_this == 0 || flag_this_is_variable))
	{
	  if (null_expr)
	    TREE_TYPE (null_expr) = type;
	  else
	    null_expr = build1 (NOP_EXPR, type, integer_zero_node);
	  if (TREE_SIDE_EFFECTS (expr))
	    expr = save_expr (expr);

	  return build (COND_EXPR, type,
			build (EQ_EXPR, integer_type_node, expr, integer_zero_node),
			null_expr,
			build (code, type, expr, offset));
	}
      else return build (code, type, expr, offset);
    }

  if (null_expr || TREE_CODE (expr) == NOP_EXPR)
    {
      TREE_TYPE (expr) = type;
      return expr;
    }
  else
    return build1 (NOP_EXPR, type, expr);
}

/* Virtual function things.  */

/* Virtual functions to be dealt with after laying out our
   base classes.  Usually this is used only when classes have virtual
   baseclasses, but it can happen also when classes have non-virtual
   baseclasses if the derived class overrides baseclass functions
   at different offsets.  */
static tree pending_hard_virtuals;
static int doing_hard_virtuals;

/* The names of the entries in the virtual table structure.  */
static tree delta_name, pfn_name;

/* Temporary assoc list to memoize lookups of the left-most non-virtual
   baseclass B in a lattice topped by T.  B can appear multiple times
   in the lattice.
   TREE_PURPOSE is B's TYPE_MAIN_VARIANT.
   TREE_VALUE is the path by which B is reached from T.
   TREE_TYPE is B's real type.

   If TREE_TYPE is NULL_TREE, it means that B was reached via
   a virtual baseclass.
   N.B.: This list consists of nodes on the temporary obstack.  */
static tree leftmost_baseclasses;

/* Build an entry in the virtual function table.
   DELTA is the offset for the `this' pointer.
   PFN is an ADDR_EXPR containing a pointer to the virtual function.
   Note that the index (DELTA2) in the virtual function table
   is always 0.  */
tree
build_vtable_entry (delta, pfn)
     tree delta, pfn;
{
  tree elems = tree_cons (NULL_TREE, delta,
			  tree_cons (NULL_TREE, integer_zero_node,
				     build_tree_list (NULL_TREE, pfn)));
  tree entry = build (CONSTRUCTOR, vtable_entry_type, NULL_TREE, elems);
  TREE_CONSTANT (entry) = 1;
  TREE_STATIC (entry) = 1;
  TREE_READONLY (entry) = 1;

#ifdef GATHER_STATISTICS
  n_vtable_entries += 1;
#endif

  return entry;
}

/* Given an object INSTANCE, return an expression which yields
   the virtual function corresponding to INDEX.  There are many special
   cases for INSTANCE which we take care of here, mainly to avoid
   creating extra tree nodes when we don't have to.  */
tree
build_vfn_ref (ptr_to_instptr, instance, index)
     tree *ptr_to_instptr, instance;
     tree index;
{
  extern int building_cleanup;
  tree vtbl, aref;
  tree basetype = TREE_TYPE (instance);

  if (TREE_CODE (basetype) == REFERENCE_TYPE)
    basetype = TREE_TYPE (basetype);

  if (instance == C_C_D)
    {
      if (current_vtable_decl == NULL_TREE
	  || current_vtable_decl == error_mark_node
	  || get_base_type (DECL_FCONTEXT (CLASSTYPE_VFIELD (current_class_type)), basetype, 0) == NULL_TREE)
	vtbl = build_indirect_ref (build_vfield_ref (instance, basetype));
      else
	vtbl = current_vtable_decl;
    }
  else
    {
      if (optimize)
	{
	  /* Try to figure out what a reference refers to, and
	     access its virtual function table directly.  */
	  tree ref = NULL_TREE;

	  if (TREE_CODE (instance) == INDIRECT_REF
	      && TREE_CODE (TREE_TYPE (TREE_OPERAND (instance, 0))) == REFERENCE_TYPE)
	    ref = TREE_OPERAND (instance, 0);
	  else if (TREE_CODE (TREE_TYPE (instance)) == REFERENCE_TYPE)
	    ref = instance;

	  if (ref && TREE_CODE (ref) == VAR_DECL
	      && DECL_INITIAL (ref))
	    {
	      tree init = DECL_INITIAL (ref);

	      while (TREE_CODE (init) == NOP_EXPR
		     || TREE_CODE (init) == NON_LVALUE_EXPR)
		init = TREE_OPERAND (init, 0);
	      if (TREE_CODE (init) == ADDR_EXPR)
		{
		  init = TREE_OPERAND (init, 0);
		  if (IS_AGGR_TYPE (TREE_TYPE (init))
		      && (TREE_CODE (init) == PARM_DECL
			  || TREE_CODE (init) == VAR_DECL))
		    instance = init;
		}
	    }
	}

      if (IS_AGGR_TYPE (TREE_TYPE (instance))
	  && (TREE_CODE (instance) == RESULT_DECL
	      || TREE_CODE (instance) == PARM_DECL
	      || TREE_CODE (instance) == VAR_DECL))
	vtbl = CLASS_ASSOC_VTABLE (basetype);
      else
	vtbl = build_indirect_ref (build_vfield_ref (instance, basetype), 0);
    }
  aref = build_array_ref (vtbl, index);
  if (!building_cleanup && TREE_CODE (aref) == INDIRECT_REF)
    TREE_OPERAND (aref, 0) = save_expr (TREE_OPERAND (aref, 0));

  *ptr_to_instptr = build (PLUS_EXPR, TREE_TYPE (*ptr_to_instptr),
			   *ptr_to_instptr,
			   convert (integer_type_node, build_component_ref (aref, delta_name, 0, 0)));
  return build_component_ref (aref, pfn_name, 0, 0);
}

/* Build a virtual function for type TYPE.
   If ASSOC is non-NULL, build the vtable starting with the intial
   approximation that it is the same as the one which is the head of
   the assocation list.  */
static tree
build_vtable (assoc, type)
     tree assoc, type;
{
  tree name = get_vtable_name (type);
  tree virtuals, decl;

  if (assoc)
    {
      virtuals = copy_list (ASSOC_VIRTUALS (assoc));
      decl = build_decl (VAR_DECL, name, TREE_TYPE (ASSOC_VTABLE (assoc)));
    }
  else
    {
      virtuals = NULL_TREE;
      decl = build_decl (VAR_DECL, name, void_type_node);
    }

#ifdef GATHER_STATISTICS
  n_vtables += 1;
  n_vtable_elems += list_length (virtuals);
#endif

  if (write_virtuals >= 2)
    {
      if (CLASSTYPE_INTERFACE_UNKNOWN (type) == 0)
	{
	  TREE_PUBLIC (decl) = CLASSTYPE_VTABLE_NEEDS_WRITING (type);
	  TREE_EXTERNAL (decl) = ! CLASSTYPE_VTABLE_NEEDS_WRITING (type);
	}
    }
  else if (write_virtuals != 0)
    TREE_PUBLIC (decl) = 1;
  if (write_virtuals < 0)
    TREE_EXTERNAL (decl) = 1;

  IDENTIFIER_GLOBAL_VALUE (name) = decl = pushdecl_top_level (decl);
  /* Initialize the association list for this type, based
     on our first approximation.  */
  CLASS_ASSOC_VTABLE (type) = decl;
  CLASS_ASSOC_VIRTUALS (type) = virtuals;

  TREE_STATIC (decl) = 1;
  DECL_ALIGN (decl) = MAX (TYPE_ALIGN (double_type_node),
			   DECL_ALIGN (decl));

  if (assoc && write_virtuals >= 0)
    DECL_VIRTUAL_P (decl) = 1;
  /* Remember which class this vtable is really for.  */
  DECL_VPARENT (decl) = type;
  DECL_CONTEXT (decl) = type;
  CLASSTYPE_MARKED3 (type) = 1;
  CLASSTYPE_MARKED4 (type) = 1;
  return decl;
}

/* Give TYPE a new virtual function table which is initialized
   with a skeleton-copy of its original initialization.  The only
   entry that changes is the `delta' entry, so we can really
   share a lot of structure.

   FOR_TYPE is the derived type which caused this table to
   be needed.

   ASSOC is the type association which provided TYPE for FOR_TYPE.

   The way we update BASE_ASSOC's vtable information is just to change the
   association information in FOR_TYPE's association list.  */
static void
prepare_fresh_vtable (assoc, base_assoc, for_type)
     tree assoc, base_assoc, for_type;
{
  tree basetype = ASSOC_TYPE (assoc);
  tree orig_decl = ASSOC_VTABLE (assoc);
  tree name = build_type_pathname (VTABLE_NAME_FORMAT, basetype, for_type);
  tree new_decl = build_decl (VAR_DECL, name, TREE_TYPE (orig_decl));
  tree path;
  int result;

  assert (TREE_USED (assoc) == 0);

  /* Remember which class this vtable is really for.  */
  DECL_VPARENT (new_decl) = ASSOC_TYPE (base_assoc);
  DECL_CONTEXT (new_decl) = for_type;

  TREE_STATIC (new_decl) = 1;
  ASSOC_VTABLE (assoc) = pushdecl_top_level (new_decl);
  DECL_VIRTUAL_P (new_decl) = 1;
  DECL_ALIGN (new_decl) = DECL_ALIGN (orig_decl);

  /* Make fresh virtual list, so we can smash it later.  */
  assert (ASSOC_VIRTUALS (assoc));
  ASSOC_VIRTUALS (assoc) = copy_list (ASSOC_VIRTUALS (assoc));
  /* Install the value for `headof' if that's what we're doing.  */
  if (flag_dossier)
    TREE_VALUE (TREE_CHAIN (ASSOC_VIRTUALS (assoc)))
      = build_vtable_entry (size_binop (MINUS_EXPR, integer_zero_node, ASSOC_OFFSET (assoc)),
			    FNADDR_FROM_VTABLE_ENTRY (TREE_VALUE (TREE_CHAIN (ASSOC_VIRTUALS (assoc)))));

#ifdef GATHER_STATISTICS
  n_vtables += 1;
  n_vtable_elems += list_length (ASSOC_VIRTUALS (assoc));
#endif

  /* Set `new_decl's PUBLIC and EXTERNAL bits.  */
  if (write_virtuals >= 2)
    {
      if (CLASSTYPE_INTERFACE_UNKNOWN (for_type) == 0)
	{
	  TREE_PUBLIC (new_decl) = CLASSTYPE_VTABLE_NEEDS_WRITING (for_type);
	  TREE_EXTERNAL (new_decl) = ! CLASSTYPE_VTABLE_NEEDS_WRITING (for_type);
	}
    }
  else if (write_virtuals > 0)
    TREE_PUBLIC (new_decl) = 1;
  else if (write_virtuals < 0)
    TREE_EXTERNAL (new_decl) = 1;

  CLASSTYPE_MARKED3 (basetype) = 1;
  CLASSTYPE_MARKED4 (basetype) = 1;

  /* Mark all types between FOR_TYPE and TYPE as having been
     touched, so that if we change virtual function table entries,
     new vtables will be initialized.  We may reach the virtual
     baseclass via ambiguous intervening baseclasses.  This
     loop makes sure we get through to the actual baseclass we marked.

     Also, update the vtable entries to reflect the overrides
     of the top-most class (short of the top type).  */

  do
    {
      result = get_base_distance (basetype, for_type, 0, &path);
      for_type = TREE_VALUE (path);
      while (path)
	{
	  CLASSTYPE_MARKED3 (TREE_VALUE (path)) = 1;
	  if (TREE_CHAIN (path)
	      && CLASSTYPE_VFIELD (TREE_VALUE (path)) != NULL_TREE
	      && (DECL_NAME (CLASSTYPE_VFIELD (ASSOC_VALUE (assoc)))
		  == DECL_NAME (CLASSTYPE_VFIELD (TREE_VALUE (path))))
	      /* This is the baseclass just before the original FOR_TYPE.  */
	      && TREE_CHAIN (TREE_CHAIN (path)) == NULL_TREE)
	    {
	      tree old_virtuals = TREE_CHAIN (ASSOC_VIRTUALS (assoc));
	      tree new_virtuals = TREE_CHAIN (CLASS_ASSOC_VIRTUALS (TREE_VALUE (path)));
	      if (flag_dossier)
		{
		  old_virtuals = TREE_CHAIN (old_virtuals);
		  new_virtuals = TREE_CHAIN (new_virtuals);
		}
	      while (old_virtuals)
		{
		  TREE_VALUE (old_virtuals) = TREE_VALUE (new_virtuals);
		  old_virtuals = TREE_CHAIN (old_virtuals);
		  new_virtuals = TREE_CHAIN (new_virtuals);
		}
	    }
	  path = TREE_CHAIN (path);
	}
    }
  while (result == -2);
}

/* Access the virtual function table entry that logically
   contains BASE_FNDECL.  VIRTUALS is the virtual function table's
   initializer.  */
static tree
get_vtable_entry (virtuals, base_fndecl)
     tree virtuals, base_fndecl;
{
  int i = (HOST_BITS_PER_INT >= BITS_PER_WORD
#ifdef VTABLE_USES_MASK
	   && 0
#endif
	   ? (TREE_INT_CST_LOW (DECL_VINDEX (base_fndecl))
	      & ((1<<(BITS_PER_WORD-1))-1))
	   : TREE_INT_CST_LOW (DECL_VINDEX (base_fndecl)));

#ifdef GATHER_STATISTICS
  n_vtable_searches += i;
#endif

  while (i > 0)
    {
      virtuals = TREE_CHAIN (virtuals);
      i -= 1;
    }
  return virtuals;
}

/* Put new entry ENTRY into virtual function table initializer
   VIRTUALS.  The virtual function table is for type CONTEXT.

   Also update DECL_VINDEX (FNDECL).  */

static void
modify_vtable_entry (old_entry_in_list, new_entry, fndecl, context)
     tree old_entry_in_list, new_entry, fndecl, context;
{
  tree base_pfn = FNADDR_FROM_VTABLE_ENTRY (TREE_VALUE (old_entry_in_list));
  tree vindex;

  /* We can't put in the really right offset information
     here, since we have not yet laid out the class to
     take into account virtual base classes.  */
  TREE_VALUE (old_entry_in_list) = new_entry;
  vindex = DECL_VINDEX (TREE_OPERAND (base_pfn, 0));
  if (TREE_CODE (DECL_VINDEX (fndecl)) != INTEGER_CST)
    SET_DECL_VINDEX (fndecl, vindex);
  else
    {
      if (! tree_int_cst_equal (DECL_VINDEX (fndecl), vindex))
	{
	  tree elts = CONSTRUCTOR_ELTS (new_entry);
	  tree vfield = CLASSTYPE_VFIELD (context);

	  if (! doing_hard_virtuals)
	    {
	      pending_hard_virtuals
		= tree_cons (fndecl, FNADDR_FROM_VTABLE_ENTRY (new_entry),
			     pending_hard_virtuals);
	      TREE_TYPE (pending_hard_virtuals) = TREE_OPERAND (base_pfn, 0);
	      return;
	    }

	  abort ();
	  /* Compute the relative offset of vtable we are really looking for.  */
	  TREE_VALUE (elts) = size_binop (PLUS_EXPR,
					  size_int (TREE_INT_CST_LOW (DECL_FIELD_BITPOS (vfield))
/* ??? This may be wrong. */
						    / BITS_PER_UNIT),
					  TREE_VALUE (elts));
	  /* Say what index to use when we use that vtable.  */
#ifndef VTABLE_USES_MASK
	  vindex = build_int_2 (TREE_INT_CST_LOW (vindex) & ~(1 << (BITS_PER_WORD -1)), 0);
#endif
	  TREE_VALUE (TREE_CHAIN (elts)) = vindex;
	}
    }
}

/* Modify virtual function tables in lattice topped by T to
   place FNDECL in tables which previously held BASE_FNDECL.
   PFN is just FNDECL wrapped in an ADDR_EXPR, so that it
   is suitable for placement directly into an initializer.

   All distinct virtual function tables that this type uses
   must be updated.  */
static void
modify_vtable_entries (t, fndecl, base_fndecl, pfn)
     tree t;
     tree fndecl, base_fndecl, pfn;
{
  tree base_offset, offset;
  tree context = DECL_CONTEXT (base_fndecl);
  tree vfield = CLASSTYPE_VFIELD (t);
  tree vfields, vbases;

  DECL_VCONTEXT (fndecl) = DECL_VCONTEXT (base_fndecl);

  offset = integer_zero_node;
  if (DECL_CONTEXT (fndecl) != t && TYPE_USES_COMPLEX_INHERITANCE (t))
    {
      offset = virtual_offset (DECL_CONTEXT (fndecl), CLASSTYPE_VBASECLASSES (t), offset);
      if (offset == NULL_TREE)
	{
	  tree assoc = assoc_value (DECL_CONTEXT (fndecl), t, 0);
	  assert (assoc != NULL_TREE);
	  offset = ASSOC_OFFSET (assoc);
	}
    }

  /* For each layer of base class (i.e., the first base class, and each
     virtual base class from that one), modify the virtual function table
     of the derived class to contain the new virtual function.
     A class has as many vfields as it has virtual base classes (total).  */
  for (vfields = CLASSTYPE_VFIELDS (t); vfields; vfields = TREE_CHAIN (vfields))
    {
      int normal = 1;
      tree assoc, this_offset;
      tree base, path;

      /* Find the right base class for this derived class, call it BASE.  */
      base = TREE_VALUE (vfields);

      if (base != context)
	{
	  /* If BASE_FNDECL is not contained in the vtable accessed by
	     the vslot, don't try to modify the vtable.
	     
	     Virtual functions from virtual baseclasses are not in derived
	     virtual function tables.  This is an implementation decision;
	     it keeps there from being a combinatorial exposion in the
	     number of different vtables which must be maintained.  */

	  if (get_base_distance (base, context, 0, 0) == -1)
	    continue;

	  /* BASE_FNDECL is defined in a class derived from
	     the base class owning this VFIELD.  */
	}
      /* Get the path starting from the deepest base class CONTEXT
	 of T (i.e., first defn of BASE_FNDECL).  */
      get_base_distance (context, t, 0, &path);

      /* Get our best approximation of what to use for constructing
	 the virtual function table for T.  */
      do
	{
	  /* Walk from base toward derived, stopping at the
	     most derived baseclass that matters.  That baseclass
	     is exactly the one which provides the vtable along
	     the VFIELD spine, but no more.  */
	  if (TREE_VIA_VIRTUAL (path))
	    {
	      base = TREE_VALUE (path);
	      assoc = value_member (TYPE_MAIN_VARIANT (base), CLASSTYPE_VBASECLASSES (t));
	      break;
	    }
	  if (TREE_CHAIN (path) == NULL_TREE
	      || (CLASSTYPE_BASECLASS (TREE_VALUE (TREE_CHAIN (path)), 0)
		  != TREE_VALUE (path))
	      || TREE_CHAIN (TREE_CHAIN (path)) == NULL_TREE)
	    {
	      base = TREE_VALUE (path);
	      assoc = assoc_value (TYPE_MAIN_VARIANT (base), t, 1);
	      break;
	    }
	  path = TREE_CHAIN (path);
	}
      while (1);

      /* Find the right offset for the this pointer based on the base
	 class we just found.  */
      base_offset = ASSOC_OFFSET (assoc);
      if (base_offset == integer_zero_node)
	this_offset = offset;
      else
	this_offset = size_binop (MINUS_EXPR, offset, base_offset);

      /* Make sure we can modify the derived association with immunity.  */
      if (TREE_USED (CLASSTYPE_ASSOC (t)))
	CLASSTYPE_ASSOC (t) = copy_assoc (CLASSTYPE_ASSOC (t));

      /* We call this case NORMAL iff this virtual function table
	 pointer field has its storage reserved in this class.
	 This is normally the case without virtual baseclasses
	 or off-center multiple baseclasses.  */
      normal = (vfield != NULL_TREE
		&& TREE_VALUE (vfields) == DECL_FCONTEXT (vfield)
		&& (TREE_PURPOSE (vfields) == NULL_TREE
		    || ! TREE_VIA_VIRTUAL (TREE_PURPOSE (vfields))));

      if (normal && TREE_PURPOSE (vfields))
	/* Everything looks normal so far...check that we are really
	   working from VFIELD's basetype, and not some other appearance
	   of that basetype in the lattice.  */
	normal = (TREE_PURPOSE (vfields) == get_base_type (TREE_VALUE (vfields), t, 0));

      if (normal)
	{
	  /* In this case, it is *type*'s vtable we are modifying.
	     We start with the approximation that it's vtable is that
	     of the immediate base class.  */
	  context = t;
	  if (! CLASSTYPE_MARKED4 (t))
	    build_vtable (CLASSTYPE_ASSOC (DECL_CONTEXT (vfield)), t);
	  assoc = CLASSTYPE_ASSOC (t);
	}
      else
	{
	  /* This is our very own copy of `basetype' to play with.
	     Later, we will fill in all the virtual functions
	     that override the virtual functions in these base classes
	     which are not defined by the current type.  */
	  if (! CLASSTYPE_MARKED4 (ASSOC_TYPE (assoc)))
	    prepare_fresh_vtable (assoc, CLASSTYPE_ASSOC (base), t);
	}

      modify_vtable_entry (get_vtable_entry (ASSOC_VIRTUALS (assoc), base_fndecl),
			   build_vtable_entry (this_offset, pfn),
			   fndecl, context);
    }
  for (vbases = CLASSTYPE_VBASECLASSES (t); vbases; vbases = TREE_CHAIN (vbases))
    {
      tree this_offset;
      tree base, path;

      if (! ASSOC_VTABLE (vbases))
	/* There are only two ways that a type can fail to have
	   virtual functions: neither it nor any of its base
	   types define virtual functions (in which case
	   no updating need be done), or virtual functions
	   accessible to it come from virtual base classes
	   (in which case we have or will get them modified
	   in other passes of this loop).  */
	continue;

      base = TREE_VALUE (vbases);
      path = NULL_TREE;

      if (base != context
	  && get_base_distance (context, base, 0, &path) == -1)
	continue;

      /* Doesn't matter if not actually from this virtual base class,
         but shouldn't come from deeper virtual baseclasses.  The enclosing
	 loop should take care of such baseclasses.  */
      while (path)
	{
	  if (TREE_VIA_VIRTUAL (path))
	    goto skip;
	  path = TREE_CHAIN (path);
	}

      base_offset = ASSOC_OFFSET (vbases);
      if (base_offset == integer_zero_node)
	this_offset = offset;
      else
	this_offset = size_binop (MINUS_EXPR, offset, base_offset);

      /* Make sure we can modify the derived association with immunity.  */
      if (TREE_USED (CLASSTYPE_ASSOC (t)))
	CLASSTYPE_ASSOC (t) = copy_assoc (CLASSTYPE_ASSOC (t));

      /* This is our very own copy of `basetype' to play with.  */
      if (! CLASSTYPE_MARKED4 (ASSOC_TYPE (vbases)))
	{
	  tree context_assoc = assoc_value (context, base, 0);
	  prepare_fresh_vtable (vbases, context_assoc, t);
	}
      modify_vtable_entry (get_vtable_entry (ASSOC_VIRTUALS (vbases), base_fndecl),
			   build_vtable_entry (this_offset, pfn),
			   fndecl, context);
    skip: {}
    }
}

static tree
add_virtual_function (pending_virtuals, has_virtual, x)
     tree pending_virtuals;
     int *has_virtual;
     tree x;
{
  int debug_vbase = 1;

  /* FUNCTION_TYPEs and OFFSET_TYPEs no longer freely
     convert to void *.  Make such a conversion here.  */
  tree vfn = build1 (ADDR_EXPR, ptr_type_node, x);
  TREE_CONSTANT (vfn) = 1;
  TREE_ADDRESSABLE (x) = CLASSTYPE_VTABLE_NEEDS_WRITING (current_class_type);

  /* If the virtual function is a redefinition of a prior one,
     figure out in which base class the new definition goes,
     and if necessary, make a fresh virtual function table
     to hold that entry.  */
  if (DECL_VINDEX (x) == NULL_TREE)
    {
      tree entry = build_vtable_entry (integer_zero_node, vfn);

      /* Build a new INT_CST for this DECL_VINDEX.  */
#ifdef VTABLE_USES_MASK
      SET_DECL_VINDEX (x, build_int_2 (++(*has_virtual), 0));
#else
      SET_DECL_VINDEX (x, build_int_2 (((1 << (BITS_PER_WORD - 1)) | ++(*has_virtual)), ~0));
#endif
      pending_virtuals = tree_cons (DECL_VINDEX (x), entry, pending_virtuals);
    }
  /* Happens if declared twice in class.  We will give error
     later.  */
  else if (TREE_CODE (DECL_VINDEX (x)) == INTEGER_CST)
    return pending_virtuals;
  else if (debug_vbase && TYPE_USES_VIRTUAL_BASECLASSES (current_class_type))
    {
      /* Need an entry in some other virtual function table.
         Deal with this after we have laid out our virtual base classes.  */
      pending_hard_virtuals = temp_tree_cons (x, vfn, pending_hard_virtuals);
    }
  else
    {
      /* Need an entry in some other virtual function table.
         We can do this now.  */
      tree base_fndecl_list = DECL_VINDEX (x), base_fndecls, prev = 0;
      tree vtable_context = DECL_FCONTEXT (CLASSTYPE_VFIELD (current_class_type));
      tree true_base_fndecl = 0;

      /* First assign DECL_VINDEX from the base vfn with which
	 we share our vtable.  */
      base_fndecls = base_fndecl_list;
      while (base_fndecls)
	{
	  if (TREE_CHAIN (base_fndecls) == NULL_TREE
	      || DECL_FCONTEXT (CLASSTYPE_VFIELD (DECL_CONTEXT (TREE_VALUE (base_fndecls)))) == vtable_context)
	    {
	      true_base_fndecl = TREE_VALUE (base_fndecls);
	      modify_vtable_entries (current_class_type, x,
				     true_base_fndecl, vfn);
	      if (prev)
		TREE_CHAIN (prev) = TREE_CHAIN (base_fndecls);
	      else
		base_fndecl_list = prev;
	      break;
	    }
	  prev = base_fndecls;
	  base_fndecls = TREE_CHAIN (base_fndecls);
	}

      /* Now fill in the rest of the vtables.  */
      base_fndecls = base_fndecl_list;
      while (base_fndecls)
	{
	  /* If we haven't found one we like, first one wins.  */
	  if (true_base_fndecl == 0)
	    true_base_fndecl = TREE_VALUE (base_fndecls);

	  modify_vtable_entries (current_class_type, x,
				 TREE_VALUE (base_fndecls), vfn);
	  base_fndecls = TREE_CHAIN (base_fndecls);
	}

      DECL_VCONTEXT (x) = DECL_VCONTEXT (true_base_fndecl);
    }
  return pending_virtuals;
}

/* Obstack on which to build the vector of class methods.  */
struct obstack class_obstack;
extern struct obstack *current_obstack;

/* Add method METHOD to class TYPE.  This is used when a method
   has been defined which did not initially appear in the class definition,
   and helps cut down on spurious error messages.

   FIELDS is the entry in the METHOD_VEC vector entry of the class type where
   the method should be added.  */
void
add_method (type, fields, method)
     tree type, *fields, method;
{
  /* We must make a copy of METHOD here, since we must be sure that
     we have exclusive title to this method's DECL_CHAIN.  */
  int temp = allocation_temporary_p ();
  tree decl;

  if (temp)
    end_temporary_allocation ();
  {
    decl = copy_node (method);
    if (DECL_RTL (decl) == 0)
      make_function_rtl (decl);
  }

  if (fields && *fields)
    {
      /* Take care not to hide destructor.  */
      DECL_CHAIN (decl) = DECL_CHAIN (*fields);
      DECL_CHAIN (*fields) = decl;
    }
  else if (CLASSTYPE_METHOD_VEC (type) == 0)
    {
      tree method_vec = make_node (TREE_VEC);
      if (TYPE_IDENTIFIER (type) == DECL_ORIGINAL_NAME (decl))
	{
	  TREE_VEC_ELT (method_vec, 0) = decl;
	  TREE_VEC_LENGTH (method_vec) = 1;
	}
      else
	{
	  obstack_free (current_obstack, method_vec);
	  obstack_blank (current_obstack, sizeof (struct tree_vec) + sizeof (tree *));
	  TREE_VEC_ELT (method_vec, 1) = decl;
	  TREE_VEC_LENGTH (method_vec) = 2;
	  obstack_finish (current_obstack);
	}
      CLASSTYPE_METHOD_VEC (type) = method_vec;
    }
  else
    {
      tree method_vec = CLASSTYPE_METHOD_VEC (type);
      int len = TREE_VEC_LENGTH (method_vec);

      /* Adding a new ctor or dtor.  */
      if (TYPE_IDENTIFIER (type) == DECL_ORIGINAL_NAME (decl))
	TREE_VEC_ELT (method_vec, 0) = decl;
      else
	{
	  tree *end = (tree *)obstack_next_free (&class_obstack);
	  if (end != TREE_VEC_END (method_vec))
	    {
	      tree tmp_vec = copy_node (method_vec);
	      obstack_copy (current_obstack, &TREE_VEC_ELT (method_vec, 1), len);
	      obstack_blank (current_obstack, sizeof (tree *));
	      obstack_finish (current_obstack);
	      method_vec = tmp_vec;
	    }
	  else
	    {
	      /* We can easily extend the last such method_vec created.  */
	      obstack_free (&class_obstack, method_vec);
	      obstack_blank (&class_obstack,
			     ((char *)end - (char *)method_vec) + sizeof (tree *));
	      method_vec = (tree)obstack_base (&class_obstack);
	      obstack_finish (&class_obstack);
	    }
	  TREE_VEC_ELT (method_vec, len) = decl;
	  TREE_VEC_LENGTH (method_vec) = len + 1;
	  CLASSTYPE_METHOD_VEC (type) = method_vec;

	  if (TYPE_BASETYPES (type) && CLASSTYPE_BASELINK_VEC (type))
	    {
	      /* ??? May be better to know whether these can be extended?  */
	      tree baselink_vec = copy_node (CLASSTYPE_BASELINK_VEC (type));

	      obstack_copy (current_obstack, &TREE_VEC_ELT (CLASSTYPE_BASELINK_VEC (type), 1), len);
	      TREE_VEC_ELT (baselink_vec, len) = 0;
	      TREE_VEC_LENGTH (baselink_vec) = len + 1;
	      CLASSTYPE_BASELINK_VEC (type) = baselink_vec;
	    }
	}
    }
  DECL_CONTEXT (decl) = type;
  DECL_VCONTEXT (decl) = type;

  if (temp)
    resume_temporary_allocation ();
}

/* Subroutines of finish_struct.  */

/* Look through the list of fields for this struct, deleting
   duplicates as we go.  This must be recursive to handle
   anonymous unions.

   FIELD is the field which may not appear anywhere in FIELDS.
   FIELD_PTR, if non-null, is the starting point at which
   chained deletions may take place.
   The value returned is the first acceptable entry found
   in FIELDS.

   Note that anonymous fields which are not of UNION_TYPE are
   not duplicates, they are just anonymous fields.  This happens
   when we have unnamed bitfields, for example.  */
static tree
delete_duplicate_fields_1 (field, field_ptr, fields)
     tree field, *field_ptr, fields;
{
  tree x;
  tree prev = field_ptr ? *field_ptr : 0;
  if (DECL_NAME (field) == 0)
    {
      if (TREE_CODE (TREE_TYPE (field)) != UNION_TYPE)
	return fields;

      for (x = TYPE_FIELDS (TREE_TYPE (field)); x; x = TREE_CHAIN (x))
	fields = delete_duplicate_fields_1 (x, field_ptr, fields);
      if (prev)
	TREE_CHAIN (prev) = fields;
      return fields;
    }
  else
    {
      for (x = fields; x; prev = x, x = TREE_CHAIN (x))
	{
	  if (DECL_NAME (x) == 0)
	    {
	      if (TREE_CODE (TREE_TYPE (x)) != UNION_TYPE)
		continue;
	      TYPE_FIELDS (TREE_TYPE (x))
		= delete_duplicate_fields_1 (field, 0, TYPE_FIELDS (TREE_TYPE (x)));
	      if (TYPE_FIELDS (TREE_TYPE (x)) == 0)
		{
		  if (prev == 0)
		    fields = TREE_CHAIN (fields);
		  else
		    TREE_CHAIN (prev) = TREE_CHAIN (x);
		}
	    }
	  else
	    {
	      if (DECL_NAME (field) == DECL_NAME (x))
		{
		  if (TREE_CODE (field) == CONST_DECL
		      && TREE_CODE (x) == CONST_DECL)
		    error_with_decl (x, "duplicate enum value `%s'");
		  else if (TREE_CODE (field) == CONST_DECL
			   || TREE_CODE (x) == CONST_DECL)
		    error_with_decl (x, "duplicate field `%s' (as enum and non-enum)");
		  else
		    error_with_decl (x, "duplicate member `%s'");
		  if (prev == 0)
		    fields = TREE_CHAIN (fields);
		  else
		    TREE_CHAIN (prev) = TREE_CHAIN (x);
		}
	    }
	}
    }
  return fields;
}

static void
delete_duplicate_fields (fields)
     tree fields;
{
  tree x;
  for (x = fields; x && TREE_CHAIN (x); x = TREE_CHAIN (x))
    TREE_CHAIN (x) = delete_duplicate_fields_1 (x, &x, TREE_CHAIN (x));
}

/* Add OFFSET to all child types of T.

   OFFSET, which is a type offset, is number of bytes.

   Note that we don't have to worry about having two paths to the
   same base type, since this type owns its association list.  */
static void
propagate_basetype_offsets (for_type, t, offset)
     tree for_type, t;
     tree offset;
{
  int i, n_baselinks = CLASSTYPE_N_BASECLASSES (t);

  for (i = 0; i < n_baselinks; i++)
    if (! CLASSTYPE_VIA_VIRTUAL (t, i))
      {
	int j;
	tree basetype = CLASSTYPE_BASECLASS (t, i);
	tree assoc = assoc_value (TYPE_MAIN_VARIANT (basetype), for_type, 0);
	tree other_assoc;
	tree delta;

	for (j = i+1; j < n_baselinks; j++)
	  if (! CLASSTYPE_VIA_VIRTUAL (t, j))
	    {
	      /* The next basetype offset must take into account the space
		 between the classes, not just the size of each class.  */
	      delta = size_binop (MINUS_EXPR,
				  CLASSTYPE_OFFSET (CLASSTYPE_BASECLASS (t, j)),
				  CLASSTYPE_OFFSET (basetype));
	      break;
	    }

	if (CLASSTYPE_OFFSET_ZEROP (basetype))
	  basetype = build_classtype_variant (basetype, offset, 0);
	else
	  basetype = build_classtype_variant (basetype,
					      size_binop (PLUS_EXPR, CLASSTYPE_OFFSET (basetype), offset), 0);
	/* Now make our own copy of this base type we can munge.  */
	basetype = copy_node (basetype);
	copy_type_lang_specific (basetype);

	CLASSTYPE_BASECLASS (t, i) = basetype;
	other_assoc = assoc_value (TYPE_MAIN_VARIANT (basetype), for_type, 0);
	/* ??? Why aren't assocs with eq types the same?  */
	if (ASSOC_TYPE (other_assoc) != ASSOC_TYPE (assoc))
	  abort ();
	ASSOC_TYPE (assoc) = basetype;
	ASSOC_OFFSET (assoc) = CLASSTYPE_OFFSET (basetype);
	propagate_basetype_offsets (for_type, basetype, offset);

	/* Go to our next class that counts for offset propagation.  */
	i = j;
	if (i < n_baselinks)
	  offset = size_binop (PLUS_EXPR, offset, delta);
      }
}

/* Change the visibility of T::FDECL to VISIBILITY.
   Return 1 if change was legit, otherwise return 0.  */
static int
alter_visibility (t, fdecl, visibility)
     tree t;
     tree fdecl;
     enum visibility_type visibility;
{
  tree elem = purpose_member (t, DECL_VISIBILITY (fdecl));
  if (elem && TREE_VALUE (elem) != (tree)visibility)
    {
      if (TREE_CODE (TREE_TYPE (fdecl)) == FUNCTION_DECL)
	{
	  error_with_decl (TREE_TYPE (fdecl), "conflicting visibility specifications for method `%s', ignored");
	}
      else error ("conflicting visibility specifications for field `%s', ignored", IDENTIFIER_POINTER (DECL_NAME (fdecl)));
    }
  else if (TREE_PRIVATE (fdecl) && visibility != visibility_private)
    error_with_decl (fdecl, "cannot make private %s non-private");
  else if (TREE_PROTECTED (fdecl) && visibility == visibility_public)
		    
    error_with_decl (fdecl, "cannot make protected %s public");
  else if (elem == NULL_TREE)
    {
      DECL_VISIBILITY (fdecl) = tree_cons (t, (tree)visibility,
					   DECL_VISIBILITY (fdecl));
      return 1;
    }
  return 0;
}

static tree
get_vfield_offset (type)
     tree type;
{
  return size_binop (PLUS_EXPR,
		     DECL_FIELD_BITPOS (CLASSTYPE_VFIELD (type)),
		     CLASSTYPE_OFFSET (type));
}

/* If FOR_TYPE needs to reinitialize virtual function table pointers
   for TYPE's sub-objects, add such reinitializations to BASE_INIT_LIST.
   Returns BASE_INIT_LIST appropriately modified.  */

static tree
maybe_fixup_vptrs (for_type, type, base_init_list)
     tree for_type, type, base_init_list;
{
  /* Now reinitialize any slots that don't fall under our virtual
     function table pointer.  */
  tree vfields = CLASSTYPE_VFIELDS (type);
  while (vfields)
    {
      tree basetype = get_base_type (TREE_VALUE (vfields), for_type, 0);
      if (CLASSTYPE_NEEDS_VIRTUAL_REINIT (basetype))
	{
	  tree base_offset = get_vfield_offset (basetype);
	  if (! tree_int_cst_equal (base_offset, get_vfield_offset (for_type))
	      && ! tree_int_cst_equal (base_offset, get_vfield_offset (type)))
	    base_init_list = tree_cons (error_mark_node, basetype,
					base_init_list);
	}
      vfields = TREE_CHAIN (vfields);
    }
  return base_init_list;
}

/* If TYPE does not have a constructor, then the compiler must
   manually deal with all of the initialization this type requires.

   If a base initializer exists only to fill in the virtual function
   table pointer, then we mark that fact with the TREE_VIRTUAL bit.
   This way, we avoid multiple initializations of the same field by
   each virtual function table up the class hierarchy.

   Virtual base class pointers are not initialized here.  They are
   initialized only at the "top level" of object creation.  If we
   initialized them here, we would have to skip a lot of work.  */

static void
build_class_init_list (type)
     tree type;
{
  tree base_init_list = NULL_TREE;
  tree member_init_list = NULL_TREE;

  /* Since we build member_init_list and base_init_list using
     tree_cons, backwards fields the all through work.  */
  tree x;
  int i, n_baseclasses = CLASSTYPE_N_BASECLASSES (type);

  for (x = TYPE_FIELDS (type); x; x = TREE_CHAIN (x))
    {
      if (TREE_CODE (x) != FIELD_DECL)
	continue;

      if (TYPE_NEEDS_CONSTRUCTING (TREE_TYPE (x))
	  || DECL_INITIAL (x) != NULL_TREE)
	member_init_list = tree_cons (x, type, member_init_list);
    }
  member_init_list = nreverse (member_init_list);

  /* We will end up doing this last.  Need special marker
     to avoid infinite regress.  */
  if (TYPE_VIRTUAL_P (type))
    {
      base_init_list = build_tree_list (error_mark_node, type);
      if (CLASSTYPE_NEEDS_VIRTUAL_REINIT (type) == 0)
	TREE_VALUE (base_init_list) = NULL_TREE;
      TREE_ADDRESSABLE (base_init_list) = 1;
    }

  /* Each base class which needs to have initialization
     of some kind gets to make such requests known here.  */
  for (i = n_baseclasses-1; i >= 0; i--)
    {
      tree basetype = CLASSTYPE_BASECLASS (type, i);
      tree blist;

      /* Don't initialize virtual baseclasses this way.  */
      if (TREE_VIA_VIRTUAL (basetype))
	continue;

      if (TYPE_HAS_CONSTRUCTOR (basetype))
	{
	  /* ...and the last shall come first...  */
	  base_init_list = maybe_fixup_vptrs (type, basetype, base_init_list);
	  base_init_list = tree_cons (NULL_TREE, basetype,
				      base_init_list);
	  continue;
	}

      if ((blist = CLASSTYPE_BASE_INIT_LIST (basetype)) == NULL_TREE)
	/* Nothing to initialize.  */
	continue;

      /* ...ditto...  */
      base_init_list = maybe_fixup_vptrs (type, basetype, base_init_list);

      /* This is normally true for single inheritance.
	 The win is we can shrink the chain of initializations
	 to be done by only converting to the actual type
	 we are interested in.  */
      if (TREE_VALUE (blist)
	  && TREE_CODE (TREE_VALUE (blist)) == RECORD_TYPE
	  && tree_int_cst_equal (CLASSTYPE_OFFSET (basetype),
				 CLASSTYPE_OFFSET (TREE_VALUE (blist))))
	{
	  if (base_init_list)
	    {
	      /* Does it do more than just fill in a
		 virtual function table pointer?  */
	      if (! TREE_ADDRESSABLE (blist))
		base_init_list = build_tree_list (blist, base_init_list);
	      /* Can we get by just with the virtual function table
		 pointer that it fills in?  */
	      else if (TREE_ADDRESSABLE (base_init_list)
		       && TREE_VALUE (base_init_list) == 0)
		base_init_list = blist;
	      /* Maybe, but it is not obvious as the previous case.  */
	      else if (! CLASSTYPE_NEEDS_VIRTUAL_REINIT (type))
		{
		  tree last = tree_last (base_init_list);
		  while (TREE_VALUE (last)
			 && TREE_CODE (TREE_VALUE (last)) == TREE_LIST)
		    last = tree_last (TREE_VALUE (last));
		  if (TREE_VALUE (last) == 0)
		    base_init_list = build_tree_list (blist, base_init_list);
		}
	    }
	  else
	    base_init_list = blist;
	}
      else
	{
	  /* The function expand_aggr_init knows how to do the
	     initialization of `basetype' without getting
	     an explicit `blist'.  */
	  if (base_init_list)
	    base_init_list = tree_cons (NULL_TREE, basetype, base_init_list);
	  else
	    base_init_list = CLASSTYPE_AS_LIST (basetype);
	}
    }

  if (base_init_list)
    if (member_init_list)
      CLASSTYPE_BASE_INIT_LIST (type) = build_tree_list (base_init_list, member_init_list);
    else
      CLASSTYPE_BASE_INIT_LIST (type) = base_init_list;
  else if (member_init_list)
    CLASSTYPE_BASE_INIT_LIST (type) = member_init_list;
}

struct base_info
{
  int has_virtual;
  int max_has_virtual;
  int n_ancestors;
  tree vfield;
  tree vfields;
  char needs_default_ctor;
  char cant_have_default_ctor;
  char needs_const_ctor;
  char cant_have_const_ctor;
};

/* Record information about type T derived from its base classes.
   Store most of that information in T itself, and place the
   remaining information in the struct BASE_INFO.

   Returns the index of the first base class to have virtual functions,
   or zero if no such base class.  */

static int
finish_base_struct (t, b)
     tree t;
     struct base_info *b;
{
  int i, n_baseclasses = CLASSTYPE_N_BASECLASSES (t);
  int first_vfn_base_index = -1;

  bzero (b, sizeof (struct base_info));

  for (i = 0; i < n_baseclasses; i++)
    {
      tree basetype = CLASSTYPE_BASECLASS (t, i);

      /* If the type of basetype is incomplete, then
	 we already complained about that fact
	 (and we should have fixed it up as well).  */
      if (TYPE_SIZE (basetype) == 0)
	{
	  int j;
	  /* The base type is of incomplete type.  It is
	     probably best to pretend that it does not
	     exist.  */
	  if (i == n_baseclasses-1)
	    CLASSTYPE_BASECLASS (t, i) = NULL_TREE;
	  TREE_VEC_LENGTH (TYPE_BASETYPES (t)) -= 1;
	  n_baseclasses -= 1;
	  for (j = i; j+1 < n_baseclasses; j++)
	    {
	      CLASSTYPE_BASECLASS (t, j) = CLASSTYPE_BASECLASS (t, j+1);
	      SET_CLASSTYPE_VIAS (t, j,
				  CLASSTYPE_VIA_PUBLIC (t, j+1),
				  CLASSTYPE_VIA_VIRTUAL (t, j+1));
	    }
	}

      if (TYPE_WRAP_TYPE (t) == NULL_TREE)
	TYPE_WRAP_TYPE (t) = TYPE_WRAP_TYPE (basetype);
      else if (TYPE_WRAP_TYPE (basetype)
	       && TYPE_WRAP_TYPE (t) != TYPE_WRAP_TYPE (basetype))
	/* Must have its own.  */
	TYPE_WRAP_TYPE (t) = error_mark_node;

      if (TYPE_HAS_DEFAULT_CONSTRUCTOR (basetype))
	b->needs_default_ctor = 1;
      else if (TYPE_HAS_CONSTRUCTOR (basetype))
	b->cant_have_default_ctor = 1;
      if (TYPE_GETS_CONST_INIT_REF (basetype))
	b->needs_const_ctor = 1;
      else if (TYPE_GETS_INIT_REF (basetype))
	b->cant_have_const_ctor = 1;

      CLASSTYPE_ALTERS_VISIBILITIES_P (t)
	|= CLASSTYPE_ALTERS_VISIBILITIES_P (basetype);

      b->n_ancestors += CLASSTYPE_N_SUPERCLASSES (basetype);
      TYPE_NEEDS_CONSTRUCTING (t) |= TYPE_NEEDS_CONSTRUCTING (basetype);
      TYPE_NEEDS_CONSTRUCTOR (t) |= TYPE_NEEDS_CONSTRUCTOR (basetype);
      TYPE_NEEDS_DESTRUCTOR (t) |= TYPE_NEEDS_DESTRUCTOR (basetype);
      TYPE_GETS_ASSIGNMENT (t) |= TYPE_GETS_ASSIGNMENT (basetype);
      TYPE_GETS_INIT_REF (t) |= TYPE_GETS_INIT_REF (basetype);

      TYPE_OVERLOADS_CALL_EXPR (t) |= TYPE_OVERLOADS_CALL_EXPR (basetype);
      TYPE_OVERLOADS_ARRAY_REF (t) |= TYPE_OVERLOADS_ARRAY_REF (basetype);

      if (! CLASSTYPE_OFFSET_ZEROP (basetype))
	{
	  /* Completely unshare potentially shared data, and
	     update what is ours.  */
	  tree assoc = assoc_value (TYPE_MAIN_VARIANT (basetype), t, 1);
	  basetype = copy_node (basetype);
	  copy_type_lang_specific (basetype);
	  CLASSTYPE_BASECLASS (t, i) = basetype;
	  ASSOC_TYPE (assoc) = basetype;

	  /* Propagate this offset through all the children.  Do this
	     before uniquizing baseclasses for virtual functions.  */
	  CLASSTYPE_ASSOC (basetype) = copy_assoc (CLASSTYPE_ASSOC (TYPE_MAIN_VARIANT (basetype)));
	  CLASS_ASSOC_OFFSET (basetype) = CLASSTYPE_OFFSET (basetype);
	  propagate_basetype_offsets (basetype, basetype, CLASSTYPE_OFFSET (basetype));
	}

      if (! CLASSTYPE_VIA_VIRTUAL (t, i))
	CLASSTYPE_N_SUPERCLASSES (t) += 1;

      if (TYPE_VIRTUAL_P (basetype))
	{
	  /* Don't borrow virtuals from virtual baseclasses.  */
	  if (TREE_VIA_VIRTUAL (basetype))
	    continue;

	  if (first_vfn_base_index < 0)
	    {
	      first_vfn_base_index = i;

	      b->has_virtual = CLASSTYPE_VSIZE (basetype);
	      b->vfield = CLASSTYPE_VFIELD (basetype);
	      b->vfields = CLASSTYPE_VFIELDS (basetype);
	      CLASSTYPE_VFIELD (t) = b->vfield;
	    }
	  else
	    {
	      /* Only add unique vfields, and flatten them out as we go.  */
	      tree vfields = CLASSTYPE_VFIELDS (basetype);
	      while (vfields)
		{
		  if (TREE_PURPOSE (vfields) == NULL_TREE
		      || ! TREE_VIA_VIRTUAL (TREE_PURPOSE (vfields)))
		    {
		      tree value = TREE_VALUE (vfields);
		      if (TYPE_MAIN_VARIANT (basetype) == value)
			b->vfields = tree_cons (basetype, value, b->vfields);
		      else
			b->vfields = tree_cons (get_base_type (value, basetype, 0),
						value, b->vfields);
		      TREE_TYPE (b->vfields) = basetype;
		    }
		  vfields = TREE_CHAIN (vfields);
		}

	      if (b->has_virtual == 0)
		{
		  first_vfn_base_index = i;
		  b->has_virtual = CLASSTYPE_VSIZE (basetype);
		  b->vfield = CLASSTYPE_VFIELD (basetype);
		  CLASSTYPE_VFIELD (t) = b->vfield;
		}
	    }
	}
    }

  {
    tree vfields;
    /* Find the base class with the largest number of virtual functions.  */
    for (vfields = b->vfields; vfields; vfields = TREE_CHAIN (vfields))
      {
	if (CLASSTYPE_VSIZE (TREE_VALUE (vfields)) > b->max_has_virtual)
	  b->max_has_virtual = CLASSTYPE_VSIZE (TREE_VALUE (vfields));
	if (TREE_TYPE (vfields)
	    && CLASSTYPE_VSIZE (TREE_TYPE (vfields)) > b->max_has_virtual)
	  b->max_has_virtual = CLASSTYPE_VSIZE (TREE_TYPE (vfields));
      }
  }

  if (b->vfield == 0)
    /* If all virtual functions come only from virtual baseclasses.  */
    return -1;
  return first_vfn_base_index;
}

static int
typecode_p (type, code)
     tree type;
     enum tree_code code;
{
  return (TREE_CODE (type) == code
	  || (TREE_CODE (type) == REFERENCE_TYPE
	      && TREE_CODE (TREE_TYPE (type)) == code));
}

/* Set memoizing fields and bits of T (and its variants) for later use.
   MAX_HAS_VIRTUAL is the largest size of any T's virtual function tables.  */
static void
finish_struct_bits (t, max_has_virtual)
     tree t;
     int max_has_virtual;
{
  int i, n_baseclasses = CLASSTYPE_N_BASECLASSES (t);
  tree method_vec = CLASSTYPE_METHOD_VEC (t);

  /* Fix up variants (if any).  */
  tree variants = TYPE_NEXT_VARIANT (t);
  while (variants)
    {
      /* These fields are in the _TYPE part of the node, not in
	 the TYPE_LANG_SPECIFIC component, so they are not shared.  */
      TYPE_HAS_CONSTRUCTOR (variants) = TYPE_HAS_CONSTRUCTOR (t);
      TYPE_HAS_DESTRUCTOR (variants) = TYPE_HAS_DESTRUCTOR (t);
      TYPE_NEEDS_CONSTRUCTOR (variants) = TYPE_NEEDS_CONSTRUCTOR (t);
      TYPE_NEEDS_CONSTRUCTING (variants) = TYPE_NEEDS_CONSTRUCTING (t);
      TYPE_NEEDS_DESTRUCTOR (variants) = TYPE_NEEDS_DESTRUCTOR (t);

      TYPE_USES_COMPLEX_INHERITANCE (variants) = TYPE_USES_COMPLEX_INHERITANCE (t);
      TYPE_VIRTUAL_P (variants) = TYPE_VIRTUAL_P (t);
      TYPE_USES_VIRTUAL_BASECLASSES (variants) = TYPE_USES_VIRTUAL_BASECLASSES (t);
      /* Copy whatever these are holding today.  */
      TYPE_MIN_VALUE (variants) = TYPE_MIN_VALUE (t);
      TYPE_MAX_VALUE (variants) = TYPE_MAX_VALUE (t);
      variants = TYPE_NEXT_VARIANT (variants);
    }

  if (n_baseclasses && max_has_virtual)
    {
      /* Done by `finish_struct' for classes without baseclasses.  */
      int has_abstract_virtuals = CLASSTYPE_ABSTRACT_VIRTUALS (t) != 0;
      if (has_abstract_virtuals == 0)
	for (i = n_baseclasses-1; i >= 0; i--)
	  has_abstract_virtuals
	    |= (CLASSTYPE_ABSTRACT_VIRTUALS (CLASSTYPE_BASECLASS (t, i)) != 0);
      if (has_abstract_virtuals)
	CLASSTYPE_ABSTRACT_VIRTUALS (t) = get_abstract_virtuals (t);
    }

  if (n_baseclasses)
    {
      /* Notice whether this class has type conversion functions defined.
	 Also report whether joining two types yields an ambiguity in the
	 virtual function table, e.g.,
	 
	 struct A { virtual int f (); };
	 struct B { virtual int f (); };
	 struct C : A, B { / * no f (); * / };	/ / error, ambiguous
	 */
      tree assoc = CLASSTYPE_ASSOC (t);
      int n_assocs = list_length (assoc);
      tree vbases = CLASSTYPE_VBASECLASSES (t), basetype;
      int n_vbases = list_length (vbases), j;

      build_mi_virtuals (n_assocs+n_vbases*n_baseclasses, max_has_virtual);
      /* Fill in virutal function table with values which do not come
	 "normal"ly, i.e., those which come from virtual and/or
	 non-leftmost base classes.  */
      for (i = 0; assoc; assoc = TREE_CHAIN (assoc))
	{
	  if (TREE_VIA_VIRTUAL (TREE_TYPE (assoc)))
	    /* Virtual functions from virtual baseclasses are done below.  */;
	  else if (CLASSTYPE_VSIZE (TREE_TYPE (assoc)))
	    {
	      tree virtuals = TREE_CHAIN (ASSOC_VIRTUALS (assoc));
	      if (flag_dossier)
		virtuals = TREE_CHAIN (virtuals);
	      add_mi_virtuals (++i, virtuals);
	    }
	}
      for (; vbases; vbases = TREE_CHAIN (vbases))
	{
	  basetype = TREE_VALUE (vbases);
	  if (CLASSTYPE_VSIZE (basetype))
	    for (j = n_baseclasses-1; j >= 0; j--)
	      {
		tree this_base = CLASSTYPE_BASECLASS (t, j);
		if (get_base_distance (basetype, TYPE_MAIN_VARIANT (this_base), 0, 0) != -1)
		  {
		    tree virtuals = TREE_CHAIN (ASSOC_VIRTUALS (vbases));
		    if (flag_dossier)
		      virtuals = TREE_CHAIN (virtuals);
		    add_mi_virtuals (++i, virtuals);
		  }
	      }
	}
      for (i = n_baseclasses-1; i >= 0; i--)
	{
	  basetype = CLASSTYPE_BASECLASS (t, i);

	  if (TYPE_HAS_CONVERSION (basetype))
	    {
	      TYPE_HAS_CONVERSION (t) = 1;
	      TYPE_HAS_INT_CONVERSION (t) |= TYPE_HAS_INT_CONVERSION (basetype);
	      TYPE_HAS_REAL_CONVERSION (t) |= TYPE_HAS_REAL_CONVERSION (basetype);
	    }
	}
      report_ambiguous_mi_virtuals (n_assocs+n_vbases*n_baseclasses, t);
#if 0
      /* Now that we know what the virtual functiond table looks like,
	 fix up offsets in the presence of virtual base classes.  */
      if (n_vbases)
	fixup_vbase_offsets (t);
#endif
    }

  /* Need to test METHOD_VEC here in case all methods
     (conversions and otherwise) are inherited.  */
  if (TYPE_HAS_CONVERSION (t) && method_vec != NULL_TREE)
    {
      tree first_conversions[last_conversion_type];
      tree last_conversions[last_conversion_type];
      enum conversion_type conv_index;
      tree *tmp;
      int i;

      bzero (first_conversions, sizeof (first_conversions));
      bzero (last_conversions, sizeof (last_conversions));
      for (tmp = &TREE_VEC_ELT (method_vec, 1);
	   tmp != TREE_VEC_END (method_vec); tmp += 1)
	{
	  if (OPERATOR_TYPENAME_P (DECL_ORIGINAL_NAME (*tmp)))
	    {
	      tree fntype = TREE_TYPE (*tmp);
	      tree return_type = TREE_TYPE (fntype);
	      assert (TREE_CODE (fntype) == METHOD_TYPE);

	      if (typecode_p (return_type, POINTER_TYPE))
		{
		  if (TYPE_READONLY (TREE_TYPE (return_type)))
		    conv_index = constptr_conv;
		  else
		    conv_index = ptr_conv;
		}
	      else if (typecode_p (return_type, INTEGER_TYPE))
		{
		  TYPE_HAS_INT_CONVERSION (t) = 1;
		  conv_index = int_conv;
		}
	      else if (typecode_p (return_type, REAL_TYPE))
		{
		  TYPE_HAS_REAL_CONVERSION (t) = 1;
		  conv_index = real_conv;
		}
	      else
		continue;

	      if (first_conversions[(int) conv_index] == NULL_TREE)
		first_conversions[(int) conv_index] = *tmp;
	      last_conversions[(int) conv_index] = *tmp;
	    }
	}

      for (i = 0; i < (int) last_conversion_type; i++)
	if (first_conversions[i] != last_conversions[i])
	  CLASSTYPE_CONVERSION (t, i) = error_mark_node;
	else
	  CLASSTYPE_CONVERSION (t, i) = first_conversions[i];
    }

  /* If this type has constructors, force its mode to be BLKmode,
     and force its TREE_ADDRESSABLE bit to be nonzero.  */
  if (TYPE_NEEDS_CONSTRUCTING (t) || TYPE_NEEDS_DESTRUCTOR (t))
    {
      tree variants = t;

      if (TREE_CODE (TYPE_NAME (t)) == TYPE_DECL)
	DECL_MODE (TYPE_NAME (t)) = BLKmode;
      while (variants)
	{
	  TYPE_MODE (variants) = BLKmode;
	  TREE_ADDRESSABLE (variants) = 1;
	  variants = TYPE_NEXT_VARIANT (variants);
	}
    }
}

/* Create a RECORD_TYPE or UNION_TYPE node for a C struct or union declaration
   (or C++ class declaration).

   For C++, we must handle the building of derived classes.
   Also, C++ allows static class members.  The way that this is
   handled is to keep the field name where it is (as the DECL_NAME
   of the field), and place the overloaded decl in the DECL_FIELD_BITPOS
   of the field.  layout_record and layout_union will know about this.

   More C++ hair: inline functions have text in their
   DECL_PENDING_INLINE_INFO nodes which must somehow be parsed into
   meaningful tree structure.  After the struct has been laid out, set
   things up so that this can happen.

   And still more: virtual functions.  In the case of single inheritance,
   when a new virtual function is seen which redefines a virtual function
   from the base class, the new virtual function is placed into
   the virtual function table at exactly the same address that
   it had in the base class.  When this is extended to multiple
   inheritance, the same thing happens, except that multiple virtual
   function tables must be maintained.  The first virtual function
   table is treated in exactly the same way as in the case of single
   inheritance.  Additional virtual function tables have different
   DELTAs, which tell how to adjust `this' to point to the right thing.

   LIST_OF_FIELDLISTS is just that.  The elements of the list are
   TREE_LIST elements, whose TREE_PURPOSE field tells what visibility
   the list has, and the TREE_VALUE slot gives the actual fields.

   EMPTY is non-zero if this structure has no declarations following it.

   If flag_all_virtual == 1, then we lay all functions into
   the virtual function table, as though they were declared
   virtual.  Constructors do not lay down in the virtual function table.

   If flag_all_virtual == 2, then we lay all functions into
   the virtual function table, such that virtual functions
   occupy a space by themselves, and then all functions
   of the class occupy a space by themselves.  This is illustrated
   in the following diagram:

   class A; class B : A;

	Class A's vtbl:			Class B's vtbl:
    --------------------------------------------------------------------
   | A's virtual functions|		| B's virtual funcitions	|
   |			  |		| (may inherit some from A).	|
    --------------------------------------------------------------------
   | All of A's functions |		| All of A's functions		|
   | (such as a->A::f).	  |		| (such as b->A::f)		|
    --------------------------------------------------------------------
					| B's new virtual functions	|
					| (not defined in A.)		|
					 -------------------------------
					| All of B's functions		|
					| (such as b->B::f)		|
					 -------------------------------

   this allows the program to make references to any function, virtual
   or otherwise in a type-consistant manner.  */

extern tree constructor_name ();
tree
finish_struct (t, list_of_fieldlists, empty, warn_anon)
     tree t;
     tree list_of_fieldlists;
     int empty;
     int warn_anon;
{
  extern int interface_only, interface_unknown;
  int old;
  int round_up_size = 1;
  /* Set non-zero to debug using default functions.
     Not set by program.  */
  static int debug_default_functions = 0;

  enum tree_code code = TREE_CODE (t);
  register tree x, y, method_vec;
  int needs_ctor = 0, needs_dtor = 0;
  int members_need_dtors = 0;
  tree name = TYPE_NAME (t), fields, fn_fields, tail;
  enum visibility_type visibility;
  int all_virtual;
  int has_virtual;
  int max_has_virtual;
  tree pending_virtuals = NULL_TREE;
  tree abstract_virtuals = NULL_TREE;
  tree vfield;
  tree vfields;
  int needs_default_ctor;
  int cant_have_default_ctor;
  int needs_const_ctor;
  int cant_have_const_ctor;

  /* The index of the first base class which has virtual
     functions.  Only applied to non-virtual baseclasses.  */
  int first_vfn_base_index;

  int i, n_baseclasses;
  int any_default_members = 0;
  char *err_name;
  int const_sans_init = 0;
  int ref_sans_init = 0;
  int nonprivate_method = 0;

  if (TREE_CODE (name) == TYPE_DECL)
    {
      extern int lineno;

      DECL_SOURCE_FILE (name) = input_filename;
      DECL_SOURCE_LINE (name) = lineno;
      name = DECL_NAME (name);
    }
  err_name = IDENTIFIER_POINTER (name);

  if (warn_anon && code != UNION_TYPE && ANON_AGGRNAME_P (name))
    {
      warning ("un-usable class ignored (anonymous classes and unions are useless)");
      err_name = "(anon)";
    }

  leftmost_baseclasses = NULL_TREE;
  if (TYPE_SIZE (t))
    {
      if (TREE_CODE (t) == UNION_TYPE)
	error ("redefinition of `union %s'", err_name);
      else if (TREE_CODE (t) == RECORD_TYPE)
	error ("redefinition of `struct %s'", err_name);
      else
	assert (0);
      popclass (0);
      return t;
    }

  GNU_xref_decl (current_function_decl, t);

  /* If this type was previously laid out as a forward reference,
     make sure we lay it out again.  */

  TYPE_SIZE (t) = 0;
  CLASSTYPE_GOT_SEMICOLON (t) = 0;
  CLASSTYPE_INTERFACE_ONLY (t) = interface_only;
  CLASSTYPE_INTERFACE_UNKNOWN (t) = interface_unknown;

  if (flag_dossier
      /* Unions cannot have dossiers.  */
      && TREE_CODE (t) == RECORD_TYPE
      && current_lang_name == lang_name_cplusplus)
    {
      CLASSTYPE_DOSSIER (t) = build_t_desc (t, 0);
    }
  /* Just build the dossier info, without sticking it into T.  */
  else if (flag_dossier)
    build_t_desc (t, 0);

  old = suspend_momentary ();

  /* Install struct as DECL_FIELD_CONTEXT of each field decl.
     Also process specified field sizes.
     Set DECL_FRAME_SIZE to the specified size, or 0 if none specified.
     The specified size is found in the DECL_INITIAL.
     Store 0 there, except for ": 0" fields (so we can find them
     and delete them, below).  */

  n_baseclasses = CLASSTYPE_N_BASECLASSES (t);
  if (n_baseclasses > 0)
    {
      struct base_info base_info;

      /* If using multiple inheritance, this may cause variants of our
	 basetypes to be used (instead of their canonical forms).  */
      fields = layout_basetypes (t);
      y = tree_last (fields);

      first_vfn_base_index = finish_base_struct (t, &base_info);
      has_virtual = base_info.has_virtual;
      max_has_virtual = base_info.max_has_virtual;
      CLASSTYPE_N_SUPERCLASSES (t) += base_info.n_ancestors;
      vfield = base_info.vfield;
      vfields = base_info.vfields;
      needs_default_ctor = base_info.needs_default_ctor;
      cant_have_default_ctor = base_info.cant_have_default_ctor;
      needs_const_ctor = base_info.needs_const_ctor;
      cant_have_const_ctor = base_info.cant_have_const_ctor;
      n_baseclasses = CLASSTYPE_N_BASECLASSES (t);
    }
  else
    {
      first_vfn_base_index = -1;
      has_virtual = 0;
      max_has_virtual = has_virtual;
      vfield = NULL_TREE;
      vfields = NULL_TREE;
      fields = NULL_TREE;
      y = NULL_TREE;
      needs_default_ctor = 0;
      cant_have_default_ctor = 0;
      needs_const_ctor = 0;
      cant_have_const_ctor = 0;
    }

  /* Every class gets its own new slot in the vtable.  */
  if (CLASSTYPE_DOSSIER (t))
    has_virtual += 1;

  if (write_virtuals == 3 && ! CLASSTYPE_INTERFACE_UNKNOWN (t)
      && current_lang_name == lang_name_cplusplus)
    {
      CLASSTYPE_INTERFACE_ONLY (t) = interface_only;
      CLASSTYPE_VTABLE_NEEDS_WRITING (t) = ! interface_only;
    }

  /* The three of these are approximations which may later be
     modified.  Needed at this point to make add_virtual_function
     and modify_vtable_entries work.  */
  CLASSTYPE_ASSOC (t) = make_assoc (integer_zero_node, t,
				    0, 0, CLASSTYPE_ASSOC (t));
  CLASSTYPE_VFIELDS (t) = vfields;
  CLASSTYPE_VFIELD (t) = vfield;

  fn_fields = NULL_TREE;
  tail = NULL_TREE;
  if (y && list_of_fieldlists)
    TREE_CHAIN (y) = TREE_VALUE (list_of_fieldlists);

#ifdef SOS
  if (flag_all_virtual == 2)
    all_virtual = 2;
  else
#endif
    {
      if (flag_all_virtual == 1 && TYPE_OVERLOADS_METHOD_CALL_EXPR (t))
	all_virtual = 1;
      else
	all_virtual = 0;
    }

  if (CLASSTYPE_DECLARED_CLASS (t) == 0)
    {
      nonprivate_method = 1;
      if (list_of_fieldlists
	  && TREE_PURPOSE (list_of_fieldlists) == (tree)visibility_default)
	TREE_PURPOSE (list_of_fieldlists) = (tree)visibility_public;
    }
  else if (list_of_fieldlists
	   && TREE_PURPOSE (list_of_fieldlists) == (tree)visibility_default)
    TREE_PURPOSE (list_of_fieldlists) = (tree)visibility_private;

  while (list_of_fieldlists)
    {
      visibility = (enum visibility_type)TREE_PURPOSE (list_of_fieldlists);

      for (x = TREE_VALUE (list_of_fieldlists); x; x = TREE_CHAIN (x))
	{
	  TREE_PRIVATE (x) = visibility == visibility_private;
	  TREE_PROTECTED (x) = visibility == visibility_protected;
	  GNU_xref_member (current_class_name, x);
	  if (TREE_CODE (x) == FUNCTION_DECL)
	    {
	      /* Clear out this flag.

	         @@ Doug may figure out how to break
		 @@ this with nested classes and friends.  */
	      DECL_IN_AGGR_P (x) = 0;

	      nonprivate_method |= ! TREE_PRIVATE (x);

	      /* If this was an evil function, don't keep it in class.  */
	      if (IDENTIFIER_ERROR_LOCUS (DECL_NAME (x)))
		continue;

	      if (y) TREE_CHAIN (y) = TREE_CHAIN (x);
	      if (! fn_fields) fn_fields = x;
	      else TREE_CHAIN (tail) = x;
	      tail = x;
	      if (DECL_CONTEXT (x))
		continue;

	      DECL_CONTEXT (x) = t;
	      DECL_VCONTEXT (x) = t;

	      DECL_FRAME_SIZE (x) = 0;

	      /* The name of the field is the original field name
		 Save this in auxiliary field for later overloading.  */
	      if (DECL_VIRTUAL_P (x)
		  || (all_virtual == 1 && ! DECL_CONSTRUCTOR_P (x)))
		{
		  pending_virtuals = add_virtual_function (pending_virtuals,
							   &has_virtual, x);
		  if (DECL_ABSTRACT_VIRTUAL_P (x))
		    abstract_virtuals = tree_cons (NULL_TREE, x, abstract_virtuals);
		}
	      continue;
	    }

	  /* Handle visibility declarations.  */
	  if (DECL_NAME (x) && TREE_CODE (DECL_NAME (x)) == SCOPE_REF)
	    {
	      tree fdecl = TREE_OPERAND (DECL_NAME (x), 1);

	      if (y) TREE_CHAIN (y) = TREE_CHAIN (x);
	      /* Make type T see field decl FDECL with
		 the visibility VISIBILITY.  */
	      if (TREE_CODE (fdecl) == TREE_LIST)
		{
		  fdecl = TREE_VALUE (fdecl);
		  while (fdecl)
		    {
		      if (alter_visibility (t, fdecl, visibility) == 0)
			break;
		      fdecl = TREE_CHAIN (fdecl);
		    }
		}
	      else alter_visibility (t, fdecl, visibility);
	      CLASSTYPE_ALTERS_VISIBILITIES_P (t) = 1;
	      continue;
	    }

	  /* Perform error checking that did not get done in grokdeclarator.  */
	  if (TREE_CODE (TREE_TYPE (x)) == FUNCTION_TYPE)
	    {
	      error_with_decl (x, "field `%s' invalidly declared function type");
	      TREE_TYPE (x) = build_pointer_type (TREE_TYPE (x));
	    }
	  else if (TREE_CODE (TREE_TYPE (x)) == METHOD_TYPE)
	    {
	      error_with_decl (x, "field `%s' invalidly declared method type");
	      TREE_TYPE (x) = build_pointer_type (TREE_TYPE (x));
	    }
	  else if (TREE_CODE (TREE_TYPE (x)) == OFFSET_TYPE)
	    {
	      error_with_decl (x, "field `%s' invalidly declared offset type");
	      TREE_TYPE (x) = build_pointer_type (TREE_TYPE (x));
	    }
	  /* If this is of reference type, check if it needs an init.  */
	  if (TREE_CODE (TREE_TYPE (x)) == REFERENCE_TYPE
	      && DECL_INITIAL (x) == 0)
	    ref_sans_init = 1;

	  /* When this goes into scope, it will be a non-local reference.  */
	  TREE_NONLOCAL (x) = 1;

	  if (TREE_CODE (x) == FIELD_DECL)
	    {
	      /* Never let anything with uninheritable virtuals
		 make it through without complaint.  */
	      if (TYPE_LANG_SPECIFIC (TREE_TYPE (x))
		  && CLASSTYPE_ABSTRACT_VIRTUALS (TREE_TYPE (x)))
		abstract_virtuals_error (x, TREE_TYPE (x));

	      if (TYPE_LANG_SPECIFIC (TREE_TYPE (x)))
		{
		  if (TYPE_HAS_DEFAULT_CONSTRUCTOR (TREE_TYPE (x)))
		    needs_default_ctor = 1;
		  if (TYPE_GETS_CONST_INIT_REF (TREE_TYPE (x)))
		    needs_const_ctor = 1;
		  else if (TYPE_GETS_INIT_REF (TREE_TYPE (x)))
		    cant_have_const_ctor = 1;
		}
	      else if (DECL_INITIAL (x) == NULL_TREE
		       && (TYPE_HAS_CONSTRUCTOR (TREE_TYPE (x))
			   || TREE_CODE (TREE_TYPE (x)) == REFERENCE_TYPE))
		cant_have_default_ctor = 1;

	      /* If any field is const, the structure type is pseudo-const.  */
	      if (TREE_READONLY (x))
		{
		  C_TYPE_FIELDS_READONLY (t) = 1;
		  if (DECL_INITIAL (x) == 0)
		    const_sans_init = 1;
		}
	      else 
		{
		  /* A field that is pseudo-const makes the structure likewise.  */
		  tree t1 = TREE_TYPE (x);
		  while (TREE_CODE (t1) == ARRAY_TYPE)
		    t1 = TREE_TYPE (t1);
		  if (IS_AGGR_TYPE (t1))
		    {
		      if (C_TYPE_FIELDS_READONLY (t1))
			C_TYPE_FIELDS_READONLY (t) = 1;
		      if (CLASSTYPE_READONLY_FIELDS_NEED_INIT (t1))
			const_sans_init = 1;
		    }
		}
	    }
	  else if (TREE_STATIC (x) && TREE_CODE (t) == UNION_TYPE)
	    /* Unions cannot have static members.  */
	    error_with_decl (x, "field `%s' declared static in union");

	  if (! fields) fields = x;
	  DECL_FIELD_CONTEXT (x) = t;
	  DECL_FRAME_SIZE (x) = 0;

	  if (DECL_BIT_FIELD (x))
	    {
	      /* Invalid bit-field size done by grokfield.  */
	      /* Detect invalid bit-field type.  */
	      if (DECL_INITIAL (x)
		  && TREE_CODE (TREE_TYPE (x)) != INTEGER_TYPE
		  && TREE_CODE (TREE_TYPE (x)) != ENUMERAL_TYPE)
		{
		  error_with_decl (x, "bit-field `%s' has invalid type");
		  DECL_INITIAL (x) = NULL;
		}
	      if (DECL_INITIAL (x) && pedantic
		  && TREE_TYPE (x) != integer_type_node
		  && TREE_TYPE (x) != unsigned_type_node)
		warning_with_decl (x, "bit-field `%s' type invalid in ANSI C");

	      /* Detect and ignore out of range field width.  */
	      if (DECL_INITIAL (x))
		{
		  register int width = TREE_INT_CST_LOW (DECL_INITIAL (x));

		  if (width < 0)
		    {
		      DECL_INITIAL (x) = NULL;
		      warning_with_decl (x, "negative width in bit-field `%s'");
		    }
		  else if (width == 0 && DECL_NAME (x) != 0)
		    {
		      error_with_decl (x, "zero width for bit-field `%s'");
		      DECL_INITIAL (x) = NULL;
		    }
		  else if (width > TYPE_PRECISION (TREE_TYPE (x)))
		    {
		      DECL_INITIAL (x) = NULL;
		      warning_with_decl (x, "width of `%s' exceeds its type");
		    }
		}

	      /* Process valid field width.  */
	      if (DECL_INITIAL (x))
		{
		  register int width = TREE_INT_CST_LOW (DECL_INITIAL (x));

		  if (width == 0)
		    {
#ifdef EMPTY_FIELD_BOUNDARY
		      /* field size 0 => mark following field as "aligned" */
		      if (TREE_CHAIN (x))
			DECL_ALIGN (TREE_CHAIN (x))
			  = MAX (DECL_ALIGN (TREE_CHAIN (x)), EMPTY_FIELD_BOUNDARY);
		      /* field of size 0 at the end => round up the size.  */
		      else
			round_up_size = EMPTY_FIELD_BOUNDARY;
#endif
#ifdef PCC_BITFIELD_TYPE_MATTERS
		      DECL_ALIGN (x) = MAX (DECL_ALIGN (x),
					    TYPE_ALIGN (TREE_TYPE (x)));
#endif
		    }
		  else
		    {
		      DECL_INITIAL (x) = NULL_TREE;
		      DECL_FRAME_SIZE (x) = width;
		      DECL_BIT_FIELD (x) = 1;
		      /* Traditionally a bit field is unsigned
			 even if declared signed.  */
		      if (flag_traditional
			  && TREE_CODE (TREE_TYPE (x)) == INTEGER_TYPE)
			TREE_TYPE (x) = unsigned_type_node;
		    }
		}
	      else
		/* Non-bit-fields are aligned for their type.  */
		DECL_ALIGN (x) = MAX (DECL_ALIGN (x), TYPE_ALIGN (TREE_TYPE (x)));
	    }
	  else if (TREE_CODE (x) == FIELD_DECL)
	    {
	      tree type = TREE_TYPE (x);
	      if (TREE_CODE (type) == ARRAY_TYPE)
		type = TREE_TYPE (type);
	      if (code == UNION_TYPE)
		{
		  if (TYPE_NEEDS_CONSTRUCTING (type)
		      || TYPE_NEEDS_DESTRUCTOR (type))
		    error_with_decl (x, "member `%s' with constructor or destructor not allowed in union");
		}
	      else if (code == RECORD_TYPE)
		{
		  /* Array of record type doesn't matter for this bit.  */
		  TYPE_NEEDS_CONSTRUCTING (t) |= TYPE_NEEDS_CONSTRUCTING (type);
		  if (IS_AGGR_TYPE (type))
		    {
		      needs_ctor |= TYPE_NEEDS_CONSTRUCTOR (type);
		      needs_dtor |= TYPE_NEEDS_DESTRUCTOR (type);
		      members_need_dtors |= TYPE_NEEDS_DESTRUCTOR (type);
		      TYPE_GETS_ASSIGNMENT (t) |= TYPE_GETS_ASSIGNMENT (type);
		      TYPE_GETS_INIT_REF (t) |= TYPE_GETS_INIT_REF (type);
		      TYPE_GETS_CONST_INIT_REF (t) |= TYPE_GETS_CONST_INIT_REF (type);
		    }
		}
	      if (DECL_INITIAL (x) != NULL_TREE)
		{
		  /* `build_class_init_list' does not recognize non-FIELD_DECLs.  */
		  if (code == UNION_TYPE && any_default_members != 0)
		    error ("multiple fields in union initialized");
		  any_default_members = 1;
		}
	    }
	  y = x;
	}
      list_of_fieldlists = TREE_CHAIN (list_of_fieldlists);
      /* link the tail while we have it! */
      if (y)
	{
	  TREE_CHAIN (y) = NULL_TREE;

	  if (list_of_fieldlists
	      && TREE_VALUE (list_of_fieldlists)
	      && TREE_CODE (TREE_VALUE (list_of_fieldlists)) != FUNCTION_DECL)
	    TREE_CHAIN (y) = TREE_VALUE (list_of_fieldlists);
	}
    }

  if (tail) TREE_CHAIN (tail) = NULL_TREE;

  /* If this type has any constant members which did not come
     with their own initialization, mark that fact here.  It is
     not an error here, since such types can be saved either by their
     constructors, or by fortuitous initialization.  */
  CLASSTYPE_READONLY_FIELDS_NEED_INIT (t) = const_sans_init;
  CLASSTYPE_REF_FIELDS_NEED_INIT (t) = ref_sans_init;
  CLASSTYPE_ABSTRACT_VIRTUALS (t) = abstract_virtuals;

  if (vfield == 0
      && (has_virtual
#ifdef SOS
	  || TYPE_DYNAMIC (t)
#endif
	  ))
    {
      /* We build this decl with ptr_type_node, and
	 change the type when we know what it should be.  */
      vfield = build_decl (FIELD_DECL, get_vfield_name (t), ptr_type_node);
      CLASSTYPE_VFIELD (t) = vfield;
      DECL_VIRTUAL_P (vfield) = 1;
      DECL_FIELD_CONTEXT (vfield) = t;
      SET_DECL_FCONTEXT (vfield, t);
      DECL_FRAME_SIZE (vfield) = 0;
      if (CLASSTYPE_DOSSIER (t))
	{
	  /* vfield is always first entry in structure.  */
	  TREE_CHAIN (vfield) = fields;
	  fields = vfield;
	}
      else if (y)
	{
	  assert (TREE_CHAIN (y) == 0);
	  TREE_CHAIN (y) = vfield;
	  y = vfield;
	}
      else fields = vfield;
      vfields = chainon (vfields, CLASSTYPE_AS_LIST (t));
    }

  /* Now DECL_INITIAL is null on all members except for zero-width bit-fields.
     And they have already done their work.

     C++: maybe we will support default field initialization some day...  */

  /* Delete all zero-width bit-fields from the front of the fieldlist */
  while (fields && DECL_BIT_FIELD (fields)
	 && DECL_INITIAL (fields))
    fields = TREE_CHAIN (fields);
  /* Delete all such fields from the rest of the fields.  */
  for (x = fields; x;)
    {
      if (TREE_CHAIN (x) && DECL_BIT_FIELD (TREE_CHAIN (x))
	  && DECL_INITIAL (TREE_CHAIN (x)))
	TREE_CHAIN (x) = TREE_CHAIN (TREE_CHAIN (x));
      else x = TREE_CHAIN (x);
    }
  /* Delete all duplicate fields from the fields */
  delete_duplicate_fields (fields);

  /* Now we have the final fieldlist for the data fields.  Record it,
     then lay out the structure or union (including the fields).  */

  TYPE_FIELDS (t) = fields;

  /* If there's a :0 field at the end, round the size to the
     EMPTY_FIELD_BOUNDARY.  */
  TYPE_ALIGN (t) = round_up_size;

  if (debug_default_functions)
    {
      if ((TYPE_NEEDS_CONSTRUCTOR (t) || TYPE_HAS_CONSTRUCTOR (t) || needs_ctor)
	  && ! TYPE_HAS_INIT_REF (t))
	{
	  tree default_fn = cons_up_default_function (t, name, 1);
	  TREE_CHAIN (default_fn) = fn_fields;
	  DECL_CONTEXT (default_fn) = t;
	  DECL_VCONTEXT (default_fn) = t;
	  fn_fields = default_fn;
	  TYPE_HAS_INIT_REF (t) = 1;
	  default_fn = cons_up_default_function (t, name, 3);
	  TREE_CHAIN (default_fn) = fn_fields;
	  DECL_CONTEXT (default_fn) = t;
	  DECL_VCONTEXT (default_fn) = t;
	  fn_fields = default_fn;
	  nonprivate_method = 1;
	}

      if (! TYPE_HAS_DEFAULT_CONSTRUCTOR (t)
	  && needs_default_ctor && ! cant_have_default_ctor)
	{
	  tree default_fn = cons_up_default_function (t, name, 2);
	  TREE_CHAIN (default_fn) = fn_fields;
	  DECL_CONTEXT (default_fn) = t;
	  DECL_VCONTEXT (default_fn) = t;
	  fn_fields = default_fn;
	  TYPE_HAS_DEFAULT_CONSTRUCTOR (t) = 1;
	  nonprivate_method = 1;
	}
    }
  /* Warn about duplicate methods in fn_fields.  Also compact
     method lists so that lookup can be made faster.

     Algorithm:  Outer loop builds lists by method name.
     Inner loop checks for redundant method names within a list.

     Data Structure:  List of method lists.  The outer list
     is a TREE_LIST, whose TREE_PURPOSE field is the field name
     and the TREE_VALUE is the TREE_CHAIN of the FUNCTION_DECLs.
     Friends are chained in the same way as member functions, but
     they live in the TREE_TYPE field of the outer list.
     That allows them to be quicky deleted, and requires
     no extra storage.

     If there are any constructors/destructors, they are moved to
     the front of the list.  This makes pushclass more efficient.

     We also link each field which has shares a name with its
     baseclass to the head of the list of fields for that base class.
     This allows us to reduce search time in places like `build_method_call'
     to consider only reasonably likely functions.  */

  if (fn_fields)
    {
      /* Now prepare to gather fn_fields into vector.  */
      struct obstack *ambient_obstack = current_obstack;
      current_obstack = &class_obstack;
      method_vec = make_node (TREE_VEC);
      /* Room has been saved for constructors and destructors.  */
      current_obstack = ambient_obstack;
      /* Now make this a live vector.  */
      obstack_free (&class_obstack, method_vec);
      obstack_blank (&class_obstack, sizeof (struct tree_vec));

      while (fn_fields)
	{
	  /* NEXT Pointer, TEST Pointer, and BASE Pointer.  */
	  tree nextp, *testp;

	  nextp = TREE_CHAIN (fn_fields);
	  TREE_CHAIN (fn_fields) = NULL_TREE;
	  /* Constrcutors are handled easily in search routines.
	     Besides, we know we wont find any, so do not bother looking.  */
	  if (DECL_ORIGINAL_NAME (fn_fields) == constructor_name (name)
	      && TREE_VEC_ELT (method_vec, 0) == 0)
	    TREE_VEC_ELT (method_vec, 0) = fn_fields;
	  else
	    {
	      testp = &TREE_VEC_ELT (method_vec, 0);
	      if (*testp == NULL_TREE)
		testp++;
	      while ((int)testp < (int)obstack_next_free (&class_obstack)
		     && DECL_ORIGINAL_NAME (*testp) != DECL_ORIGINAL_NAME (fn_fields))
		testp++;
	      if ((int)testp < (int)obstack_next_free (&class_obstack))
		{
		  for (x = *testp; x; x = DECL_CHAIN (x))
		    {
		      if (DECL_NAME (fn_fields) == DECL_NAME (x))
			{
			  /* We complain about multiple destructors on sight,
			     so we do not repeat the warning here.  Friend-friend
			     ambiguities are warned about outside this loop.  */
			  if (! DESTRUCTOR_NAME_P (DECL_NAME (fn_fields)))
			    error_with_file_and_line (DECL_SOURCE_FILE (fn_fields),
						      DECL_SOURCE_LINE (fn_fields),
						      "ambiguous method `%s' in structure",
						      lang_printable_name (fn_fields));
			  break;
			}
		      y = x;
		    }
		  if (x == 0)
		    if (*testp)
		      DECL_CHAIN (y) = fn_fields;
		    else
		      *testp = fn_fields;
		}
	      else
		{
		  obstack_ptr_grow (&class_obstack, fn_fields);
		  method_vec = (tree)obstack_base (&class_obstack);
		}
	    }
	  fn_fields = nextp;
	}

      TREE_VEC_LENGTH (method_vec)
	= (tree *)obstack_next_free (&class_obstack) - (&TREE_VEC_ELT (method_vec, 0));
      obstack_finish (&class_obstack);
      CLASSTYPE_METHOD_VEC (t) = method_vec;

      if (nonprivate_method == 0
	  && CLASSTYPE_FRIEND_CLASSES (t) == NULL_TREE
	  && DECL_FRIENDLIST (TYPE_NAME (t)) == NULL_TREE)
	{
	  for (i = 0; i < n_baseclasses; i++)
	    if (CLASSTYPE_VIA_PUBLIC (t, i))
	      {
		nonprivate_method = 1;
		break;
	      }
	  if (nonprivate_method == 0)
	    warning ("all class member functions are private");
	}
    }
  else
    {
      method_vec = 0;

      /* Just in case these got accidently
	 filled in by syntax errors.  */
      TYPE_HAS_CONSTRUCTOR (t) = 0;
      TYPE_HAS_DESTRUCTOR (t) = 0;
    }

  /* If there are constructors (and destructors), they are at the
     front.  Place destructors at very front.  Also warn if all
     constructors and/or destructors are private (in which case this
     class is effectively unusable.  */
  if (TYPE_HAS_DESTRUCTOR (t))
    {
      tree dtor, prev;

      for (dtor = TREE_VEC_ELT (method_vec, 0); dtor; prev = dtor, dtor = DECL_CHAIN (dtor))
	{
	  if (DESTRUCTOR_NAME_P (DECL_NAME (dtor)))
	    {
	      if (TREE_PRIVATE (dtor)
		  && CLASSTYPE_FRIEND_CLASSES (t) == NULL_TREE
		  && DECL_FRIENDLIST (TYPE_NAME (t)) == NULL_TREE)
		warning_with_decl (TYPE_NAME (t), "class `%s' only defines a private destructor and has no friends");
	      break;
	    }
	}
      /* Wild parse errors can cause this to happen.  */
      if (dtor == NULL_TREE)
	TYPE_HAS_DESTRUCTOR (t) = 0;
      else if (dtor != TREE_VEC_ELT (method_vec, 0))
	{
	  DECL_CHAIN (prev) = DECL_CHAIN (dtor);
	  DECL_CHAIN (dtor) = TREE_VEC_ELT (method_vec, 0);
	  TREE_VEC_ELT (method_vec, 0) = dtor;
	}
    }
  else if (members_need_dtors || TYPE_USES_COMPLEX_INHERITANCE (t))
    {
      /* Here we must cons up a destructor on the fly.  */
      tree dtor = cons_up_default_function (t, name, 0);

      /* If we couldn't make it work, then pretend we didn't need it.  */
      if (dtor == void_type_node)
	TYPE_NEEDS_DESTRUCTOR (t) = 0;
      else
	{
	  DECL_CONTEXT (dtor) = t;
	  DECL_VCONTEXT (dtor) = t;
	  if (DECL_VIRTUAL_P (dtor))
	    pending_virtuals = add_virtual_function (pending_virtuals,
						     &has_virtual, dtor);
	  if (TYPE_HAS_CONSTRUCTOR (t))
	    DECL_CHAIN (dtor) = TREE_VEC_ELT (method_vec, 0);
	  else if (method_vec == 0)
	    {
	      /* Now prepare to gather fn_fields into vector.  */
	      struct obstack *ambient_obstack = current_obstack;
	      current_obstack = &class_obstack;
	      method_vec = make_node (TREE_VEC);
	      /* Room has been saved for constructors and destructors.  */
	      current_obstack = ambient_obstack;
	      TREE_VEC_LENGTH (method_vec) = 1;
	      CLASSTYPE_METHOD_VEC (t) = method_vec;
	    }
	  TREE_VEC_ELT (method_vec, 0) = dtor;
	  TYPE_HAS_DESTRUCTOR (t) = 1;
	}
    }
  if (TYPE_HAS_CONSTRUCTOR (t)
      && ! CLASSTYPE_DECLARED_EXCEPTION (t)
      && CLASSTYPE_FRIEND_CLASSES (t) == NULL_TREE
      && DECL_FRIENDLIST (TYPE_NAME (t)) == NULL_TREE)
    {
      int nonprivate_ctor = 0;
      tree ctor;

      for (ctor = TREE_VEC_ELT (method_vec, 0); ctor; ctor = DECL_CHAIN (ctor))
	if (! TREE_PRIVATE (ctor))
	  {
	    nonprivate_ctor = 1;
	    break;
	  }
      if (nonprivate_ctor == 0)
	warning ("class %s only defines private constructors and has no friends",
		 TYPE_NAME_STRING (t));
    }

  /* Now for each member function (except for constructors and
     destructors), compute where member functions of the same
     name reside in base classes.  */
  if (n_baseclasses != 0
      && method_vec != NULL_TREE
      && TREE_VEC_LENGTH (method_vec) > 1)
    {
      int len = TREE_VEC_LENGTH (method_vec);
      tree baselink_vec = make_tree_vec (len);
      int any_links = 0;

      for (i = 1; i < len; i++)
	{
	  TREE_VEC_ELT (baselink_vec, i)
	    = get_baselinks (t, DECL_ORIGINAL_NAME (TREE_VEC_ELT (method_vec, i)));
	  if (TREE_VEC_ELT (baselink_vec, i) != 0)
	    any_links = 1;
	}
      if (any_links != 0)
	CLASSTYPE_BASELINK_VEC (t) = baselink_vec;
      else
	obstack_free (current_obstack, baselink_vec);
    }

  /* Pass layout information about base classes to layout_type, if any.  */

  if (n_baseclasses)
    {
      tree pseudo_basetype = TREE_TYPE (base_layout_decl);

      TREE_CHAIN (base_layout_decl) = TYPE_FIELDS (t);
      TYPE_FIELDS (t) = base_layout_decl;

      TYPE_SIZE (pseudo_basetype) = CLASSTYPE_SIZE (t);
      TYPE_MODE (pseudo_basetype) = TYPE_MODE (t);
      TYPE_ALIGN (pseudo_basetype) = CLASSTYPE_ALIGN (t);
      DECL_ALIGN (base_layout_decl) = TYPE_ALIGN (pseudo_basetype);
    }

  layout_type (t);

  if (n_baseclasses)
    TYPE_FIELDS (t) = TREE_CHAIN (TYPE_FIELDS (t));

  /* C++: do not let empty structures exist.  */
  if (integer_zerop (TYPE_SIZE (t)))
    TYPE_SIZE (t) = TYPE_SIZE (char_type_node);

  /* Set the TYPE_DECL for this type to contain the right
     value for DECL_OFFSET, so that we can use it as part
     of a COMPONENT_REF for multiple inheritance.  */

  if (TREE_CODE (TYPE_NAME (t)) == TYPE_DECL)
    layout_decl (TYPE_NAME (t), 0);

  /* Now fix up any virtual base class types that we
     left lying around.  We must get these done
     before we try to lay out the virtual function table.  */
  doing_hard_virtuals = 1;
  pending_hard_virtuals = nreverse (pending_hard_virtuals);

  if (TYPE_USES_VIRTUAL_BASECLASSES (t))
    {
      tree vbases;

      max_has_virtual = layout_vbasetypes (t, max_has_virtual);
      vbases = CLASSTYPE_VBASECLASSES (t);
      CLASSTYPE_N_VBASECLASSES (t) = list_length (vbases);

      /* This loop makes all the entries in the virtual function tables
	 of interest contain the "latest" version of the functions
	 we have defined.  */

      while (vbases)
	{
	  tree virtuals = ASSOC_VIRTUALS (vbases);

	  if (virtuals)
	    {
	      /* Get past the `null' vtable entry...  */
	      virtuals = TREE_CHAIN (virtuals);
	      /* and the `dossier' vtable entry if we're doing dossiers.  */
	      if (flag_dossier)
		virtuals = TREE_CHAIN (virtuals);
	    }

	  while (virtuals != NULL_TREE)
	    {
	      tree pfn = FNADDR_FROM_VTABLE_ENTRY (TREE_VALUE (virtuals));
	      tree base_fndecl = TREE_OPERAND (pfn, 0);
	      tree decl = get_first_matching_virtual (t, base_fndecl, 0);
	      tree context = DECL_CONTEXT (decl);
	      if (decl != base_fndecl && context != t)
		{
		  tree assoc = NULL_TREE, these_virtuals;
		  int i = TREE_INT_CST_LOW (DECL_VINDEX (base_fndecl)) & ((1<<(BITS_PER_WORD-1))-1);

		  if (TYPE_USES_VIRTUAL_BASECLASSES (context))
		    assoc = virtual_member (DECL_CONTEXT (base_fndecl),
					    CLASSTYPE_VBASECLASSES (context));
		  if (assoc == NULL_TREE)
		    assoc = assoc_value (DECL_CONTEXT (base_fndecl), context, 0);
		  if (assoc != NULL_TREE)
		    {
		      these_virtuals = ASSOC_VIRTUALS (assoc);

		      while (i-- > 0)
			these_virtuals = TREE_CHAIN (these_virtuals);
		      pfn = FNADDR_FROM_VTABLE_ENTRY (TREE_VALUE (these_virtuals));
		      modify_vtable_entries (t, decl, base_fndecl, pfn);
		    }
		}
	      virtuals = TREE_CHAIN (virtuals);
	    }
	  /* Update dossier info with offsets for virtual baseclasses.  */
	  if (flag_dossier && ! CLASSTYPE_MARKED4 (ASSOC_TYPE (vbases)))
	    prepare_fresh_vtable (vbases, vbases, t);

	  vbases = TREE_CHAIN (vbases);
	}
    }

  while (pending_hard_virtuals)
    {
      /* Need an entry in some other virtual function table.  */
      if (TREE_TYPE (pending_hard_virtuals))
	{
	  /* This is how we modify entries when a vfn's index changes
	     between derived and base type.  */
	  modify_vtable_entries (t, TREE_PURPOSE (pending_hard_virtuals),
				 TREE_TYPE (pending_hard_virtuals),
				 TREE_VALUE (pending_hard_virtuals));
	}
      else
	{
	  /* This is how we modify entries when a vfn comes from
	     a virtual baseclass.  */
	  tree base_fndecls = DECL_VINDEX (TREE_PURPOSE (pending_hard_virtuals));
	  while (base_fndecls)
	    {
	      modify_vtable_entries (t, TREE_PURPOSE (pending_hard_virtuals),
				     TREE_VALUE (base_fndecls),
				     TREE_VALUE (pending_hard_virtuals));
	      base_fndecls = TREE_CHAIN (base_fndecls);
	    }
	}
      pending_hard_virtuals = TREE_CHAIN (pending_hard_virtuals);
    }
  doing_hard_virtuals = 0;

  /* Under our model of GC, every C++ class gets its own virtual
     function table, at least virtually.  */
  if (pending_virtuals || CLASSTYPE_DOSSIER (t))
    {
      pending_virtuals = nreverse (pending_virtuals);
      /* We must enter these virtuals into the table.  */
      if (first_vfn_base_index < 0)
	{
	  if (flag_dossier)
	    pending_virtuals = tree_cons (NULL_TREE,
					  build_vtable_entry (integer_zero_node,
							      build_t_desc (t, 0)),
					  pending_virtuals);
	  pending_virtuals = tree_cons (NULL_TREE, the_null_vtable_entry,
					pending_virtuals);
	  build_vtable (0, t);
	}
      else
	{
	  /* Here we know enough to change the type of our virtual
	     function table, but we will wait until later this function.  */
	  assert (first_vfn_base_index >= 0);

	  if (! CLASSTYPE_MARKED4 (t))
	    build_vtable (assoc_value (TYPE_MAIN_VARIANT (CLASSTYPE_BASECLASS (t, first_vfn_base_index)), t, 0), t);

	  /* Update the dossier pointer for this class.  */
	  if (flag_dossier)
	    TREE_VALUE (TREE_CHAIN (CLASS_ASSOC_VIRTUALS (t)))
	      = build_vtable_entry (integer_zero_node, build_t_desc (t, 0));
	}

      /* If this type has basetypes with constructors, then those
	 constructors might clobber the virtual function table.  But
	 they don't if the derived class shares the exact vtable of the base
	 class.  */

      CLASSTYPE_NEEDS_VIRTUAL_REINIT (t) = 1;
    }
  else if (first_vfn_base_index >= 0)
    {
      tree basetype = get_base_type (DECL_FIELD_CONTEXT (vfield), t, 0);
      tree assoc;
      
      if (TREE_VIA_VIRTUAL (basetype))
	assoc = virtual_member (DECL_FIELD_CONTEXT (vfield), CLASSTYPE_VBASECLASSES (t));
      else
	assoc = assoc_value (TYPE_MAIN_VARIANT (basetype), t, 0);

      /* This class contributes nothing new to the virtual function
	 table.  However, it may have declared functions which
	 went into the virtual function table "inherited" from the
	 base class.  If so, we grab a copy of those updated functions,
	 and pretend they are ours.  */

#ifdef SOS
      /* Don't define this ahead of time if we have more
	 fields to add later.  */
      if (all_virtual == 2 && fn_fields != NULL_TREE)
	;
      else
#endif
	{
	  /* See if we should steal the virtual info from base class.  */
	  if (CLASS_ASSOC_VTABLE (t) == NULL_TREE)
	    CLASS_ASSOC_VTABLE (t) = ASSOC_VTABLE (assoc);
	  if (CLASS_ASSOC_VIRTUALS (t) == NULL_TREE)
	    CLASS_ASSOC_VIRTUALS (t) = ASSOC_VIRTUALS (assoc);
	}
      if (CLASS_ASSOC_VTABLE (t) != ASSOC_VTABLE (assoc))
	CLASSTYPE_NEEDS_VIRTUAL_REINIT (t) = 1;
    }

  if (has_virtual > max_has_virtual)
    max_has_virtual = has_virtual;
  if (max_has_virtual || first_vfn_base_index >= 0)
    {
#ifdef VTABLE_USES_MASK
      if (max_has_virtual >= VINDEX_MAX)
	{
	  error ("too many virtual functions for class `%s' (VINDEX_MAX < %d)", TYPE_NAME_STRING (t), has_virtual);
	}
#endif
      TYPE_VIRTUAL_P (t) = 1;
      CLASSTYPE_VSIZE (t) = has_virtual;
      if (first_vfn_base_index >= 0)
	{
	  if (pending_virtuals)
	    CLASS_ASSOC_VIRTUALS (t) = chainon (CLASS_ASSOC_VIRTUALS (t),
						pending_virtuals);
	}
      else if (has_virtual)
	{
	  CLASS_ASSOC_VIRTUALS (t) = pending_virtuals;
	  if (write_virtuals >= 0)
	    DECL_VIRTUAL_P (CLASS_ASSOC_VTABLE (t)) = 1;
	}
    }

#ifdef SOS
  if (all_virtual == 2 && (max_has_virtual || method_vec))
    {
      /* Now that we know the size of the virtual table, lay out
	 the absolute table following it.  */
      int i;
      tree tmp;
      tree pending_absolutes = NULL_TREE;
      int has_absolute = has_virtual;

      /* Local variables for building a table filled with strings
	 containing the names of interesting things.  */
      tree decl, init;
      tree start = NULL_TREE, next = NULL_TREE;
      tree *outer = &TREE_VEC_ELT (method_vec, 0);

      while (outer != TREE_VEC_END (method_vec))
	{
	  tree inner;
	  for (inner = *outer; inner; inner = DECL_CHAIN (inner))
	    {
	      tree entry;
	      tree fn;

	      /* Don't bother with functions which appear
		 for visibility reasons.  */
	      if (DECL_FIELD_CONTEXT (inner) != t)
		continue;

	      /* Must lay this function into its absolute table as well.
		 This forces an inline function to be written out.  */
	      fn = build1 (ADDR_EXPR, ptr_type_node, inner);
	      TREE_CONSTANT (fn) = 1;
	      DECL_DINDEX (inner) = build_int_2 (++has_absolute, 0);
	      entry = build_vtable_entry (integer_zero_node, fn);
	      pending_absolutes = tree_cons (DECL_DINDEX (inner), entry,
					     pending_absolutes);
	    }
	  outer++;
	}

      CLASS_ASSOC_VIRTUALS (t) = chainon (CLASS_ASSOC_VIRTUALS (t),
					  nreverse (pending_absolutes));
      if (TYPE_DYNAMIC (t))
	{
	  for (outer = &TREE_VEC_ELT (method_vec, 0);
	       outer != TREE_VEC_END (method_vec);
	       outer++)
	    {
	      tree inner;
	      for (inner = *outer; inner; inner = DECL_CHAIN (inner))
		{
		  tree str = make_node (STRING_CST);
		  TREE_STRING_LENGTH (str) = IDENTIFIER_LENGTH (DECL_NAME (inner));
		  TREE_STRING_POINTER (str) = IDENTIFIER_POINTER (DECL_NAME (inner));
		  TREE_CONSTANT (str) = 1;
		  TREE_STATIC (str) = 1;
		  TREE_TYPE (str)
		    = build_cplus_array_type (char_type_node,
					      build_index_type (build_int_2 (TREE_STRING_LENGTH (str) - 1, 0)));
	      
		  if (start)
		    {
		      TREE_CHAIN (next) = build_tree_list (NULL_TREE, str);
		      next = TREE_CHAIN (next);
		    }
		  else
		    {
		      start = build_tree_list (NULL_TREE, str);
		      next = start;
		    }
		}
	    }

	  /* Lay out dynamic link table for SOS.  */

	  decl = finish_table (get_linktable_name (t),
			       string_type_node, start, 0);
	}
      has_virtual = has_absolute;
      CLASSTYPE_VSIZE (t) = has_virtual;
      if (has_virtual > max_has_virtual)
	max_has_virtual = has_virtual;
      if (vfield == 0)
	{
	  /* We build this decl with ptr_type_node, and
	     change the type when we know what it should be.  */
	  vfield = build_decl (FIELD_DECL, get_vfield_name (t), ptr_type_node);
	  CLASSTYPE_VFIELD (t) = vfield;
	  DECL_VIRTUAL_P (vfield) = 1;
	  DECL_FIELD_CONTEXT (vfield) = t;
	  SET_DECL_FCONTEXT (vfield, t);
	  DECL_FRAME_SIZE (vfield) = 0;
	  y = tree_last (fields);
	  if (y)
	    TREE_CHAIN (y) = vfield;
	  else
	    fields = vfield;
	  vfields = chainon (vfields, CLASSTYPE_AS_LIST (t));
	}
    }
#endif

  /* Now lay out the virtual function table.  */
  if (has_virtual)
    {
      tree atype, itype;

      if (TREE_TYPE (vfield) == ptr_type_node)
	{
	  /* We must create a pointer to this table because
	     the one inherited from base class does not exist.
	     We will fill in the type when we know what it
	     should really be.  */
	  itype = build_index_type (build_int_2 (has_virtual, 0));
	  atype = build_array_type (vtable_entry_type, itype);
	  layout_type (atype);
	  TREE_TYPE (vfield) = build_pointer_type (atype);
	}
      else
	{
	  atype = TREE_TYPE (TREE_TYPE (vfield));

	  if (has_virtual != TREE_INT_CST_LOW (TYPE_MAX_VALUE (TYPE_DOMAIN (atype))))
	    {
	      /* We must extend (or create) the boundaries on this array,
		 because we picked up virtual functions from multiple
		 base classes.  */
	      itype = build_index_type (build_int_2 (has_virtual, 0));
	      atype = build_array_type (vtable_entry_type, itype);
	      layout_type (atype);
	      vfield = copy_node (vfield);
	      TREE_TYPE (vfield) = build_pointer_type (atype);
#if 0
	      /* In the case of single inheritance, we can
		 just move up the tree, since we share the
		 same vptr slot.  */
	      if (TREE_CHAIN (vfields) == NULL_TREE)
		vfields = CLASSTYPE_AS_LIST (t);
#endif
	    }
	}

      CLASSTYPE_VFIELD (t) = vfield;
      if (TREE_TYPE (CLASS_ASSOC_VTABLE (t)) != atype)
	{
	  TREE_TYPE (CLASS_ASSOC_VTABLE (t)) = atype;
	  layout_decl (CLASS_ASSOC_VTABLE (t), 0);
	  DECL_ALIGN (CLASS_ASSOC_VTABLE (t))
	    = MAX (TYPE_ALIGN (double_type_node),
		   DECL_ALIGN (CLASS_ASSOC_VTABLE (t)));
	}
    }
  else if (first_vfn_base_index >= 0)
    CLASSTYPE_VFIELD (t) = vfield;
  CLASSTYPE_VFIELDS (t) = vfields;

  /* Set all appropriate CLASSTYPE_... flags for this type
     and its variants.  */
  TYPE_NEEDS_CONSTRUCTOR (t) |= needs_ctor || TYPE_HAS_CONSTRUCTOR (t);
  TYPE_NEEDS_CONSTRUCTING (t)
    |= ((TYPE_NEEDS_CONSTRUCTOR (t)|TYPE_USES_VIRTUAL_BASECLASSES (t))
	|| has_virtual || any_default_members
	|| first_vfn_base_index >= 0);
  TYPE_NEEDS_DESTRUCTOR (t) |= needs_dtor || TYPE_HAS_DESTRUCTOR (t);
  finish_struct_bits (t, max_has_virtual);

  /* Promote each bit-field's type to int if it is narrower than that.
     Also warn (or error) if static members are specified for a class
     which takes a constructor.  */
  for (x = fields; x; x = TREE_CHAIN (x))
    {
      if (DECL_BIT_FIELD (x)
	  && TREE_CODE (TREE_TYPE (x)) == INTEGER_TYPE
	  && (TREE_INT_CST_LOW (DECL_SIZE (x))
	      < TYPE_PRECISION (integer_type_node)))
	TREE_TYPE (x) = integer_type_node;
    }

  /* Now add the tags, if any, to the list of TYPE_DECLs
     defined for this type.  */
  if (CLASSTYPE_TAGS (t))
    {
      x = CLASSTYPE_TAGS (t);
      y = tree_last (TYPE_FIELDS (t));
      while (x)
	{
	  tree tag = build_lang_decl (TYPE_DECL, TREE_PURPOSE (x), TREE_VALUE (x));
	  DECL_CONTEXT (tag) = t;
	  x = TREE_CHAIN (x);
	  y = chainon (y, tag);
	}
      if (TYPE_FIELDS (t) == 0)
	TYPE_FIELDS (t) = y;
      CLASSTYPE_LOCAL_TYPEDECLS (t) = 1;
    }

  if (TYPE_HAS_CONSTRUCTOR (t))
    {
      tree vfields = CLASSTYPE_VFIELDS (t);

      while (vfields)
	{
	  /* Mark the fact that constructor for T
	     could affect anybody inheriting from T
	     who wants to initialize vtables for VFIELDS's type.  */
	  if (TREE_TYPE (vfields))
	    TREE_ADDRESSABLE (vfields) = 1;
	  vfields = TREE_CHAIN (vfields);
	}
      if (any_default_members != 0)
	build_class_init_list (t);
    }
  else if (TYPE_NEEDS_CONSTRUCTING (t))
    build_class_init_list (t);

  if (current_lang_name == lang_name_cplusplus)
    {
      if (! CLASSTYPE_DECLARED_EXCEPTION (t))
	embrace_waiting_friends (t);

      /* Write out inline function definitions.  */
      do_inline_function_hair (t, CLASSTYPE_INLINE_FRIENDS (t));
      CLASSTYPE_INLINE_FRIENDS (t) = 0;
    }

  if (CLASSTYPE_VSIZE (t) != 0)
    {
#if 0
      if (!TYPE_USES_COMPLEX_INHERITANCE (t))
	TYPE_NONCOPIED_PARTS (t) = build_tree_list (default_conversion (CLASS_ASSOC_VTABLE (t)), vfield);
#endif

      if ((flag_this_is_variable & 1) == 0)
	{
	  tree vtbl_ptr = build_decl (VAR_DECL, get_identifier (VPTR_NAME),
				      TREE_TYPE (vfield));
	  TREE_REGDECL (vtbl_ptr) = 1;
	  CLASSTYPE_VTBL_PTR (t) = vtbl_ptr;
	}
      if (DECL_FIELD_CONTEXT (vfield) != t)
	{
	  tree assoc = assoc_value (DECL_FIELD_CONTEXT (vfield), t, 0);
	  tree offset = ASSOC_OFFSET (assoc);

	  vfield = copy_node (vfield);

	  if (! integer_zerop (offset))
	    offset = size_binop (MULT_EXPR, offset, size_int (BITS_PER_UNIT));
	  DECL_FIELD_CONTEXT (vfield) = t;
	  DECL_FIELD_BITPOS (vfield)
	    = size_binop (PLUS_EXPR, offset, DECL_FIELD_BITPOS (vfield));
	  CLASSTYPE_VFIELD (t) = vfield;
	}
    }

  /* Make the rtl for any new vtables we have created, and unmark
     the base types we marked.  */
  unmark_finished_struct (t);
  TYPE_BEING_DEFINED (t) = 0;

  if (flag_dossier && CLASSTYPE_VTABLE_NEEDS_WRITING (t))
    {
      tree variants;
      tree tdecl;

      /* Now instantiate its type descriptors.  */
      tdecl = TREE_OPERAND (build_t_desc (t, 1), 0);
      variants = TYPE_POINTER_TO (t);
      build_type_variant (variants, 1, 0);
      while (variants)
	{
	  build_t_desc (variants, 1);
	  variants = TYPE_NEXT_VARIANT (variants);
	}
      variants = build_reference_type (t);
      build_type_variant (variants, 1, 0);
      while (variants)
	{
	  build_t_desc (variants, 1);
	  variants = TYPE_NEXT_VARIANT (variants);
	}
      DECL_VPARENT (tdecl) = t;
      DECL_CONTEXT (tdecl) = t;
    }
  /* Still need to instantiate this C struct's type descriptor.  */
  else if (flag_dossier && ! CLASSTYPE_DOSSIER (t))
    build_t_desc (t, 1);

  popclass (0);

  hack_incomplete_structures (t);

  resume_momentary (old);

  if (flag_cadillac)
    cadillac_finish_struct (t);

  return t;
}

/* Return non-zero if the effective type of INSTANCE is static.
   Used to determine whether the virtual function table is needed
   or not.

   *NONNULL is set iff INSTANCE can be known to be nonnull, regardless
   of our knownledge of its type.  */
int
resolves_to_fixed_type_p (instance, nonnull)
     tree instance;
     int *nonnull;
{
  switch (TREE_CODE (instance))
    {
    case INDIRECT_REF:
      /* Check that we are not going through a cast of some sort.  */
      if (TREE_TYPE (instance)
	  == TREE_TYPE (TREE_TYPE (TREE_OPERAND (instance, 0))))
	instance = TREE_OPERAND (instance, 0);
      /* fall through...  */
    case CALL_EXPR:
      /* This is a call to a constructor, hence it's never zero.  */
      if (TREE_HAS_CONSTRUCTOR (instance))
	{
	  if (nonnull)
	    *nonnull = 1;
	  return 1;
	}
      return 0;

      /* This is a call to a constructor, hence it's never zero.  */
      if (TREE_HAS_CONSTRUCTOR (instance))
	{
	  if (nonnull)
	    *nonnull = 1;
	  return 1;
	}
      return 0;

    case SAVE_EXPR:
      /* This is a call to a constructor, hence it's never zero.  */
      if (TREE_HAS_CONSTRUCTOR (instance))
	{
	  if (nonnull)
	    *nonnull = 1;
	  return 1;
	}
      return resolves_to_fixed_type_p (TREE_OPERAND (instance, 0), nonnull);

    case RTL_EXPR:
      /* This is a call to `new', hence it's never zero.  */
      if (TREE_CALLS_NEW (instance))
	{
	  if (nonnull)
	    *nonnull = 1;
	  return 1;
	}
      return 0;

    case PLUS_EXPR:
    case MINUS_EXPR:
      if (TREE_CODE (TREE_OPERAND (instance, 1)) == INTEGER_CST)
	/* Propagate nonnull.  */
	resolves_to_fixed_type_p (TREE_OPERAND (instance, 0), nonnull);
      return 0;

    case NOP_EXPR:
    case CONVERT_EXPR:
      return resolves_to_fixed_type_p (TREE_OPERAND (instance, 0), nonnull);

    case ADDR_EXPR:
      if (nonnull)
	*nonnull = 1;
      return resolves_to_fixed_type_p (TREE_OPERAND (instance, 0), nonnull);

    case COMPONENT_REF:
      return resolves_to_fixed_type_p (TREE_OPERAND (instance, 1), nonnull);

    case VAR_DECL:
    case PARM_DECL:
    case FIELD_DECL:
    case TARGET_EXPR:
      if (IS_AGGR_TYPE (TREE_TYPE (instance)))
	{
	  if (nonnull)
	    *nonnull = 1;
	  return 1;
	}
      else if (nonnull)
	{
	  if (instance == current_class_decl
	      && ! flag_this_is_variable)
	    {
	      /* Some people still use `this = 0' inside destructors.  */
	      *nonnull = ! DESTRUCTOR_NAME_P (DECL_NAME (current_function_decl));
	      /* In a constructor, we know our type.  */
	      if (DECL_CONSTRUCTOR_P (current_function_decl))
		return 1;
	    }
	  else if (TREE_CODE (TREE_TYPE (instance)) == REFERENCE_TYPE)
	    /* Reference variables should be refernces to objects.  */
	    *nonnull = 1;
	}
      return 0;

    default:
      return 0;
    }
}

void
init_class_processing ()
{
  current_class_depth = 0;
  current_class_stacksize = 10;
  current_class_base = (tree *)xmalloc(current_class_stacksize * sizeof (tree));
  current_class_stack = current_class_base;

  current_lang_stacksize = 10;
  current_lang_base = (tree *)xmalloc(current_lang_stacksize * sizeof (tree));
  current_lang_stack = current_lang_base;

  delta_name = get_identifier (VTABLE_DELTA_NAME);
  pfn_name = get_identifier (VTABLE_PFN_NAME);

  /* Keep these values lying around.  */
  minus_one_node = build_int_2 (-1, 0);
  the_null_vtable_entry = build_vtable_entry (integer_zero_node, integer_zero_node);
  base_layout_decl = build_lang_field_decl (FIELD_DECL, NULL_TREE, error_mark_node);
  TREE_TYPE (base_layout_decl) = make_node (RECORD_TYPE);

  gcc_obstack_init (&class_obstack);
}

/* Set current scope to NAME. CODE tells us if this is a
   STRUCT, UNION, or ENUM environment.

   NAME may end up being NULL_TREE if this is an anonymous or
   late-bound struct (as in "struct { ... } foo;")  */

/* Here's a subroutine we need because C lacks lambdas.  */
static void
unuse_fields (type)
     tree type;
{
  tree fields;

  for (fields = TYPE_FIELDS (type); fields; fields = TREE_CHAIN (fields))
    {
      if (TREE_CODE (fields) != FIELD_DECL)
	continue;

      TREE_USED (fields) = 0;
      if (DECL_ANON_UNION_ELEM (fields))
	unuse_fields (TREE_TYPE (fields));
    }
}

/* Set global variables CURRENT_CLASS_NAME and CURRENT_CLASS_TYPE to
   appropriate values, found by looking up the type definition of
   NAME (as a CODE).

   If MODIFY is 1, we set IDENTIFIER_CLASS_VALUE's of names
   which can be seen locally to the class. They are shadowed by
   any subsequent local declaration (including parameter names).

   If MODIFY is 2, we set IDENTIFIER_CLASS_VALUE's of names
   which have static meaning (i.e., static members, static
   member functions, enum declarations, etc).

   So that we may avoid calls to lookup_name, we cache the _TYPE
   nodes of local TYPE_DECLs in the TREE_TYPE field of the name.

   For multiple inheritance, we perform a two-pass depth-first search
   of the type lattice.  The first pass performs a pre-order search,
   marking types after the type has had its fields installed in
   the appropriate IDENTIFIER_CLASS_VALUE slot.  The second pass merely
   unmarks the marked types.  If a field or member function name
   appears in an ambiguous way, the IDENTIFIER_CLASS_VALUE of
   that name becomes `error_mark_node'.  */

void
pushclass (type, modify)
     tree type;
     int modify;
{
  push_memoized_context (type, modify);

  current_class_depth++;
  *current_class_stack++ = current_class_name;
  *current_class_stack++ = current_class_type;
  if (current_class_stack >= current_class_base + current_class_stacksize)
    {
      current_class_base =
	(tree *)xrealloc (current_class_base,
			  sizeof (tree) * (current_class_stacksize + 10));
      current_class_stack = current_class_base + current_class_stacksize;
      current_class_stacksize += 10;
    }

  type = TYPE_MAIN_VARIANT (type);
  current_class_name = TYPE_NAME (type);
  if (TREE_CODE (current_class_name) == TYPE_DECL)
    current_class_name = DECL_NAME (current_class_name);
  current_class_type = type;

  if (type != prev_class_type && prev_class_type != NULL_TREE
      && current_class_depth == 1)
    {
      /* Forcibly remove any old class remnants.  */
      popclass (-1);
      prev_class_type = 0;
    }

  pushlevel_class ();

  if (modify)
    {
      tree tags;
      tree this_fndecl = current_function_decl;

      if (current_function_decl
	  && DECL_CONTEXT (current_function_decl)
	  && TREE_CODE (DECL_CONTEXT (current_function_decl)) == FUNCTION_DECL)
	current_function_decl = DECL_CONTEXT (current_function_decl);
      else
	current_function_decl = NULL_TREE;

      if (type != prev_class_type)
	{
	  build_mi_matrix (type);
	  push_class_decls (type);
	  free_mi_matrix ();
	  prev_class_type = type;
	}
      else
	unuse_fields (type);

      tags = CLASSTYPE_TAGS (type);
      while (tags)
	{
	  TREE_NONLOCAL_FLAG (TREE_VALUE (tags)) = 1;
	  pushtag (TREE_PURPOSE (tags), TREE_VALUE (tags));
	  if (IDENTIFIER_CLASS_VALUE (TREE_PURPOSE (tags)) == NULL_TREE
	      && TREE_CODE (TYPE_NAME (TREE_VALUE (tags))) == TYPE_DECL)
	    IDENTIFIER_CLASS_VALUE (TREE_PURPOSE (tags))
	      = TYPE_NAME (TREE_VALUE (tags));
	  tags = TREE_CHAIN (tags);
	}

      current_function_decl = this_fndecl;
    }

  if (flag_cadillac)
    cadillac_push_class (type);
}
 
/* Get out of the current class scope. If we were in a class scope
   previously, that is the one popped to.  The flag MODIFY tells
   whether the current scope declarations needs to be modified
   as a result of popping to the previous scope.  */
void
popclass (modify)
     int modify;
{
  if (flag_cadillac)
    cadillac_pop_class ();

  if (modify < 0)
    {
      /* Back this old class out completely.  */
      tree tags = CLASSTYPE_TAGS (prev_class_type);

      pop_class_decls (prev_class_type);
      while (tags)
	{
	  TREE_NONLOCAL_FLAG (TREE_VALUE (tags)) = 0;
	  IDENTIFIER_CLASS_VALUE (TREE_PURPOSE (tags)) = NULL_TREE;
	  tags = TREE_CHAIN (tags);
	}
      return;
    }

  if (modify)
    {
      /* Just remove from this class what didn't make
	 it into IDENTIFIER_CLASS_VALUE.  */
      tree tags = CLASSTYPE_TAGS (current_class_type);

      while (tags)
	{
	  TREE_NONLOCAL_FLAG (TREE_VALUE (tags)) = 0;
	  IDENTIFIER_CLASS_VALUE (TREE_PURPOSE (tags)) = NULL_TREE;
	  tags = TREE_CHAIN (tags);
	}
    }
  poplevel_class ();

  current_class_depth--;
  current_class_type = *--current_class_stack;
  current_class_name = *--current_class_stack;

  if (current_class_type)
    {
      if (CLASSTYPE_VTBL_PTR (current_class_type))
	{
	  current_vtable_decl = lookup_name (DECL_NAME (CLASSTYPE_VTBL_PTR (current_class_type)));
	  if (current_vtable_decl)
	    current_vtable_decl = build_indirect_ref (current_vtable_decl, 0);
	}
      current_class_decl = lookup_name (get_identifier (THIS_NAME));
      if (current_class_decl)
	{
	  if (TREE_CODE (TREE_TYPE (current_class_decl)) == POINTER_TYPE)
	    {
	      tree temp;
	      /* Can't call build_indirect_ref here, because it has special
		 logic to return C_C_D given this argument.  */
	      C_C_D = build1 (INDIRECT_REF, current_class_type, current_class_decl);
	      temp = TREE_TYPE (TREE_TYPE (current_class_decl));
	      TREE_READONLY (C_C_D) = TYPE_READONLY (temp);
	      TREE_SIDE_EFFECTS (C_C_D) = TYPE_VOLATILE (temp);
	      TREE_THIS_VOLATILE (C_C_D) = TYPE_VOLATILE (temp);
	    }
	  else
	    C_C_D = current_class_decl;
	}
      else C_C_D = NULL_TREE;
    }
  else
    {
      current_class_decl = NULL_TREE;
      current_vtable_decl = NULL_TREE;
      C_C_D = NULL_TREE;
    }

  pop_memoized_context (modify);
}

/* Set global variables CURRENT_LANG_NAME to appropriate value
   so that behavior of name-mangline machinery is correct.  */

void
push_lang_context (name)
     tree name;
{
  *current_lang_stack++ = current_lang_name;
  if (current_lang_stack >= current_lang_base + current_lang_stacksize)
    {
      current_lang_base =
	(tree *)xrealloc (current_lang_base,
			  sizeof (tree) * (current_lang_stacksize + 10));
      current_lang_stack = current_lang_base + current_lang_stacksize;
      current_lang_stacksize += 10;
    }

  if (name == lang_name_cplusplus)
    {
      strict_prototype = strict_prototypes_lang_cplusplus;
      current_lang_name = name;
      install_reserved_words (lang_cplusplus);
    }
  else if (name == lang_name_c)
    {
      strict_prototype = strict_prototypes_lang_c;
      current_lang_name = name;
      install_reserved_words(lang_c);
    }
  else if (name == lang_name_objc)
    {
      /* for now, treat Objective-C like C, other code depends on this */
      current_lang_name = lang_name_c;
      install_reserved_words (lang_objc);
    }
  else
    error ("language string `\"%s\"' not recognized", IDENTIFIER_POINTER (name));

  if (flag_cadillac)
    cadillac_push_lang (name);
}
  
/* Get out of the current language scope.  */
void
pop_lang_context ()
{
  if (flag_cadillac)
    cadillac_pop_lang ();

  current_lang_name = *--current_lang_stack;

  if (current_lang_name == lang_name_cplusplus)
    {
      strict_prototype = strict_prototypes_lang_cplusplus;
      install_reserved_words (lang_cplusplus);
    }
  else if (current_lang_name == lang_name_c)
    {
      strict_prototype = strict_prototypes_lang_c;
      install_reserved_words (lang_c);
    }
  else if (current_lang_name == lang_name_objc)
    {
      strict_prototype = strict_prototypes_lang_c;
      install_reserved_words (lang_objc);
    }
}

int
root_lang_context_p ()
{
  return current_lang_stack == current_lang_base;
}

/* Type instantiation routines.  */

/* This function will instantiate the type of the expression given
   in RHS to match the type of LHSTYPE.  If LHSTYPE is NULL_TREE,
   or other errors exist, the TREE_TYPE of RHS will be ERROR_MARK_NODE.

   This function is used in build_modify_expr, convert_arguments,
   build_c_cast, and compute_conversion_costs.  */
tree
instantiate_type (lhstype, rhs, complain)
     tree lhstype, rhs;
     int complain;
{
  if (TREE_CODE (rhs) == OP_IDENTIFIER)
    return build_instantiated_decl (lhstype, rhs);

  if (TREE_CODE (lhstype) == UNKNOWN_TYPE)
    {
      if (complain)
	error ("not enough type information");
      return error_mark_node;
    }

  if (TREE_TYPE (rhs) != NULL_TREE && ! (type_unknown_p (rhs)))
    return rhs;

  /* This should really only be used when attempting to distinguish
     what sort of a pointer to function we have.  For now, any
     arithmethic operation which is not supported on pointers
     is rejected as an error.  */

  switch (TREE_CODE (rhs))
    {
    case TYPE_EXPR:
    case CONVERT_EXPR:
    case SAVE_EXPR:
    case CONSTRUCTOR:
    case BUFFER_REF:
      assert (0);
      return error_mark_node;

    case INDIRECT_REF:
    case ARRAY_REF:
      TREE_TYPE (rhs) = lhstype;
      lhstype = build_pointer_type (lhstype);
      TREE_OPERAND (rhs, 0)
	= instantiate_type (lhstype, TREE_OPERAND (rhs, 0), complain);
      if (TREE_OPERAND (rhs, 0) == error_mark_node)
	return error_mark_node;

      return rhs;

    case NOP_EXPR:
      rhs = copy_node (TREE_OPERAND (rhs, 0));
      TREE_TYPE (rhs) = unknown_type_node;
      return instantiate_type (lhstype, rhs, complain);

    case COMPONENT_REF:
      {
	tree field = TREE_OPERAND (rhs, 1);
	if (TREE_CODE (field) == TREE_LIST)
	  {
	    tree function = instantiate_type (lhstype, field, complain);
	    if (function == error_mark_node)
	      return error_mark_node;
	    assert (TREE_CODE (function) == FUNCTION_DECL);
	    if (DECL_VIRTUAL_P (function))
	      {
		tree base = TREE_OPERAND (rhs, 0);
		tree base_ptr = build_unary_op (ADDR_EXPR, base, 0);
		if (base_ptr == error_mark_node)
		  return error_mark_node;
		base_ptr = convert_pointer_to (DECL_VCONTEXT (function), base_ptr);
		if (base_ptr == error_mark_node)
		  return error_mark_node;
		return build_vfn_ref (&base_ptr, base, DECL_VINDEX (function));
	      }
	    return function;
	  }

	assert (TREE_CODE (field) == FIELD_DECL);
	assert (!(TREE_CODE (TREE_TYPE (field)) == FUNCTION_TYPE
		  || TREE_CODE (TREE_TYPE (field)) == METHOD_TYPE));

	TREE_TYPE (rhs) = lhstype;
	/* First look for an exact match  */

	while (field && TREE_TYPE (field) != lhstype)
	  field = TREE_CHAIN (field);
	if (field)
	  {
	    TREE_OPERAND (rhs, 1) = field;
	    return rhs;
	  }

	/* No exact match found, look for a compatible function.  */
	field = TREE_OPERAND (rhs, 1);
	while (field && ! comptypes (lhstype, TREE_TYPE (field), 0))
	  field = TREE_CHAIN (field);
	if (field)
	  {
	    TREE_OPERAND (rhs, 1) = field;
	    field = TREE_CHAIN (field);
	    while (field && ! comptypes (lhstype, TREE_TYPE (field), 0))
	      field = TREE_CHAIN (field);
	    if (field)
	      {
		if (complain)
		  error ("ambiguous overload for COMPONENT_REF requested");
		return error_mark_node;
	      }
	  }
	else
	  {
	    if (complain)
	      error ("no appropriate overload exists for COMPONENT_REF");
	    return error_mark_node;
	  }
	return rhs;
      }

    case TREE_LIST:
      {
	tree elem, baselink, name;
	int globals = overloaded_globals_p (rhs);

	/* If there's only one function we know about, return that.  */
	if (globals > 0 && TREE_CHAIN (rhs) == NULL_TREE)
	  return TREE_VALUE (rhs);

	/* First look for an exact match.  Search either overloaded
	   functions or member functions.  May have to undo what
	   `default_conversion' might do to lhstype.  */

	if (TREE_CODE (lhstype) == POINTER_TYPE)
	  if (TREE_CODE (TREE_TYPE (lhstype)) == FUNCTION_TYPE
	      || TREE_CODE (TREE_TYPE (lhstype)) == METHOD_TYPE)
	    lhstype = TREE_TYPE (lhstype);
	  else
	    {
	      if (complain)
		error ("invalid type combination for overload");
	      return error_mark_node;
	    }

	if (TREE_CODE (lhstype) != FUNCTION_TYPE && globals > 0)
	  {
	    if (complain)
	      error ("cannot resolve overloaded function `%s' based on non-function type",
		     IDENTIFIER_POINTER (TREE_PURPOSE (rhs)));
	    return error_mark_node;
	  }

	if (globals > 0)
	  {
	    assert (TREE_CODE (TREE_VALUE (rhs)) == FUNCTION_DECL);
	    elem = rhs;
	    while (elem)
	      if (TREE_TYPE (TREE_VALUE (elem)) != lhstype)
		elem = TREE_CHAIN (elem);
	      else
		return TREE_VALUE (elem);
	    /* No exact match found, look for a compatible function.  */
	    elem = rhs;
	    while (elem && ! comp_target_types (lhstype, TREE_TYPE (TREE_VALUE (elem)), 1))
	      elem = TREE_CHAIN (elem);
	    if (elem)
	      {
		tree save_elem = TREE_VALUE (elem);
		elem = TREE_CHAIN (elem);
		while (elem && ! comp_target_types (lhstype, TREE_TYPE (TREE_VALUE (elem)), 0))
		  elem = TREE_CHAIN (elem);
		if (elem)
		  {
		    if (complain)
		      error ("ambiguous overload for overloaded function requested");
		    return error_mark_node;
		  }
		return save_elem;
	      }
	    if (complain)
	      {
		if (TREE_CHAIN (rhs))
		  error ("no appropriate overload for overloaded function `%s' exists",
			 IDENTIFIER_POINTER (TREE_PURPOSE (rhs)));
		else
		  error ("function `%s' has inappropriate type signature",
			 IDENTIFIER_POINTER (TREE_PURPOSE (rhs)));
	      }
	    return error_mark_node;
	  }

	if (TREE_NONLOCAL_FLAG (rhs))
	  {
	    /* Got to get it as a baselink.  */
	    rhs = lookup_fnfields (CLASSTYPE_AS_LIST (current_class_type),
				   TREE_PURPOSE (rhs), 0);
	  }
	else
	  {
	    assert (TREE_CHAIN (rhs) == NULL_TREE);
	    if (TREE_CODE (TREE_VALUE (rhs)) == TREE_LIST)
	      rhs = TREE_VALUE (rhs);
	    assert (TREE_CODE (TREE_VALUE (rhs)) == FUNCTION_DECL);
	  }

	for (baselink = rhs; baselink;
	     baselink = next_baselink (baselink))
	  {
	    elem = TREE_VALUE (baselink);
	    while (elem)
	      if (TREE_TYPE (elem) != lhstype)
		elem = TREE_CHAIN (elem);
	      else
		return elem;
	  }

	/* No exact match found, look for a compatible method.  */
	for (baselink = rhs; baselink;
	     baselink = next_baselink (baselink))
	  {
	    elem = TREE_VALUE (baselink);
	    while (elem && ! comp_target_types (lhstype, TREE_TYPE (elem), 1))
	      elem = TREE_CHAIN (elem);
	    if (elem)
	      {
		tree save_elem = elem;
		elem = TREE_CHAIN (elem);
		while (elem && ! comp_target_types (lhstype, TREE_TYPE (elem), 0))
		  elem = TREE_CHAIN (elem);
		if (elem)
		  {
		    if (complain)
		      error ("ambiguous overload for overloaded method requested");
		    return error_mark_node;
		  }
		return save_elem;
	      }
	    name = DECL_ORIGINAL_NAME (TREE_VALUE (rhs));
	    if (TREE_CODE (lhstype) == FUNCTION_TYPE && globals < 0)
	      {
		/* Try to instantiate from non-member functions.  */
		rhs = IDENTIFIER_GLOBAL_VALUE (name);
		if (rhs && TREE_CODE (rhs) == TREE_LIST)
		  {
		    /* This code seems to be missing a `return'.  */
		    abort ();
		    instantiate_type (lhstype, rhs, complain);
		  }
	      }

	    if (complain)
	      error ("no static member functions named `%s'",
		     IDENTIFIER_POINTER (name));
	    return error_mark_node;
	  }
      }

    case CALL_EXPR:
      /* This is too hard for now.  */
      assert (0);
      return error_mark_node;

    case PLUS_EXPR:
    case MINUS_EXPR:
    case COMPOUND_EXPR:
      TREE_OPERAND (rhs, 0) = instantiate_type (lhstype, TREE_OPERAND (rhs, 0), complain);
      if (TREE_OPERAND (rhs, 0) == error_mark_node)
	return error_mark_node;
      TREE_OPERAND (rhs, 1) = instantiate_type (lhstype, TREE_OPERAND (rhs, 1), complain);
      if (TREE_OPERAND (rhs, 1) == error_mark_node)
	return error_mark_node;

      TREE_TYPE (rhs) = lhstype;
      return rhs;

    case MULT_EXPR:
    case TRUNC_DIV_EXPR:
    case FLOOR_DIV_EXPR:
    case CEIL_DIV_EXPR:
    case ROUND_DIV_EXPR:
    case RDIV_EXPR:
    case TRUNC_MOD_EXPR:
    case FLOOR_MOD_EXPR:
    case CEIL_MOD_EXPR:
    case ROUND_MOD_EXPR:
    case FIX_ROUND_EXPR:
    case FIX_FLOOR_EXPR:
    case FIX_CEIL_EXPR:
    case FIX_TRUNC_EXPR:
    case FLOAT_EXPR:
    case NEGATE_EXPR:
    case ABS_EXPR:
    case MAX_EXPR:
    case MIN_EXPR:
    case FFS_EXPR:

    case BIT_AND_EXPR:
    case BIT_IOR_EXPR:
    case BIT_XOR_EXPR:
    case LSHIFT_EXPR:
    case RSHIFT_EXPR:
    case LROTATE_EXPR:
    case RROTATE_EXPR:

    case PREINCREMENT_EXPR:
    case PREDECREMENT_EXPR:
    case POSTINCREMENT_EXPR:
    case POSTDECREMENT_EXPR:
      if (complain)
	error ("illegal operation on uninstantiated type");
      return error_mark_node;

    case TRUTH_AND_EXPR:
    case TRUTH_OR_EXPR:
    case LT_EXPR:
    case LE_EXPR:
    case GT_EXPR:
    case GE_EXPR:
    case EQ_EXPR:
    case NE_EXPR:
    case TRUTH_ANDIF_EXPR:
    case TRUTH_ORIF_EXPR:
    case TRUTH_NOT_EXPR:
      if (complain)
	error ("not enough type information");
      return error_mark_node;

    case COND_EXPR:
      if (type_unknown_p (TREE_OPERAND (rhs, 0)))
	{
	  if (complain)
	    error ("not enough type information");
	  return error_mark_node;
	}
      TREE_OPERAND (rhs, 1) = instantiate_type (lhstype, TREE_OPERAND (rhs, 1), complain);
      if (TREE_OPERAND (rhs, 1) == error_mark_node)
	return error_mark_node;
      TREE_OPERAND (rhs, 2) = instantiate_type (lhstype, TREE_OPERAND (rhs, 2), complain);
      if (TREE_OPERAND (rhs, 2) == error_mark_node)
	return error_mark_node;

      TREE_TYPE (rhs) = lhstype;
      return rhs;

    case MODIFY_EXPR:
      TREE_OPERAND (rhs, 1) = instantiate_type (lhstype, TREE_OPERAND (rhs, 1), complain);
      if (TREE_OPERAND (rhs, 1) == error_mark_node)
	return error_mark_node;

      TREE_TYPE (rhs) = lhstype;
      return rhs;
      
    case ADDR_EXPR:
      if (TREE_CODE (lhstype) != POINTER_TYPE)
	{
	  if (complain)
	    error ("type for resolving address of overloaded function must be pointer type");
	  return error_mark_node;
	}
      TREE_TYPE (rhs) = lhstype;
      lhstype = TREE_TYPE (lhstype);
      TREE_OPERAND (rhs, 0) = instantiate_type (lhstype, TREE_OPERAND (rhs, 0), complain);
      if (TREE_OPERAND (rhs, 0) == error_mark_node)
	return error_mark_node;

      mark_addressable (TREE_OPERAND (rhs, 0));
      return rhs;

    case ENTRY_VALUE_EXPR:
      assert (0);
      return error_mark_node;

    case ERROR_MARK:
      return error_mark_node;

    default:
      assert (0);
      return error_mark_node;
    }
}

/* This routine is called when we finally know the type of expression
   we are looking for.  If the operator encoded by EXP can take an
   argument of type TYPE, return the FUNCTION_DECL for that operator.  */
tree
build_instantiated_decl (type, exp)
     tree type, exp;
{
  tree parmtypes, decl, name;

  assert (TREE_CODE (exp) == OP_IDENTIFIER);
  type = TREE_TYPE (type);
  if (TREE_CODE (type) != POINTER_TYPE
      || (TREE_CODE (TREE_TYPE (type)) != FUNCTION_TYPE
	  && TREE_CODE (TREE_TYPE (type)) != METHOD_TYPE))
    {
      error ("invalid type used to resolve overloaded function");
      return error_mark_node;
    }


  /* Now we know the type of this function, so overload it.  */
  parmtypes = TYPE_ARG_TYPES (TREE_TYPE (type));
  name = build_operator_fnname (&exp, parmtypes, 0);
  if (name)
    {
      name = build_decl_overload (IDENTIFIER_POINTER (name), parmtypes, 1);
      decl = lookup_name (name);
      if (decl)
	return decl;
      error ("no suitable declaration of `operator %s' for overloading",
	     operator_name_string (name));
      return error_mark_node;
    }
  return error_mark_node;
}

/* Return the name of the virtual function table (as an IDENTIFIER_NODE)
   for the given TYPE.  */
static tree
get_vtable_name (type)
     tree type;
{
  char *buf = (char *)alloca (sizeof (VTABLE_NAME_FORMAT)
			      + TYPE_NAME_LENGTH (type) + 2);
  sprintf (buf, VTABLE_NAME_FORMAT, TYPE_NAME_STRING (type));
  return get_identifier (buf);
}

/* Return the name of the virtual function pointer field
   (as an IDENTIFIER_NODE) for the given TYPE.  Note that
   this may have to look back through base types to find the
   ultimate field name.  (For single inheritance, these could
   all be the same name.  Who knows for multiple inheritance).  */
static tree
get_vfield_name (type)
     tree type;
{
  char *buf;

  while (TYPE_BASETYPES (type)
	 && TYPE_VIRTUAL_P (CLASSTYPE_BASECLASS (type, 0))
	 && ! CLASSTYPE_VIA_VIRTUAL (type, 0))
    type = CLASSTYPE_BASECLASS (type, 0);

  buf = (char *)alloca (sizeof (VFIELD_NAME_FORMAT)
			+ TYPE_NAME_LENGTH (type) + 2);
  sprintf (buf, VFIELD_NAME_FORMAT, TYPE_NAME_STRING (type));
  return get_identifier (buf);
}

void
print_class_statistics ()
{
#ifdef GATHER_STATISTICS
  fprintf (stderr, "convert_harshness = %d\n", n_convert_harshness);
  fprintf (stderr, "compute_conversion_costs = %d\n", n_compute_conversion_costs);
  fprintf (stderr, "build_method_call = %d (inner = %d)\n",
	   n_build_method_call, n_inner_fields_searched);
  if (n_vtables)
    {
      fprintf (stderr, "vtables = %d; vtable searches = %d\n",
	       n_vtables, n_vtable_searches);
      fprintf (stderr, "vtable entries = %d; vtable elems = %d\n",
	       n_vtable_entries, n_vtable_elems);
    }
#endif
}
