#import "RelocManager.h"
#import <stdlib.h>
#import <string.h>
#import <mach-o/loader.h>
#import <libc.h>

@implementation RelocManager

-init
{
    rmFlags.invalid = YES;
    return self;
}

-free
{
    [self invalidate];
    return [super free];
}

-(Reloc *)relocFor: (const void *)pointer
{
    Reloc *reloc;
    int count;
    BOOL found;
    if (rmFlags.invalid)
	[self readInAllRelocs];
    if (lastReloc
	&& (lastReloc->address <= (pointer_t)pointer) 
	&& ((pointer_t)pointer < (lastReloc->maxAddress)))
	return lastReloc;
    else {
	reloc = relocs;
	for (count = numRelocs, found = NO; count && !found; count--) {
	    if ((reloc->address <= (pointer_t)pointer) 
		                && ((pointer_t)pointer < (reloc->maxAddress)))
		found = YES;
	    else
		((void *)reloc) += relocSize;
	}
	if (found && (reloc->rFlags.readIn || [self readInReloc: reloc])) {
	    lastReloc = reloc;
	    return reloc;
	} else
	    return NULL;
    }
}

-(BOOL)readInReloc: (Reloc *)reloc
{
    [self subclassResponsibility: _cmd];
    return NO;
}

-(void)readInAllRelocs
{
    [self subclassResponsibility: _cmd];
}

-invalidate
{
    if (relocs) {
	free(relocs); relocs = NULL;
    }
    numRelocs = 0;
    lastReloc = NULL;
    rmFlags.invalid = YES;
    return self;
}

-(Reloc *)oldRelocFor: (const void *)pointer
{
   int count;
   BOOL found;
   Reloc *reloc;
    if (lastReloc
	&& (lastReloc->data <= (pointer_t)pointer) 
	&& ((pointer_t)pointer < lastReloc->maxData))
	return lastReloc;
    else {
	reloc = relocs;
	for (count = numRelocs, found = NO; count && !found; count--) {
	    if (reloc->rFlags.readIn && (reloc->data <= (pointer_t)pointer) 
		&& ((pointer_t)pointer < reloc->maxData))
		found = YES;
	    else
		((void *)reloc) += relocSize;
	}
	if (found) {
	    lastReloc = reloc;
	    return reloc;
	} else
	    return NULL;
    }
}

-(void *)originalPointerFor: (const void *)pointer
{	
    Reloc *reloc = [self oldRelocFor: pointer];
    if (reloc)
	return (void *)(reloc->address + (pointer - reloc->data));
    else 
	return NULL;
}

-(void *)pointerFor: (const void *)pointer
{
    Reloc *reloc;
    if (pointer && (reloc = [self  relocFor: pointer])) 
	return (void *)((pointer_t)pointer + reloc->displacement);
    else
	return NULL;
}

-(char *)pointerForString: (const char *)pointer isNullTerminated: (BOOL *)isNT
{
    char *retPointer;
    Reloc *reloc;
    char *string;
    if (pointer && (reloc = [self relocFor: pointer])) {
    	retPointer = (char *)((pointer_t)pointer + reloc->displacement);
	for (string = retPointer;
	     *string && ((pointer_t)string < reloc->maxData);
	     string++);
	*isNT = *string ? NO : YES;
	return retPointer;
    } else{
	*isNT = NO;
	return NULL;
    }
}

-(char *)pointerForString: (const char *)string
{
    BOOL isNT;
    char *retPointer;
    if ((retPointer = [self pointerForString: string isNullTerminated: &isNT])
	&& isNT)
	return retPointer;
    else
	return NULL;
}

-(void *)pointerFor: (const void *)pointer withSize: (int)size
{
    Reloc *reloc;
    pointer_t newPointer;
    if (pointer && (reloc = [self relocFor: pointer])) {
	newPointer = (pointer_t)pointer + reloc->displacement;
	if ((newPointer + size) <= reloc->maxData)
	    return (void *)newPointer;
	else
	    return NULL;
    } else
	return NULL;
}

-(id)pointerForID: (const id)pointer
{
    Class theClass;
    struct _ObjectID {
	@defs(Object)
    } *newID;
    
    if (pointer
        && (newID = [self pointerFor: pointer withSize: sizeof(id)])
	&& (theClass = [self pointerFor: newID->isa
	                       withSize: sizeof(Class)])
	&& (newID
	    = [self pointerFor: pointer withSize: theClass->instance_size]))
	return (id)newID;
    else
	return nil;
}
		    
-(int)getDataAt: (const void *)start for: (int)numBytes into: (void *)data
{
    Reloc *reloc = [self relocFor: start];
    int numBytesInReloc;
    
    if (reloc) {
        numBytesInReloc = reloc->maxAddress - (int)start;
        if (numBytes > numBytesInReloc)
	    numBytes = numBytesInReloc;
	memcpy(data,
	       (void *)(reloc->data + ((pointer_t)start - reloc->address)),
	       numBytes);
	return numBytes;
    } else
	return 0;
}

-(struct mach_header *)getMachHeader
{
    [self subclassResponsibility: _cmd];
    return NULL;
}

-(int)getNumMachHeaders
{
    [self subclassResponsibility: _cmd];
    return 0;
}

-(struct mach_header **)getMachHeaders
{
    [self subclassResponsibility: _cmd];
    return NULL;
}

-(struct mach_header **)getMachHeadersWithNames: (char ***)names
{
    [self subclassResponsibility: _cmd];
    return NULL;
}

-(void *)getSectData: (STR)segName 
	     section: (STR)sectName 
		size: (int *)pSize 
	   forHeader: (struct mach_header *)header
{
    void *data = getsectdatafromheader(header, segName, sectName, pSize);
    return [self pointerFor: data withSize: *pSize];
}

-(void *)getSectData: (STR)segName 
	     section: (STR)sectName 
		size: (int *)pSize 
{
    return [self getSectData: segName 
		     section: sectName 
			size: pSize 
		   forHeader: [self getMachHeader]];
}

-(const struct section *)getSeg: (STR)segName sect: (STR)sectName
{
    return getsectbynamefromheader([self getMachHeader], segName, sectName);
}

@end