#import "SegmentManager.h"
#import <stdlib.h>
#import <string.h>
#import <mach-o/loader.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <mach-o/ldsyms.h>
#import <libc.h>

@implementation SegmentManager

-initFile: (STR)theName
{
    relocSize = sizeof(Segment);
    name = malloc(strlen(theName) + 1);
    strcpy(name, theName);
    [self readInAllRelocs];
    lastReloc = NULL;
    if (numRelocs)
	return self;
    else {
	[self free];
	return nil;
    }
}

+newExecutable: (STR)theName
{
    SegmentManager *sm = [[super new] initFile: (STR)theName];
    if (sm && [sm isExecutable])
	 return sm;
    else {
	[sm free];
	return nil;
    }
}

+newCore: (STR)theName
{
    SegmentManager *sm = [[super new] initFile: (STR)theName];
    if (sm && [sm isCore])
	 return sm;
    else {
	[sm free];
	return nil;
    }
}

+newFile: (STR)theName
{
    return [[super new] initFile: (STR)theName];
}

-invalidate
{
    int count;
    Image *image;
    for (count = numImages, image = images; count; count--, image++)
	vm_deallocate(task_self(),
			(vm_address_t)image->header,
			image->size);
    free(images); images = NULL;
    numImages = 0;
    free(name); name = NULL;
    return [super invalidate];
}

-(void)getImages
{
    int imageFd;
    struct stat imageStat;
    struct load_command *loadCmd;
    struct fvmlib_command *fvmCmd;
    struct mach_header *header;
    int i;
    Image *image;
    numImages = 0;
    imageFd = open(name, O_RDONLY, 0);
    if (imageFd >= 0) {
	fstat(imageFd, &imageStat);
	if (map_fd(imageFd,
		   0,
		   &(vm_address_t)header,
		   YES,
		   imageStat.st_size) == 0) {
	    if (header->magic == MH_MAGIC) {
		numImages = 1;
		for (i = 0, loadCmd = (struct load_command *)(header + 1);
		     i < header->ncmds;
		     i++, ((void *)loadCmd) += loadCmd->cmdsize) {
		    if (loadCmd->cmd == LC_LOADFVMLIB) 
			numImages++;
		}
		image
		= images
		= calloc(numImages, sizeof(Image));
		image->name = name;
		image->header = header;
		image->size = imageStat.st_size;
		image++;
		for (i = 0, loadCmd = (struct load_command *)(header + 1);
		     i < header->ncmds;
		     i++, ((void *)loadCmd) += loadCmd->cmdsize) {
		    if (loadCmd->cmd == LC_LOADFVMLIB) {
			fvmCmd = (struct fvmlib_command *)loadCmd;
			image->name
			= (void *)fvmCmd + fvmCmd->fvmlib.name.offset;
			imageFd = open(image->name, O_RDONLY, 0);
			if (imageFd >= 0) {
			    fstat(imageFd, &imageStat);
			    image->size = imageStat.st_size;
			    if (map_fd(imageFd,
				       0,
				       &(vm_address_t)image->header,
				       YES,
				       image->size)
				      == 0)
				image++;
			    else
				numImages--;
			} else
			    numImages--;
		    }
		}
	    } else {
		return;
	    }
	}
    }
}
						    
-(int)numSegments
{
    int count, numSegs = 0, i;
    Image *image;
    struct load_command *loadCmd;
    struct mach_header *header;
    for (count = numImages,
	 image = images;
	 count;
	 count--, image++) {
	header = image->header;
	for (i = 0, loadCmd = (struct load_command *)(header + 1);
	     i < header->ncmds;
	     i++, ((void *)loadCmd) += loadCmd->cmdsize) {
	    if (loadCmd->cmd == LC_SEGMENT) 
		numSegs++;
	}
    }
    return numSegs;
}

-(BOOL)validate
{
    int nImages, i;
    Image *image;
    struct mach_header *header;
    struct load_command *loadCmd;
    struct segment_command *segCmd;
    BOOL allOK = YES;
    for (nImages = numImages,
	 image = images;
	 allOK && nImages;
	 nImages--, image++) {
	header = image->header;
	for (i = 0, loadCmd = (struct load_command *)(header + 1);
	     allOK && i < header->ncmds;
	     i++, ((void *)loadCmd) += loadCmd->cmdsize) {
	    if (loadCmd->cmd == LC_SEGMENT) {
		segCmd = (struct segment_command *)loadCmd;
		if ((segCmd->fileoff + segCmd->vmsize) > image->size)
		    allOK = NO;
	    }
	    if (!loadCmd->cmdsize)
		allOK = NO;
	}
    }
    return allOK;
}

-(void)readInSegments
{
    int nImages, i;
    Image *image;
    struct mach_header *header;
    struct load_command *loadCmd;
    struct segment_command *segCmd;
    Segment *segment = relocs;
    for (nImages = numImages,
	 image = images;
	 nImages;
	 nImages--, image++) {
	header = image->header;
	for (i = 0, loadCmd = (struct load_command *)(header + 1);
	     i < header->ncmds;
	     i++, ((void *)loadCmd) += loadCmd->cmdsize) {
	    if (loadCmd->cmd == LC_SEGMENT) {
		segCmd = (struct segment_command *)loadCmd;
		segment->address = segCmd->vmaddr;
		segment->size = segCmd->vmsize;
		segment->data = (vm_address_t)header + segCmd->fileoff;
		segment->maxAddress = segment->address + segment->size;
		segment->displacement = segment->data - segment->address;
		segment->maxData = segment->data + segment->size;
		segment->segName = segCmd->segname;
		segment->rFlags.readIn = YES;
		((void *)segment) += relocSize;
	    }
	}
    }
}

-(void)readInAllRelocs
{
    [self getImages];
    numRelocs = [self numSegments];
    relocs = calloc(numRelocs, relocSize);
    [self readInSegments];
    rmFlags.invalid = NO;
}

-(struct mach_header *)getMachHeader
{
    if (numImages)
	return images[0].header;
    else
	return NULL;
}

-(int)getNumMachHeaders
{
    return numImages;
}

-(struct mach_header **)getMachHeaders
{
    struct mach_header **headers;
    int headerIndex;
    headers = malloc((numImages + 1) * sizeof(*headers));
    for (headerIndex = 0; headerIndex < numImages; headerIndex++)
	headers[headerIndex] = images[headerIndex].header;
    headers[numImages] = NULL;
    return headers;
}

-(struct mach_header **)getMachHeadersWithNames: (char ***)names
{
    struct mach_header **headers;
    char **theNames;
    int headerIndex;
    headers = malloc((numImages + 1) * sizeof(*headers));
    *names = theNames = malloc((numImages + 1) * sizeof(*theNames));
    for (headerIndex = 0; headerIndex < numImages; headerIndex++) {
	headers[headerIndex] = images[headerIndex].header;
	theNames[headerIndex] = images[headerIndex].name;
    }
    headers[numImages] = NULL;
    theNames[numImages] = NULL;
    return headers;
}

-(struct load_command *)findLoadCommand: (unsigned long)command
                              forHeader: (struct mach_header *)header
{
    int i;
    struct load_command *loadCmd, *foundCmd = NULL;
    for (i = 0, loadCmd = (struct load_command *)(header + 1);
	 i < header->ncmds && !foundCmd;
	 i++, ((void *)loadCmd) += loadCmd->cmdsize) {
	if (loadCmd->cmd == command)
	    foundCmd = loadCmd;
    }
    return foundCmd;
}

-(int)numCommands: (unsigned long)command
        forHeader: (struct mach_header *)header
{
    int i, nCommands = 0;
    struct load_command *loadCmd;
    for (i = 0, loadCmd = (struct load_command *)(header + 1);
	 i < header->ncmds;
	 i++, ((void *)loadCmd) += loadCmd->cmdsize) {
	if (loadCmd->cmd == command)
	    nCommands++;
    }
    return nCommands;
}

-(struct symtab_command *)symCmd
{
    return (struct symtab_command *)
             [self findLoadCommand: LC_SYMTAB
	                 forHeader: images[0].header];
}

-(BOOL)isCore
{
    return images && images[0].header->filetype == MH_CORE;
}

-(BOOL)isShlib
{
    return images && images[0].header->filetype == MH_FVMLIB;
}

-(BOOL)isExecutable
{
    return images && ((images[0].header->filetype == MH_EXECUTE)
		      || (images[0].header->filetype == MH_OBJECT)
                      || (images[0].header->filetype == MH_PRELOAD));
}

@end