#include <ansidecl.h>
#include "sysdep.h"
#include "bfd.h"
#include "libbfd.h"

#include <mach_oh.h>
#include <stdio.h>
#include <stddef.h>

#include <mach-o/loader.h>
#include <sys/stat.h>

#if 0
void
mach_oh_not_implemented()
{
  bfd_perror("Function not implemented by Mach_Oh bfd.\n");
  _exit(1);
}
#else /* 0 */
#define mach_oh_not_implemented NULL
#endif /* 0 */

boolean
mach_oh_close_and_cleanup (abfd)
bfd *abfd;
{
  mach_oh_ex_data *ex_data = abfd->tdata;
  vm_deallocate(task_self(), ex_data->mapped_to, ex_data->size);
  free(ex_data);
  return true;
}

char *
mach_oh_core_file_failing_command (abfd)
bfd *abfd;
{
  return 0;
}

boolean
mach_oh_core_file_matches_executable_p (abfd)
bfd *abfd;
{
  return true;
}

int
mach_oh_get_symcount_upper_bound (abfd)
bfd *abfd;
{
  mach_oh_ex_data *ex_data = abfd->tdata;
  return ex_data->n_syms;
}

boolean
mach_oh_new_section_hook (abfd, newsect)
     bfd *abfd;
     asection *newsect;
{
    return true;
}

void
mach_oh_sect_2_section (sect, section)
struct section *sect;
asection *section;
{
  section->flags = SEC_LOAD;
  section->vma = sect->addr;
  section->size = sect->size;
  section->alignment_power = sect->align;
  section->filepos = sect->offset;
}

boolean
mach_oh_get_section_contents(abfd, section, location, offset, count)
bfd *abfd;
sec_ptr section;
PTR location;
file_ptr offset;
int count;
{
  void *data;
  mach_oh_ex_data *ex_data = abfd->tdata;
  
  if ((offset + count) <= section->size)
    {
       data = ex_data->mapped_to + section->filepos + offset;
       bcopy(data, location, count);
       return true;
    }
  else
    return false;
}
    


static void *mergeNames(sectionName, segName, sectName)
char *sectionName;
char *segName;
char *sectName;
{
  int i;
  for (i = 0; i < 16; i++) {
    if (*segName) 
      *(sectionName++) = *(segName++);
    else
      break;
  }
  for (i = 0; i < 16; i++) {
    if (*sectName) 
      *(sectionName++) = *(sectName++);
    else
      break;
  }
  *sectionName = '\0';
}

#define JMPOPCODE  0x4EF9
#define TRAPOPCODE 0x50FB
#define JMPTABLEOFFSET 0x54

void *
jmp_table_end(sect_addr, sect)
void *sect_addr;
struct section *sect;
{
  struct _JmpSlot {
    unsigned short opcode;
    unsigned int arg;
  } *jmp_slot;
  void *end_of_text;
  boolean found_end = false;
  
  for (jmp_slot = sect_addr + JMPTABLEOFFSET,
       end_of_text = sect_addr + sect->size;
       !found_end && (jmp_slot < (struct _JmpSlot *)end_of_text);
       jmp_slot++) {
    if ((jmp_slot->opcode != JMPOPCODE) && (jmp_slot->opcode != TRAPOPCODE))
      found_end = true;
  }
  return (void *)(sect->addr + ((void *)jmp_slot - sect_addr));
}
	

bfd_target *
mach_oh_map (abfd, header, size)
bfd *abfd;
struct mach_header *header;
unsigned size;
{
    mach_oh_ex_data *ex_data;
    struct load_command *loadCmd;
    struct symtab_command *symCmd;
    struct segment_command *segCmd;
    struct section *sect;
    boolean first_segment;
    asection *section;
    int i, nSects;
    char *sectionName;
    
    if (header->magic == MH_MAGIC) {
	abfd->tdata = ex_data = (mach_oh_ex_data *)zalloc(sizeof(*ex_data));
	ex_data->mapped_to = header;
	ex_data->size = size;
	switch (header->filetype) {
	    case MH_EXECUTE:
		abfd->flags = EXEC_P | D_PAGED;
		break;
	    case MH_PRELOAD:
		abfd->flags = EXEC_P;
		break;
	    case MH_FVMLIB:
		ex_data->flags.is_shlib = 1;
		break;
	    default:
		abfd->flags = NO_FLAGS;
	}
	first_segment = true;
	for (i = 0, loadCmd = (struct load_command *)(header + 1);
	    i < header->ncmds;
	    i++, ((void *)loadCmd) += loadCmd->cmdsize) {
	    switch (loadCmd->cmd) {
		case LC_SYMTAB:
		    symCmd = (struct symtab_command *)loadCmd;
		    abfd->flags |= HAS_SYMS;
		    ex_data->sym_filepos = symCmd->symoff;
		    ex_data->n_syms = symCmd->nsyms;
		    ex_data->str_filepos = symCmd->stroff;
		    ex_data->str_size = symCmd->strsize;
		    break;
		case LC_SEGMENT:
		    segCmd = (struct segment_command *)loadCmd;
		    for (nSects = segCmd->nsects, 
			 sect = (struct section *)(segCmd + 1);
			 nSects;
			 nSects--, sect++) {
			sectionName = bfd_alloc(abfd, 33);
			mergeNames(sectionName, sect->segname, sect->sectname);
			section = bfd_make_section (abfd, sectionName);
			mach_oh_sect_2_section (sect, section);
			if (first_segment
			    && (ex_data->flags.is_shlib)
			    && (nSects = segCmd->nsects)) {
			    ex_data->jmp_table_start = (void *)sect->addr;
			    ex_data->jmp_table_end
			    = jmp_table_end(((void *)header)
			                     + sect->offset, sect);
			    first_segment = false;
			}
		    }
		    break;
		case LC_UNIXTHREAD:
		    ex_data->unixThread = loadCmd;
		    break;
		default:;
	    }
	}
	
	ex_data->textsec = bfd_get_section_by_name (abfd, MACH_OH_TEXT_SECT);
	ex_data->datasec = bfd_get_section_by_name (abfd, MACH_OH_DATA_SECT);
	ex_data->bsssec = bfd_get_section_by_name (abfd, MACH_OH_BSS_SECT);
	
	return abfd->xvec;
    } else
      return NULL;
}

boolean
mach_oh_core_file_p (abfd)
bfd *abfd;
{
  return true;
}

bfd_target *
mach_oh_object_p (abfd)
bfd *abfd;
{
    struct stat stat;
    vm_offset_t start;
    int desc = fileno((FILE *)abfd->iostream);
    
    fstat(desc, &stat);
    if (!map_fd (desc, 0, &start, 1, stat.st_size))
	return mach_oh_map(abfd, (struct mach_header *)start, stat.st_size);
    else
	return NULL;
}

bfd_target mach_oh_vec = {
    "Mach_Oh",                             /* name */
    bfd_target_aout_flavour_enum,	   /* FIXME...not an aout */
    true,				   /* target byte order */
    true,				   /* target headers byte order */
    (HAS_RELOC | EXEC_P |                  /* object flags */
      HAS_LINENO | HAS_DEBUG |
      HAS_SYMS | HAS_LOCALS | DYNAMIC | WP_TEXT | D_PAGED),
    (SEC_ALLOC | SEC_LOAD | SEC_RELOC),    /* section flags */
    ' ',				   /* ar_pad_char */
    16,				           /* ar_max_namelen */
    _do_getblong, _do_putblong, _do_getbshort, _do_putbshort,  /* data */
    _do_getblong, _do_putblong, _do_getbshort, _do_putbshort,  /* hdrs */
    {_bfd_dummy_target,
     mach_oh_object_p,			   /* check format */
     mach_oh_not_implemented,
     mach_oh_not_implemented},
    {bfd_false, bfd_false, bfd_false, bfd_false},
    {bfd_false, bfd_false, bfd_false, bfd_false},
    mach_oh_core_file_failing_command,     /* _core_file_failing_command */
    mach_oh_not_implemented,               /* _core_file_failing_signal */
    mach_oh_core_file_matches_executable_p,/* _core_file_matches_ex...p */
    mach_oh_not_implemented,		   /* bfd_slurp_armap */
    mach_oh_not_implemented,	           /* bfd_slurp_extended_name_table */
    mach_oh_not_implemented,		   /* bfd_truncate_arname */
    mach_oh_not_implemented,		   /* write archive */
    mach_oh_close_and_cleanup,             /* _close_and_cleanup */
    mach_oh_not_implemented,               /* bfd_set_section_contents */
    mach_oh_get_section_contents,          /* bfd_get_section_contents */
    mach_oh_new_section_hook,              /* new_section_hook */
    mach_oh_not_implemented,               /* get_symtab_upper_bound */
    mach_oh_not_implemented,               /* canonicalize_symtab */
    mach_oh_not_implemented,               /* get_reloc_upper_bound */
    mach_oh_not_implemented,               /* bfd_canonicalize_reloc */
    mach_oh_not_implemented,               /* _bfd_make_empty_symbol */
    mach_oh_not_implemented,               /* _bfd_print_symbol */
    mach_oh_not_implemented,               /* _get_lineno */
    mach_oh_not_implemented,               /* _bfd_set_arch_mach */
    mach_oh_not_implemented,               /* openr_next_archived_file */
    mach_oh_not_implemented,               /* _bfd_find_nearest_line */
    mach_oh_not_implemented,               /* _bfd_stat_arch_elt */
    mach_oh_not_implemented,               /* _bfd_sizeof_headers */
};

bfd *
mach_oh_bfd_from_header(header, size)
struct mach_header *header;
unsigned size;
{
    bfd *abfd = new_bfd();
    abfd->xvec = &mach_oh_vec;
    abfd->iostream = NULL;
    abfd->format = bfd_object;
    abfd->direction = read_direction;
    abfd->filename = "MAPPED_FILE";
    abfd->mtime_set = 1;
    mach_oh_map(abfd, header, size);
    return abfd;
}