#import "gdb.h"
#include "objc.h"

#import <string.h>
#import <mach-o/ldsyms.h>
#import <stdio.h>
#import <mach-o/loader.h>
#import <streams/streams.h>
#import <mach_oh.h>

extern int baseFileLoaded;
#ifndef STATIC

asection *
findSegmentSection(bfd *abfd, char *segName, char *sectName)
{
  char sectionName[strlen(segName) + strlen(sectName) + 1];
  sprintf(sectionName, "%s%s", segName, sectName);
  return bfd_get_section_by_name(abfd, sectionName);
}

typedef struct _FileInfo {
    struct _FileInfo *next;
    char *name;
    vm_address_t address;
    size_t size;
    bfd *bfd;
} FileInfo;

#ifndef YES
#define YES 1
#endif YES

#ifndef NO
#define NO 0
#endif

int dbSupported = 1;
int stop_data_break_address; 
int filesLoaded = NO;
FileInfo *fileInfos;
    
static long db_length = 0;
static void *db_address = NULL;
const char *handlerFile = "/usr/lib/dbHandler.o";
char *handlerName;
static int exitHandlerLoaded = NO;
static int dbEnabled = NO;
static FileInfo *dbInfo;

extern RegionManager *regionManager;
extern bfd *exec_bfd;
extern task_t inferior_task;

unsigned char dbStub[] = {
    0x2f, 0x00, 0x2f, 0x08, 0x2f, 0x09, 0x2f, 0x0e,
    0x2f, 0x01, 0x4e, 0xb9, 0x00, 0x00, 0x3e, 0xaa,
    0x22, 0x1f, 0x2c, 0x5f, 0x22, 0x5f, 0x20, 0x5f,
    0x4a, 0x80, 0x67, 0xff, 0x00, 0x00, 0x00, 0x08,
    0x20, 0x1f, 0x4e, 0x4f, 0x20, 0x1f, 0x4e, 0x5e,
    0x4e, 0x75, 0x4e, 0x75
};

vm_address_t static inline vm_last_page(vm_address_t address)
//static inline vm_address_t vm_last_page(vm_address_t address)
{
    return address - (address % vm_page_size);
}

size_t static inline vm_next_page(size_t size)
//static inline size_t vm_next_page(size_t size)
{
    return size + (vm_page_size - (size % vm_page_size));
}

void replaceInText(char *text, int textSize, unsigned jumpToStub,
                   CORE_ADDR handler)
{
    const unsigned retCode = 0x4e5e4e75;
    char *gdbText, *endText, *gdbHandler;
    int changedText = NO;
    kern_return_t ret;
    if (gdbText = [regionManager pointerFor: text withSize: textSize]) {
	endText = gdbText + textSize - 3;
	if (((unsigned)text <= handler)
	                   && (handler < (unsigned)(text + textSize))) {
	    gdbHandler = gdbText + (handler - (unsigned)text);
	    for (; gdbText < gdbHandler; gdbText += 2) { 
		if (*(unsigned *)gdbText == retCode) {
		    *(unsigned *)gdbText = jumpToStub;
		    gdbText += 2;
		}
	    }
	    for (; *(unsigned *)gdbText != retCode; gdbText++);
	    gdbText += 2;
	}
	for (; gdbText < endText; gdbText += 2) {
	    if (*(unsigned *)gdbText == retCode) {
		*(unsigned *)gdbText = jumpToStub;
		gdbText += 2;
	    }
	}
	[regionManager writeDataAt: text for: textSize];
    }
}

short putInStub(CORE_ADDR handler)
{
    struct mach_header *header;
    char *endOfCmds;
    int found, index;
    struct load_command *loadCmd;
    struct segment_command *segCmd;
    struct section *firstSect;
    void *stub = NULL;
    kern_return_t ret;
    mach_oh_ex_data *ex_data = exec_bfd->tdata;
    header = [regionManager getMachHeader];
    endOfCmds = (char *)header + header->sizeofcmds;
    for (found = NO, loadCmd = (struct load_command *)(header + 1);
         !found && ((char *)loadCmd < endOfCmds);
         ((char *)loadCmd) += loadCmd->cmdsize) {
	if (loadCmd->cmd == LC_SEGMENT) {
	    segCmd = (struct segment_command *)loadCmd;
	    if (!segCmd->fileoff && segCmd->vmaddr)
		found = YES;
	}
    }
    if (found && segCmd->nsects) {
	firstSect = (struct section *)(segCmd + 1);
	if ((firstSect->offset - header->sizeofcmds) > sizeof(dbStub)) {
	    *(int *)(dbStub + 12) = handler;
	    stub = UNWARP(endOfCmds);
	    [regionManager putDataAt: stub for: sizeof(dbStub) from: dbStub];
	    stop_data_break_address = (int)(stub + 34);
	} else
	    error("Couldn't insert exit handler stub");
    }
    return (short)(int)stub;
}

void patchUpRefs(bfd *abfd)
{
    asection *refSect;
    char **refs, **ref;
    int sel;
    int numRefs;
    if (refSect = findSegmentSection(abfd, "__OBJC", "__message_refs")) {
	refs = [regionManager pointerFor: (void *)refSect->vma
	                        withSize: refSect->size];
	for (numRefs = refSect->size / sizeof(*refs), ref = refs;
	     numRefs;
	     numRefs--, ref++) {
	    if (sel = lookup_child_selector(WARP(*ref)))
		*ref = (char *)sel;
	}
	[regionManager writeDataAt: (void *)refSect->vma for: refSect->size];
    }
}     

void blastText(short stub, CORE_ADDR handler)
{
    bfd *eBfd;
    asection *textSect;
    unsigned jump = 0x4eb80000 | stub;
    for (eBfd = exec_bfd; eBfd; eBfd = eBfd->next) {
	if (textSect = findSegmentSection(eBfd, "__TEXT", "__text"))
	    replaceInText((char *)textSect->vma, 
	                  textSect->size,
			  jump,
			  handler);
    }
}

static vm_address_t rld_address;
static unsigned long rld_size;

unsigned long getInfMemory(unsigned long size)
{
    kern_return_t ret;
    rld_size = size;
    ret = vm_allocate(inferior_task, &rld_address, rld_size, YES);
    if (ret == KERN_SUCCESS) 
	return rld_address;
    else
	return 0;
}

int loadOneFile(FileInfo *fileInfo, NXStream *errorStream)
{
    char dbgName[200], *begin;
    struct mach_header *mach_header;
    char *objects[2] = {(char *)fileInfo->name, NULL};
    begin = rindex(fileInfo->name, '/');
    if (begin)
	begin++;
    else
	begin = fileInfo->name;
    sprintf(dbgName, "/tmp/%s", begin);
    if (!baseFileLoaded) {
	rld_load_basefile(NULL, bfd_get_filename(exec_bfd));
	baseFileLoaded = YES;
    }
    rld_address_func(getInfMemory);    
    if (rld_load(errorStream, &mach_header, objects, dbgName) == 1) {
	[regionManager invalidate];
	vm_write(inferior_task,
	         rld_address,
		 (pointer_t)mach_header,
		 vm_next_page(rld_size));
	fileInfo->address = rld_address;
	fileInfo->size = rld_size;		
//	vm_deallocate(task_self(), 
//	              (vm_address_t)mach_header,
//		      vm_next_page(rld_size));
	fileInfo->bfd = (bfd *)exec_file_add(dbgName, 0);
	symbol_file_add(dbgName, 0, 0, 0, fileInfo->bfd);
	patchUpRefs(fileInfo->bfd);
	unlink(dbgName);
	rld_unload(errorStream);
	return 0;
    } else
	return 1;
}

void removeFileInfo(FileInfo *fileInfo)
{
    FileInfo *fi;
    if (fileInfo == fileInfos)
	fileInfos = fileInfos->next;
    else {
	for (fi = fileInfos;
	     fi->next && (fi->next != fileInfo);
	     fi = fi->next);
	if (fi->next == fileInfo)
	    fi->next = fileInfo->next;
    }
}

void addFileInfo(FileInfo *fileInfo)
{
    FileInfo *fi;
    if (fileInfos) {
	for (fi = fileInfos; fi->next; fi = fi->next);
	fi->next = fileInfo;
    } else
	fileInfos = fileInfo;
}

FileInfo *loadFile(const char *fileName)
{
    NXStream *errorStream;
    FileInfo *fileInfo = xmalloc(sizeof(*fileInfo));
    bzero(fileInfo, sizeof(*fileInfo));
    fileInfo->name = savestring(fileName, strlen(fileName));
    addFileInfo(fileInfo);
    if (inferior_pid) {
	errorStream = NXOpenFile(fileno(stderr), NX_WRITEONLY);
	if (loadOneFile(fileInfo, errorStream) != 0) {
	    printf("Error loading %s\n", fileInfo->name);
	    removeFileInfo(fileInfo);
	    free(fileInfo->name);
	    free(fileInfo); fileInfo = NULL;
	}
	NXClose(errorStream);
    }
    return fileInfo;
}

static void loadFiles()
{
    NXStream *errorStream = NXOpenFile(fileno(stderr), NX_WRITEONLY);
    FileInfo *fileInfo, *fi;
    fileInfo = fileInfos;
    while (fileInfo) {
	if (loadOneFile(fileInfo, errorStream) == 0) 
	    fileInfo = fileInfo->next;
	else {
	    printf("Error loading %s\n", fileInfo->name);
	    fi = fileInfo->next;
	    removeFileInfo(fileInfo);
	    free(fileInfo->name);
	    free(fileInfo);
	    fileInfo = fi;
	}
    }
    NXClose(errorStream);
}

static void loadExitHandler()
{
    short stub;
    int handlerIndex;
    CORE_ADDR handler;
    if (!exitHandlerLoaded) {
	handlerIndex = lookup_misc_func(handlerName);
	if (handlerIndex != -1) {
	    handler = misc_function_vector[handlerIndex].address;
	    if (stub = putInStub(handler))
		blastText(stub, handler);
	} else
	    error("No handler named %s.\n", handlerName);
	exitHandlerLoaded = YES;
    }
}

static void set_exit_handler_command(char *args)
{
    if (handlerName)
	free(handlerName);
    dbEnabled = NO;
    if (args && *args) {
	handlerName = savestring(args, strlen(args));
	printf("Exit handler set to %s\n", args);
    } else {
	handlerName = NULL;
	printf("No exit handler\n");
    }
}

static void doDb()
{
    char *data;
    int dbBufferI, dbBuffer,
        dbAddressI, dbAddress,
	dbLengthI, dbLength;
    dbBufferI = lookup_misc_func("db_buffer");
    dbAddressI = lookup_misc_func("db_address");
    dbLengthI = lookup_misc_func("db_length");
    if ((dbBufferI >= 0) && (dbAddressI >= 0) && (dbLengthI >= 0)){
	dbBuffer = misc_function_vector[dbBufferI].address;
	dbAddress = misc_function_vector[dbAddressI].address;
	dbLength = misc_function_vector[dbLengthI].address;
	data = [regionManager pointerFor: db_address withSize: db_length];
	[regionManager putDataAt: (void *)dbBuffer  for: db_length from: data];
	[regionManager putDataAt: (void *)dbAddress for: 4 from: &db_address];
	[regionManager putDataAt: (void *)dbLength  for: 4 from: &db_length];
    } else
	error("Couldn't find db symbols");
}

void db_clear()
{
    FileInfo *fileInfo;
    for (fileInfo = fileInfos; fileInfo; fileInfo = fileInfo->next) {
	removeSymbols(fileInfo->address, fileInfo->size);
	removeObjC(fileInfo->bfd);
	removeBfd(fileInfo->bfd);
    }
    filesLoaded = NO;
    exitHandlerLoaded = NO;
}

void db_run()
{
    if (!filesLoaded) {
	loadFiles();
	filesLoaded = YES;
    }
    if (!exitHandlerLoaded) {
	if (handlerName)
	    loadExitHandler();
	if (dbEnabled)
	    doDb();
    }
}

void db_stop()
{
    int stack;
    stack = read_register(SP_REGNUM);
    stop_pc = read_memory_integer(stack, 4) - 4;
    write_register(PC_REGNUM, stop_pc);
    write_register(SP_REGNUM, stack + 4);
    select_frame (get_current_frame (), 0);
    printf("Data break at: %p.\n", stop_pc);
}

static void load_file_command(char *args)
{
    if (args)
	loadFile(args);
    else
	error("You must specify a file to be loaded");
}

static void db_command(char *args)
{
    char *address, *bytes;
    value val;
    void *newAddress;
    int newLength;
    if (address = args) {
	if (bytes = index(address, ' ')) {
	    *bytes = '\0';
	    bytes++;
	    val = evaluate_expression(parse_c_expression(address));
	    newAddress = (void *)value_as_long(val);
	    val = evaluate_expression(parse_c_expression(bytes));
	    newLength = value_as_long(val);
	} else {
	    val = evaluate_expression(parse_c_expression(address));
	    newAddress = (void *)(VALUE_ADDRESS(val) + VALUE_OFFSET(val));
	    if (newAddress)
		newLength = TYPE_LENGTH(VALUE_TYPE(val));
	    else {
		newAddress = (void *)value_as_long(val);
		newLength = 1;
	    }
	}
	if (db_length > 1024) {
	    error("You tried to data break at %p for %d bytes.\n"
	          "There is a limit of 1024 bytes.",
		  newAddress,
		  newLength);
	}
	dbEnabled = YES; 
	db_address = newAddress;
	db_length = newLength;
	if (handlerName)
	    free(handlerName);
	handlerName = savestring("dbHandler", 10);
	printf("Data break inserted at %p, for %d bytes.\n",
	       db_address, db_length);
    } else {
	dbEnabled = NO;
	printf("No data break\n");
	db_address = 0;
	db_length = 0;
    }
    if (!dbInfo)
	dbInfo = loadFile(handlerFile);
    if (inferior_pid) {
	if (!exitHandlerLoaded)
	    loadExitHandler();
	doDb();
    } 
}

#else STATIC
int dbSupported = 0;
int stop_data_break_address; 
void db_run() {}
void db_clear() {}
void db_stop() {}

static void db_command(char *args)
{
    error("The static version of gdb does not support data break points.");
}

static void load_file_command(char *args)
{
    error("The static version of gdb does not support dynamic loading.\n");
}

#endif STATIC

void _initialize_db()
{
    add_com("data-break", class_breakpoint, db_command,
            "break if data specified changes\n");
    add_com("load-file", class_files, load_file_command,
            "dynamically load file into debugged program\n");
    add_com("set-exit-handler", class_breakpoint, set_exit_handler_command,
            "Sets function to be called at exit of every other function\n");
    add_com_alias("db", "data-break", class_breakpoint, 1);
}	