
#include <mach/mach.h>
#include <servers/bootstrap.h>
#include <servers/netname.h>


/* Don't really need to be globals */
port_t          my_bootstrap_port;
port_t          my_nm_port;
port_t          micro_nm_port;
port_t          micro_nm_hostport;
port_t          micro_nm_hostprivport;
port_t          initialtask;

int gettaskverbose =0;

gettask(machinename, newtask)
char * machinename;
port_t *newtask;
{
	kern_return_t   result;
	task_array_t    task_list;
	unsigned int    task_count;
	processor_set_name_t default_set;
	processor_set_t default_set_priv;
	int i;

	/* Get bootstrap port */
	result = task_get_bootstrap_port(task_self(), &my_bootstrap_port);
	if (result != KERN_SUCCESS) {
 	   mach_error("Couldn't get bootstrap port", result);
	   return KERN_FAILURE;
	}


	/* Look up local nmserver */
	result=bootstrap_look_up(my_bootstrap_port, "NetMessage", &my_nm_port);
	if (result!=BOOTSTRAP_SUCCESS){
    		mach_error("lookup failed ", result);
	        return KERN_FAILURE;
	} else {

		/* Look up remote nmserver */
		result = netname_look_up(my_nm_port, machinename, "NetMessage",
			 &micro_nm_port);
		if (result!=NETNAME_SUCCESS){
			mach_error("netname_look_up nmserver failed ",result);
	        	return KERN_FAILURE;
		}
		if (gettaskverbose) 
			printf("Got nmserver port for %s\n", machinename);


		/* Get the remote host and host priv ports */
		result = netname_look_up(micro_nm_port, "", "HOST",
			 &micro_nm_hostport);
		if (result!=NETNAME_SUCCESS){
			mach_error("netname_look_up host failed ",result);
	        	return KERN_FAILURE;
		}
		if (gettaskverbose) 
			printf("Got host port for %s\n", machinename);

		result = netname_look_up(micro_nm_port, "", "HOST_PRIV",
			 &micro_nm_hostprivport);
		if (result!=NETNAME_SUCCESS){
			mach_error("netname_look_up hostpriv failed ",result);
		        return KERN_FAILURE;
		}
		if (gettaskverbose) 
			printf("Got hostpriv port for %s\n", machinename);



		/* Find us a task */
		result = processor_set_default(micro_nm_hostport, &default_set);
		if (result !=KERN_SUCCESS) {
    			mach_error("Error calling processor_set_default()", 
			result);
		        return KERN_FAILURE;
		}

		result = host_processor_set_priv(micro_nm_hostprivport,
			 default_set, &default_set_priv);
		if (result != KERN_SUCCESS) {
    			mach_error("Call to host_processor_set_priv() failed",
			 result);
	        	return KERN_FAILURE;
		}

		result = processor_set_tasks(default_set_priv, &task_list, 
			&task_count);

		if (result != KERN_SUCCESS) {
    			mach_error("Call to processor_set_tasks() failed",
			 result); 
		        return KERN_FAILURE;
		}
		if (gettaskverbose) {
			printf("Task count 0x%x\n", task_count);
			for (i = 0; i < task_count; i++){
				printf("Task array 0x%x\n", task_list[i]);
			}
		}


		/* Tuck away a task to use, then free the resources */
		*newtask = task_list[task_count-1];

		result = vm_deallocate(task_self(), (vm_address_t)task_list, 
    			sizeof(task_list)*task_count);
		if (result != KERN_SUCCESS)
    			mach_error("Trouble freeing task_list", result);


	}

}
