#include "defs.h"
#include "param.h"
#include "inferior.h"
#include "target.h"
#include <objc/objc.h>
#include <sys/signal.h>

extern struct symbol *step_start_function;
extern int stop_soon_quietly;
extern int stop_print_frame;

void
kernel_attach_command(char *args, int from_tty)
{
    extern struct target_ops child_ops;
    char portname[50];
    task_t kernel_task;
    int kdbPid;
    int tries;
    BOOL found;
    
    dont_repeat();
    
    if (!args)
	error_no_arg ("hostname of kernel to attach to");
    
    if (inferior_pid) {
	if (query ("A program is being debugged already.  Kill it? "))
	    target_kill ((char *)0, from_tty);
	else
	    error ("Inferior not killed.");
    }
    
    if (from_tty)
	printf("Attaching program: %s to kernel on %s.\n",
	       get_exec_file(1), args);

    sprintf (portname, "gdb_%s", args);
    if (netname_look_up(name_server_port, "", portname, &kernel_task) != 0) {
	kdbPid = vfork();
	if (kdbPid) {
	    for (tries = 0, found = NO; (tries < 20) && !found; tries++) {
		usleep(500000);
		found = (netname_look_up(name_server_port,
		                         "", portname, &kernel_task) == 0);
	    }
	    if (!found) {
		kill(kdbPid, SIGKILL);
		error("Couldn't find kernel rdebug port, %s", portname);
	    }
	} else
	    execl("/usr/lib/kdb", "kdb", args, NULL);
    }

    push_target (&child_ops);
    
    kernel_attached = YES;
    
    create_inferior_for_task(kernel_task, YES);
    
    mark_breakpoints_out ();
    target_terminal_init ();
    clear_proceed_status ();

    task_suspend(kernel_task);
    stop_pc = read_pc();
    step_start_function = find_pc_function(stop_pc);
    inferior_pid = -1;
    stop_soon_quietly = 1;
    stop_print_frame = 1;
    attach_flag = 1;
    normal_stop ();
}

void
kernel_reboot_command(char *args)
{
    if (kernel_attached) {
	if (!args)
	    args = "";
	mach_call(task_reboot(inferior_task, args, strlen(args)),
	          "task_reboot", "kernel_reboot_command");
	target_mourn_inferior();
    } else
	error("Not attached to kernel.");
}

void
kselect_frame_command (args)
char *args;
{
  value val;
  int fp,pc;

  if (!args)
    error ("select-frame fp, pc.");
  val = parse_to_comma_and_eval (&args);
  if (!val) 
    error ("select-frame fp, pc.");
  fp = value_as_long(val);
  if (*args++ != ',')
     error ("select-frame fp, pc.");
  val = parse_and_eval (args);
  if (!val) 
    error ("select-frame fp, pc.");
  pc = value_as_long(val);
  flush_cached_frames ();
  set_current_frame (create_new_frame (fp, pc));
  select_frame (get_current_frame (), 0);
}

void
_initialize_kernel()
{
    add_com("kattach", class_run, kernel_attach_command,
	    "Attach to a kernel on HOSTNAME.");
    add_com ("select-frame", class_support, kselect_frame_command,
             "select frame at fp, pc.");
    add_com("kreboot", class_run, kernel_reboot_command,
            "Reboot an attached kernel with ARGS");
}
