/* Parameters for execution on a Sun, for GDB, the GNU debugger.
   Copyright (C) 1986, 1987, 1989 Free Software Foundation, Inc.

This file is part of GDB.

GDB is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GDB is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GDB; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#define HOST_BYTE_ORDER BIG_ENDIAN

/* Do implement the attach and detach commands.  */

#define ATTACH_DETACH

/* Override copies of {fetch,store}_inferior_registers in infptrace.c.  */

#define FETCH_INFERIOR_REGISTERS

/* We have to grab the regs since we store all regs at once.  */

#define CHILD_PREPARE_TO_STORE() \
	read_register_bytes (0, (char *)NULL, REGISTER_BYTES)

/* Interface definitions for kernel debugger KDB.  */

/* Map machine fault codes into signal numbers.
   First subtract 0, divide by 4, then index in a table.
   Faults for which the entry in this table is 0
   are not handled by KDB; the program's own trap handler
   gets to handle then.  */

#define FAULT_CODE_ORIGIN 0
#define FAULT_CODE_UNITS 4
#define FAULT_TABLE    \
{ 0, 0, 0, 0, SIGTRAP, 0, 0, 0, \
  0, SIGTRAP, 0, 0, 0, 0, 0, SIGKILL, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  SIGILL }

/* Start running with a stack stretching from BEG to END.
   BEG and END should be symbols meaningful to the assembler.
   This is used only for kdb.  */

#define INIT_STACK(beg, end)  \
{ asm (".globl end");         \
  asm ("movel #end, sp");      \
  asm ("movel #0,a6"); }

/* Push the frame pointer register on the stack.  */
#define PUSH_FRAME_PTR        \
  asm ("movel a6,sp@-");

/* Copy the top-of-stack to the frame pointer register.  */
#define POP_FRAME_PTR  \
  asm ("movl sp@,a6");

/* After KDB is entered by a fault, push all registers
   that GDB thinks about (all NUM_REGS of them),
   so that they appear in order of ascending GDB register number.
   The fault code will be on the stack beyond the last register.  */

#define PUSH_REGISTERS        \
{ asm ("clrw -(sp)");	      \
  asm ("pea sp@(10)");	      \
  asm ("movem #0xfffe,sp@-"); }

/* Assuming the registers (including processor status) have been
   pushed on the stack in order of ascending GDB register number,
   restore them and return to the address in the saved PC register.  */

#define POP_REGISTERS          \
{ asm ("subil #8,sp@(28)");     \
  asm ("movem sp@,#0xffff"); \
  asm ("rte"); }
  
struct mach_oh_nlist {
	union {
		char *n_name;	/* for use when in-core */
		long  n_strx;	/* index into the string table */
	} n_un;
	unsigned char n_type;	/* type flag, see below */
	unsigned char n_sect;	/* section number or NO_SECT */
	short	      n_desc;	/* see <stab.h> */
	unsigned long n_value;	/* value of this symbol (or stab offset) */
};

#define	MACH_OH_N_SECT	0xe		/* defined in section number n_sect */
#define	MACH_OH_N_TYPE	0x1e            /* mask for the type bits */
#define BSD_N_TEXT 4
#define BSD_N_DATA 6
#define BSD_N_BSS 8

void static inline modify_symbol(struct mach_oh_nlist *bufp)
{
    if ((bufp->n_type & MACH_OH_N_TYPE) == MACH_OH_N_SECT) { \
	bufp->n_type &= ~MACH_OH_N_TYPE; \
	switch (bufp->n_sect) { \
	    case 1: /* TEXT */ \
		bufp->n_type |= BSD_N_TEXT; \
		break; \
	    case 3: /* BSS */ \
		bufp->n_type |= BSD_N_BSS; \
		break; \
	    case 2: /* DATA */ \
	    default: /* ??? Anything else defaults to data */ \
		bufp->n_type |= BSD_N_DATA; \
		break; \
	} \
    }
}

#define MODIFY_SYMBOL(bufp) modify_symbol((struct mach_oh_nlist *)bufp)

#define TEXT_SECT_NAME "__TEXT__text"

#define VALIDATE_MISC_FUNCTION(address) NeXT_validate_misc_function(address)

#define COMPLETION_LIST_HOOK(text) objc_completion_list_hook(text)

#define DECODE_LINE_1_HOOK(args, value) lookup_method(args, value)

#define CONDENSE_MISC_BUNCHES_HOOK() combine_misc_functions_psymtabs()

#define FIND_PC_SYMTAB find_pc_symtab

extern int pc_off_limits (CORE_ADDR stop_pc, char *name, CORE_ADDR *new_pc);
extern struct type *typeForObject(CORE_ADDR object);
struct symtab *symtab_for_misc_function(int ind);

/* We relocated the blocks */
#define BLOCK_ADDRESS_ABSOLUTE

/* If a program is launched by Workspace, it has no controlling tty */
#define NO_CONTROLLING_TTY

#define MAKE_BLOCK_VECTOR_HOOK(bv) make_block_vector_hook(bv)

#define LINETABLE_HOOK(lt) linetable_hook(lt)
#define CREATE_INFERIOR_HOOK create_inferior_hook
#define STOP_HOOK stop_hook
#define WAIT_FOR_INFERIOR_HOOK wait_for_inferior_hook

static inline char *
STRCHR (const char *s, int c)
{
    while (*s) {
	if (*s == c)
	    return (char *)s;
	s++;
    }
    return NULL;
}

#define strchr(s, c) STRCHR(s, c)

/*
 * Find the first colon after the name of a function.  This name could be an
   Objective C method name, which could have colons in it itself.
 */
static inline char *
colon_from_name (name)
char *name;
{
  if (*name == '-' || *name == '+')
    return (char *) strchr (strchr (name, ']'), ':');
  else
    return (char *) strchr (name, ':');
}

#define COLON_FROM_NAME(name) colon_from_name(name)

#define RECORD_MISC_FUNCTION_PSYMTAB record_misc_function_psymtab

#define LOCK_EXECUTE_COMMAND

extern int kernel_attached;

#define ADDITIONAL_OPTIONS \
        {"opath", 1, 0, 'o'}, \
	{"viewhost", 1, 0, 12}, \
	{"viewprogram", 1, 0, 13},
	
#define ADDITIONAL_OPTION_CASES \
          case 'o': \
		opath_alias_command (optarg); \
		break; \
	  case 12: \
	        set_view_host (optarg); \
		break; \
	  case 13: \
	  	set_view_program (optarg); \
		break;
		
extern task_t inferior_task, gdb_task;
extern CORE_ADDR changedAddress;
#define SUN_FIXED_LBRAC_BUG

char *stringSave(char *string);
char *stringSave2(char *s1, char *s2);
char *stringSaveN(char *string, int n);
int stringEq(char *s1, char *s2);

static inline int
stringCmp(char *s1, char *s2)
{
    short diff = (*(short *)s1 - *(short *)s2);
    if (diff)
	return (int)diff;
    else
	return strcmp(s1 + 2, s2 + 2);
}

extern int isTask(task_t task);

int static inline mach_call_no_error(kern_return_t ret, char *mstring, char *estring)
{ 
    if (ret != KERN_SUCCESS) {
	mach_error(mstring, ret);
	fprintf(stderr, "%s\n", estring);
    }
    return ret;
}

void static inline mach_call_no_inferior(kern_return_t ret, char *mstring, char *estring)
{ 
    if (ret != KERN_SUCCESS) {
	mach_error(mstring, ret);
	error(stderr, "%s\n", estring);
    }
}

extern void NeXT_mourn_inferior();

void static inline mach_call(kern_return_t ret, char *mstring, char *estring)
{ 
    if (ret != KERN_SUCCESS) {
	if (isTask(inferior_task)) {
	    mach_error(mstring, ret);
	    error(estring);
 	} else {
	    mach_error(mstring, ret);
	    error(estring);
	    NeXT_mourn_inferior();
	    error("The program you were running has been terminated.");
	}
   }
}