
# line 30 "../expread.y"
#include <stdio.h>
#include "defs.h"
#include "param.h"
#include "symtab.h"
#include "frame.h"
#include "expression.h"
#include "value.h"
#include "command.h"
#include "objc.h"
#ifdef SIGNED
#undef SIGNED
#endif /* SIGNED */
#ifdef AND
#undef AND
#endif /* AND */

static struct expression *expout;
static int expout_size;
static int expout_ptr;

static int yylex ();
static void yyerror ();
static void write_exp_elt ();
static void write_exp_elt_opcode ();
static void write_exp_elt_sym ();
static void write_exp_elt_longcst ();
static void write_exp_elt_dblcst ();
static void write_exp_elt_type ();
static void write_exp_elt_intern ();
static void write_exp_string ();
static void start_arglist ();
static int end_arglist ();
static void free_funcalls ();
static char *copy_name ();
static int parse_number ();

/* If this is nonzero, this block is used as the lexical context
   for symbol names.  */

static struct block *expression_context_block;

/* The innermost context required by the stack and register variables
   we've encountered so far. */
struct block *innermost_block;

/* The block in which the most recently discovered symbol was found. */
struct block *block_found;

/* Number of arguments seen so far in innermost function call.  */
static int arglist_len;

/* Data structure for saving values of arglist_len
   for function calls whose arguments contain other function calls.  */

struct funcall
  {
    struct funcall *next;
    int arglist_len;
  };

struct funcall *funcall_chain;

/* This kind of datum is used to represent the name
   of a symbol token.  */

struct stoken
  {
    char *ptr;
    int length;
  };

struct ttype
  {
    struct stoken stoken;
    struct type *type;
  };

struct symtoken
  {
    struct stoken stoken;
    struct symbol *sym;
    int is_a_field_of_this;
  };

/* Data structure for saving the parts of an objective C selector. */

struct sclass
  {
    Class class;
    struct type *type;
  };

struct selname
  {
    struct selname *next;
    char *msglist_sel;
    int msglist_len;
  };

static int msglist_len;
static struct selname *selname_chain;
static char *msglist_sel;
static void start_msglist();
static void add_msglist(struct stoken *str, int addcolon);
static int end_msglist();

/* For parsing of complicated types.
   An array should be preceded in the list by the size of the array.  */
enum type_pieces
  {tp_end = -1, tp_pointer, tp_reference, tp_array, tp_function};
/* The stack can contain either an enum type_pieces or an int.  */
union type_stack_elt {
  enum type_pieces piece;
  int int_val;
};
static union type_stack_elt *type_stack;
static int type_stack_depth, type_stack_size;

static void push_type ();
static void push_type_int ();
static enum type_pieces pop_type ();
static int pop_type_int ();

/* Allow debugging of parsing.  */
#define YYDEBUG 1

# line 161 "../expread.y"
typedef union 
  {
    LONGEST lval;
    unsigned LONGEST ulval;
    double dval;
    struct symbol *sym;
    struct type *tval;
    struct stoken sval;
    struct ttype tsym;
    struct symtoken ssym;
    int voidval;
    struct block *bval;
    enum exp_opcode opcode;
    struct internalvar *ivar;

    struct type **tvec;
    int *ivec;
    struct sclass sclass;
  } YYSTYPE;
# define INT 257
# define CHAR 258
# define UINT 259
# define FLOAT 260
# define STRING 261
# define NAME 262
# define BLOCKNAME 263
# define TYPENAME 264
# define CLASSNAME 265
# define NAME_OR_INT 266
# define NAME_OR_UINT 267
# define STRUCT 268
# define UNION 269
# define ENUM 270
# define SIZEOF 271
# define UNSIGNED 272
# define COLONCOLON 273
# define ERROR 274
# define SIGNED 275
# define LONG 276
# define SHORT 277
# define INT_KEYWORD 278
# define LAST 279
# define REGNAME 280
# define VARIABLE 281
# define ASSIGN_MODIFY 282
# define THIS 283
# define ABOVE_COMMA 284
# define OR 285
# define AND 286
# define EQUAL 287
# define NOTEQUAL 288
# define LEQ 289
# define GEQ 290
# define LSH 291
# define RSH 292
# define UNARY 293
# define INCREMENT 294
# define DECREMENT 295
# define ARROW 296
#define yyclearin yychar = -1
#define yyerrok yyerrflag = 0
extern int yychar;
extern short yyerrflag;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif
YYSTYPE yylval, yyval;
# define YYERRCODE 256

# line 1000 "../expread.y"


/* Begin counting arguments for a function call,
   saving the data about any containing call.  */

static void
start_arglist ()
{
  register struct funcall *new = (struct funcall *) xmalloc (sizeof (struct funcall));

  new->next = funcall_chain;
  new->arglist_len = arglist_len;
  arglist_len = 0;
  funcall_chain = new;
}

/* Return the number of arguments in a function call just terminated,
   and restore the data for the containing function call.  */

static int
end_arglist ()
{
  register int val = arglist_len;
  register struct funcall *call = funcall_chain;
  funcall_chain = call->next;
  arglist_len = call->arglist_len;
  free (call);
  return val;
}

/* Free everything in the funcall chain.
   Used when there is an error inside parsing.  */

static void
free_funcalls ()
{
  register struct funcall *call, *next;

  for (call = funcall_chain; call; call = next)
    {
      next = call->next;
      free (call);
    }
}

/* This page contains the functions for adding data to the  struct expression
   being constructed.  */

/* Add one element to the end of the expression.  */

/* To avoid a bug in the Sun 4 compiler, we pass things that can fit into
   a register through here */

static void
write_exp_elt (expelt)
     union exp_element expelt;
{
  if (expout_ptr >= expout_size)
    {
      expout_size *= 2;
      expout = (struct expression *) xrealloc (expout,
					       sizeof (struct expression)
					       + expout_size * sizeof (union exp_element));
    }
  expout->elts[expout_ptr++] = expelt;
}

static void
write_exp_elt_opcode (expelt)
     enum exp_opcode expelt;
{
  union exp_element tmp;

  tmp.opcode = expelt;

  write_exp_elt (tmp);
}

static void
write_exp_elt_sym (expelt)
     struct symbol *expelt;
{
  union exp_element tmp;

  tmp.symbol = expelt;

  write_exp_elt (tmp);
}

static void
write_exp_elt_longcst (expelt)
     LONGEST expelt;
{
  union exp_element tmp;

  tmp.longconst = expelt;

  write_exp_elt (tmp);
}

static void
write_exp_elt_dblcst (expelt)
     double expelt;
{
  union exp_element tmp;

  tmp.doubleconst = expelt;

  write_exp_elt (tmp);
}

static void
write_exp_elt_type (expelt)
     struct type *expelt;
{
  union exp_element tmp;

  tmp.type = expelt;

  write_exp_elt (tmp);
}

static void
write_exp_elt_intern (expelt)
     struct internalvar *expelt;
{
  union exp_element tmp;

  tmp.internalvar = expelt;

  write_exp_elt (tmp);
}

/* Add a string constant to the end of the expression.
   Follow it by its length in bytes, as a separate exp_element.  */

static void
write_exp_string (str)
     struct stoken str;
{
  register int len = str.length;
  register int lenelt
    = (len + sizeof (union exp_element)) / sizeof (union exp_element);

  expout_ptr += lenelt;

  if (expout_ptr >= expout_size)
    {
      expout_size = max (expout_size * 2, expout_ptr + 10);
      expout = (struct expression *)
	xrealloc (expout, (sizeof (struct expression)
			   + (expout_size * sizeof (union exp_element))));
    }
  bcopy (str.ptr, (char *) &expout->elts[expout_ptr - lenelt], len);
  ((char *) &expout->elts[expout_ptr - lenelt])[len] = 0;
  write_exp_elt_longcst ((LONGEST) len);
}

/* During parsing of a C expression, the pointer to the next character
   is in this variable.  */

static char *lexptr;

/* Tokens that refer to names do so with explicit pointer and length,
   so they can share the storage that lexptr is parsing.

   When it is necessary to pass a name to a function that expects
   a null-terminated string, the substring is copied out
   into a block of storage that namecopy points to.

   namecopy is allocated once, guaranteed big enough, for each parsing.  */

static char *namecopy;

/* Current depth in parentheses within the expression.  */

static int paren_depth;

/* Nonzero means stop parsing on first comma (if not within parentheses).  */

static int comma_terminates;

/* Take care of parsing a number (anything that starts with a digit).
   Set yylval and return the token type; update lexptr.
   LEN is the number of characters in it.  */

/*** Needs some error checking for the float case ***/

static int
parse_number (p, len, parsed_float, putithere)
     register char *p;
     register int len;
     int parsed_float;
     YYSTYPE *putithere;
{
  register LONGEST n = 0;
  register int i;
  register int c;
  register int base = input_radix;
  int unsigned_p = 0;

  extern double atof ();

  if (parsed_float)
    {
      /* It's a float since it contains a point or an exponent.  */
      putithere->dval = atof (p);
      return FLOAT;
    }

  /* Handle base-switching prefixes 0x, 0t, 0d, 0 */
  if (p[0] == '0')
    switch (p[1])
      {
      case 'x':
      case 'X':
	if (len >= 3)
	  {
	    p += 2;
	    base = 16;
	    len -= 2;
	  }
	break;

      case 't':
      case 'T':
      case 'd':
      case 'D':
	if (len >= 3)
	  {
	    p += 2;
	    base = 10;
	    len -= 2;
	  }
	break;

      default:
	base = 8;
	break;
      }

  while (len-- > 0)
    {
      c = *p++;
      if (c >= 'A' && c <= 'Z')
	c += 'a' - 'A';
      if (c != 'l' && c != 'u')
	n *= base;
      if (c >= '0' && c <= '9')
	n += i = c - '0';
      else
	{
	  if (base > 10 && c >= 'a' && c <= 'f')
	    n += i = c - 'a' + 10;
	  else if (len == 0 && c == 'l')
	    ;
	  else if (len == 0 && c == 'u')
	    unsigned_p = 1;
	  else
	    return ERROR;	/* Char not a digit */
	}
      if (i >= base)
	return ERROR;		/* Invalid digit in this base */
    }

  if (unsigned_p)
    {
      putithere->ulval = n;
      return UINT;
    }
  else
    {
      putithere->lval = n;
      return INT;
    }
}

struct token
{
  char *operator;
  int token;
  enum exp_opcode opcode;
};

const static struct token tokentab3[] =
  {
    {">>=", ASSIGN_MODIFY, BINOP_RSH},
    {"<<=", ASSIGN_MODIFY, BINOP_LSH}
  };

const static struct token tokentab2[] =
  {
    {"+=", ASSIGN_MODIFY, BINOP_ADD},
    {"-=", ASSIGN_MODIFY, BINOP_SUB},
    {"*=", ASSIGN_MODIFY, BINOP_MUL},
    {"/=", ASSIGN_MODIFY, BINOP_DIV},
    {"%=", ASSIGN_MODIFY, BINOP_REM},
    {"|=", ASSIGN_MODIFY, BINOP_LOGIOR},
    {"&=", ASSIGN_MODIFY, BINOP_LOGAND},
    {"^=", ASSIGN_MODIFY, BINOP_LOGXOR},
    {"++", INCREMENT, BINOP_END},
    {"--", DECREMENT, BINOP_END},
    {"->", ARROW, BINOP_END},
    {"&&", AND, BINOP_END},
    {"||", OR, BINOP_END},
    {"::", COLONCOLON, BINOP_END},
    {"<<", LSH, BINOP_END},
    {">>", RSH, BINOP_END},
    {"==", EQUAL, BINOP_END},
    {"!=", NOTEQUAL, BINOP_END},
    {"<=", LEQ, BINOP_END},
    {">=", GEQ, BINOP_END}
  };

/* assign machine-independent names to certain registers 
 * (unless overridden by the REGISTER_NAMES table)
 */
struct std_regs {
	char *name;
	int regnum;
} std_regs[] = {
#ifdef PC_REGNUM
	{ "pc", PC_REGNUM },
#endif
#ifdef FP_REGNUM
	{ "fp", FP_REGNUM },
#endif
#ifdef SP_REGNUM
	{ "sp", SP_REGNUM },
#endif
#ifdef PS_REGNUM
	{ "ps", PS_REGNUM },
#endif
};

#define NUM_STD_REGS (sizeof std_regs / sizeof std_regs[0])

/* Read one token, getting characters through lexptr.  */

static int
yylex ()
{
  register int c;
  register int namelen;
  register unsigned i;
  register char *tokstart;

 retry:

  tokstart = lexptr;
  /* See if it is a special token of length 3.  */
  for (i = 0; i < sizeof tokentab3 / sizeof tokentab3[0]; i++)
    if (!strncmp (tokstart, tokentab3[i].operator, 3))
      {
	lexptr += 3;
	yylval.opcode = tokentab3[i].opcode;
	return tokentab3[i].token;
      }

  /* See if it is a special token of length 2.  */
  for (i = 0; i < sizeof tokentab2 / sizeof tokentab2[0]; i++)
    if (!strncmp (tokstart, tokentab2[i].operator, 2))
      {
	lexptr += 2;
	yylval.opcode = tokentab2[i].opcode;
	return tokentab2[i].token;
      }

  switch (c = *tokstart)
    {
    case 0:
      return 0;

    case ' ':
    case '\t':
    case '\n':
      lexptr++;
      goto retry;

    case '\'':
      lexptr++;
      c = *lexptr++;
      if (c == '\\')
	c = parse_escape (&lexptr);
      yylval.lval = c;
      c = *lexptr++;
      if (c != '\'')
	error ("Invalid character constant.");
      return CHAR;

    case '(':
      paren_depth++;
      lexptr++;
      return c;

    case ')':
      if (paren_depth == 0)
	return 0;
      paren_depth--;
      lexptr++;
      return c;

    case ',':
      if (comma_terminates && paren_depth == 0)
	return 0;
      lexptr++;
      return c;

    case '.':
      /* Might be a floating point number.  */
      if (lexptr[1] < '0' || lexptr[1] > '9')
	goto symbol;		/* Nope, must be a symbol. */
      /* FALL THRU into number case.  */

    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      {
	/* It's a number.  */
	int got_dot = 0, got_e = 0, toktype;
	register char *p = tokstart;
	int hex = input_radix > 10;

	if (c == '0' && (p[1] == 'x' || p[1] == 'X'))
	  {
	    p += 2;
	    hex = 1;
	  }
	else if (c == '0' && (p[1]=='t' || p[1]=='T' || p[1]=='d' || p[1]=='D'))
	  {
	    p += 2;
	    hex = 0;
	  }

	for (;; ++p)
	  {
	    if (!hex && !got_e && (*p == 'e' || *p == 'E'))
	      got_dot = got_e = 1;
	    else if (!hex && !got_dot && *p == '.')
	      got_dot = 1;
	    else if (got_e && (p[-1] == 'e' || p[-1] == 'E')
		     && (*p == '-' || *p == '+'))
	      /* This is the sign of the exponent, not the end of the
		 number.  */
	      continue;
	    /* We will take any letters or digits.  parse_number will
	       complain if past the radix, or if L or U are not final.  */
	    else if ((*p < '0' || *p > '9')
		     && ((*p < 'a' || *p > 'z')
				  && (*p < 'A' || *p > 'Z')))
	      break;
	  }
	toktype = parse_number (tokstart, p - tokstart, got_dot|got_e, &yylval);
        if (toktype == ERROR)
	  {
	    char *err_copy = (char *) alloca (p - tokstart + 1);

	    bcopy (tokstart, err_copy, p - tokstart);
	    err_copy[p - tokstart] = 0;
	    error ("Invalid number \"%s\".", err_copy);
	  }
	lexptr = p;
	return toktype;
      }

    case '+':
    case '-':
    case '*':
    case '/':
    case '%':
    case '|':
    case '&':
    case '^':
    case '~':
    case '!':
    case '@':
    case '<':
    case '>':
    case '[':
    case ']':
    case '?':
    case ':':
    case '=':
    case '{':
    case '}':
    symbol:
      lexptr++;
      return c;

    case '"':
      for (namelen = 1; (c = tokstart[namelen]) != '"'; namelen++)
	if (c == '\\')
	  {
	    c = tokstart[++namelen];
	    if (c >= '0' && c <= '9')
	      {
		c = tokstart[++namelen];
		if (c >= '0' && c <= '9')
		  c = tokstart[++namelen];
	      }
	  }
      yylval.sval.ptr = tokstart + 1;
      yylval.sval.length = namelen - 1;
      lexptr += namelen + 1;
      return STRING;
    }

  if (!(c == '_' || c == '$'
	|| (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')))
    /* We must have come across a bad character (e.g. ';').  */
    error ("Invalid character '%c' in expression.", c);

  /* It's a name.  See how long it is.  */
  namelen = 0;
  for (c = tokstart[namelen];
       (c == '_' || c == '$' || (c >= '0' && c <= '9')
	|| (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'));
       c = tokstart[++namelen])
    ;

  /* The token "if" terminates the expression and is NOT 
     removed from the input stream.  */
  if (namelen == 2 && tokstart[0] == 'i' && tokstart[1] == 'f')
    {
      return 0;
    }

  lexptr += namelen;

  /* Handle the tokens $digits; also $ (short for $0) and $$ (short for $$1)
     and $$digits (equivalent to $<-digits> if you could type that).
     Make token type LAST, and put the number (the digits) in yylval.  */

  if (*tokstart == '$')
    {
      register int negate = 0;
      c = 1;
      /* Double dollar means negate the number and add -1 as well.
	 Thus $$ alone means -1.  */
      if (namelen >= 2 && tokstart[1] == '$')
	{
	  negate = 1;
	  c = 2;
	}
      if (c == namelen)
	{
	  /* Just dollars (one or two) */
	  yylval.lval = - negate;
	  return LAST;
	}
      /* Is the rest of the token digits?  */
      for (; c < namelen; c++)
	if (!(tokstart[c] >= '0' && tokstart[c] <= '9'))
	  break;
      if (c == namelen)
	{
	  yylval.lval = atoi (tokstart + 1 + negate);
	  if (negate)
	    yylval.lval = - yylval.lval;
	  return LAST;
	}
    }

  /* Handle tokens that refer to machine registers:
     $ followed by a register name.  */

  if (*tokstart == '$') {
    for (c = 0; c < NUM_REGS; c++)
      if (namelen - 1 == strlen (reg_names[c])
	  && !strncmp (tokstart + 1, reg_names[c], namelen - 1))
	{
	  yylval.lval = c;
	  return REGNAME;
	}
    for (c = 0; c < NUM_STD_REGS; c++)
     if (namelen - 1 == strlen (std_regs[c].name)
	 && !strncmp (tokstart + 1, std_regs[c].name, namelen - 1))
       {
	 yylval.lval = std_regs[c].regnum;
	 return REGNAME;
       }
  }
  /* Catch specific keywords.  Should be done with a data structure.  */
  switch (namelen)
    {
    case 8:
      if (!strncmp (tokstart, "unsigned", 8))
	return UNSIGNED;
      break;
    case 6:
      if (!strncmp (tokstart, "struct", 6))
	return STRUCT;
      if (!strncmp (tokstart, "signed", 6))
	return SIGNED;
      if (!strncmp (tokstart, "sizeof", 6))      
	return SIZEOF;
      break;
    case 5:
      if (!strncmp (tokstart, "union", 5))
	return UNION;
      if (!strncmp (tokstart, "short", 5))
	return SHORT;
      break;
    case 4:
      if (!strncmp (tokstart, "enum", 4))
	return ENUM;
      if (!strncmp (tokstart, "long", 4))
	return LONG;
      if (!strncmp (tokstart, "this", 4))
	{
	  static const char this_name[] =
				 { CPLUS_MARKER, 't', 'h', 'i', 's', '\0' };

	  if (lookup_symbol (this_name, expression_context_block,
			     VAR_NAMESPACE, 0, NULL))
	    return THIS;
	}
      break;
    case 3:
      if (!strncmp (tokstart, "int", 3))
	return INT_KEYWORD;
      break;
    default:
      break;
    }

  yylval.sval.ptr = tokstart;
  yylval.sval.length = namelen;

  /* Any other names starting in $ are debugger internal variables.  */

  if (*tokstart == '$')
    {
      yylval.ivar =  lookup_internalvar (copy_name (yylval.sval) + 1);
      return VARIABLE;
    }

  /* Use token-type BLOCKNAME for symbols that happen to be defined as
     functions or symtabs.  If this is not so, then ...
     Use token-type TYPENAME for symbols that happen to be defined
     currently as names of types; NAME for other symbols.
     The caller is not constrained to care about the distinction.  */
  {
    char *tmp = copy_name (yylval.sval);
    struct symbol *sym;
    int is_a_field_of_this = 0;
    int hextype;
    Class class;

    if (class = lookup_objc_class(tmp))
      {
        yylval.sclass.class = class;
	yylval.sclass.type = lookup_struct_no_error(tmp, 0);
	if (!yylval.sclass.type)
	    yylval.sclass.type = builtin_type_error;
	return CLASSNAME;
      }

    sym = lookup_symbol (tmp, expression_context_block,
			 VAR_NAMESPACE, &is_a_field_of_this, NULL);
    if ((sym && SYMBOL_CLASS (sym) == LOC_BLOCK) ||
        lookup_partial_symtab (tmp))
      {
	yylval.ssym.sym = sym;
	yylval.ssym.is_a_field_of_this = is_a_field_of_this;
	return BLOCKNAME;
      }
    if (sym && SYMBOL_CLASS (sym) == LOC_TYPEDEF)
        {
	  yylval.tsym.type = SYMBOL_TYPE (sym);
	  return TYPENAME;
        }
    if ((yylval.tsym.type = lookup_primitive_typename (tmp)) != 0)
	return TYPENAME;

    /* Input names that aren't symbols but ARE valid hex numbers,
       when the input radix permits them, can be names or numbers
       depending on the parse.  Note we support radixes > 16 here.  */
    if (!sym && 
        ((tokstart[0] >= 'a' && tokstart[0] < 'a' + input_radix - 10) ||
         (tokstart[0] >= 'A' && tokstart[0] < 'A' + input_radix - 10)))
      {
 	YYSTYPE newlval;	/* Its value is ignored.  */
	hextype = parse_number (tokstart, namelen, 0, &newlval);
	if (hextype == INT)
	  {
	    yylval.ssym.sym = sym;
	    yylval.ssym.is_a_field_of_this = is_a_field_of_this;
	    return NAME_OR_INT;
	  }
	if (hextype == UINT)
	  {
	    yylval.ssym.sym = sym;
	    yylval.ssym.is_a_field_of_this = is_a_field_of_this;
	    return NAME_OR_UINT;
	  }
      }

    /* Any other kind of symbol */
    yylval.ssym.sym = sym;
    yylval.ssym.is_a_field_of_this = is_a_field_of_this;
    return NAME;
  }
}

static void
yyerror (msg)
     char *msg;
{
  error ("Invalid syntax in expression.");
}

/* Return a null-terminated temporary copy of the name
   of a string token.  */

static char *
copy_name (token)
     struct stoken token;
{
  bcopy (token.ptr, namecopy, token.length);
  namecopy[token.length] = 0;
  return namecopy;
}

/* Reverse an expression from suffix form (in which it is constructed)
   to prefix form (in which we can conveniently print or execute it).  */

static void prefixify_subexp ();

static void
prefixify_expression (expr)
     register struct expression *expr;
{
  register int len = sizeof (struct expression) +
				    expr->nelts * sizeof (union exp_element);
  register struct expression *temp;
  register int inpos = expr->nelts, outpos = 0;

  temp = (struct expression *) alloca (len);

  /* Copy the original expression into temp.  */
  bcopy (expr, temp, len);

  prefixify_subexp (temp, expr, inpos, outpos);
}

/* Return the number of exp_elements in the subexpression of EXPR
   whose last exp_element is at index ENDPOS - 1 in EXPR.  */

static int
length_of_subexp (expr, endpos)
     register struct expression *expr;
     register int endpos;
{
  register int oplen = 1;
  register int args = 0;
  register int i;

  if (endpos < 0)
    error ("?error in length_of_subexp");

  i = (int) expr->elts[endpos - 1].opcode;

  switch (i)
    {
      /* C++  */
    case OP_SCOPE:
      oplen = 4 + ((expr->elts[endpos - 2].longconst
		    + sizeof (union exp_element))
		   / sizeof (union exp_element));
      break;

    case OP_LONG:
    case OP_DOUBLE:
      oplen = 4;
      break;

    case OP_VAR_VALUE:
    case OP_LAST:
    case OP_REGISTER:
    case OP_INTERNALVAR:
      oplen = 3;
      break;

    case OP_FUNCALL:
      oplen = 3;
      args = 1 + expr->elts[endpos - 2].longconst;
      break;

    case OP_MSGCALL:
      oplen = 4;
      args = 1 + expr->elts[endpos - 2].longconst;
      break;

    case UNOP_CAST:
    case UNOP_MEMVAL:
      oplen = 3;
      args = 1;
      break;

    case STRUCTOP_STRUCT:
    case STRUCTOP_PTR:
      args = 1;
    case OP_STRING:
      oplen = 3 + ((expr->elts[endpos - 2].longconst
		    + sizeof (union exp_element))
		   / sizeof (union exp_element));
      break;

    case TERNOP_COND:
      args = 3;
      break;

    case BINOP_ASSIGN_MODIFY:
      oplen = 3;
      args = 2;
      break;

      /* C++ */
    case OP_THIS:
      oplen = 2;
      break;

    case OP_SELF:
      oplen = 2;
      break;

    default:
      args = 1 + (i < (int) BINOP_END);
    }

  while (args > 0)
    {
      oplen += length_of_subexp (expr, endpos - oplen);
      args--;
    }

  return oplen;
}

/* Copy the subexpression ending just before index INEND in INEXPR
   into OUTEXPR, starting at index OUTBEG.
   In the process, convert it from suffix to prefix form.  */

static void
prefixify_subexp (inexpr, outexpr, inend, outbeg)
     register struct expression *inexpr;
     struct expression *outexpr;
     register int inend;
     int outbeg;
{
  register int oplen = 1;
  register int args = 0;
  register int i;
  int *arglens;
  enum exp_opcode opcode;

  /* Compute how long the last operation is (in OPLEN),
     and also how many preceding subexpressions serve as
     arguments for it (in ARGS).  */

  opcode = inexpr->elts[inend - 1].opcode;
  switch (opcode)
    {
      /* C++  */
    case OP_SCOPE:
      oplen = 4 + ((inexpr->elts[inend - 2].longconst
		    + sizeof (union exp_element))
		   / sizeof (union exp_element));
      break;

    case OP_LONG:
    case OP_DOUBLE:
      oplen = 4;
      break;

    case OP_VAR_VALUE:
    case OP_LAST:
    case OP_REGISTER:
    case OP_INTERNALVAR:
      oplen = 3;
      break;

    case OP_FUNCALL:
      oplen = 3;
      args = 1 + inexpr->elts[inend - 2].longconst;
      break;

    case OP_MSGCALL:
      oplen = 4;
      args = 1 + inexpr->elts[inend - 2].longconst;
      break;

    case UNOP_CAST:
    case UNOP_MEMVAL:
      oplen = 3;
      args = 1;
      break;

    case STRUCTOP_STRUCT:
    case STRUCTOP_PTR:
      args = 1;
    case OP_STRING:
      oplen = 3 + ((inexpr->elts[inend - 2].longconst
		    + sizeof (union exp_element))
		   / sizeof (union exp_element));
		   
      break;

    case TERNOP_COND:
      args = 3;
      break;

    case BINOP_ASSIGN_MODIFY:
      oplen = 3;
      args = 2;
      break;

      /* C++ */
    case OP_THIS:
      oplen = 2;
      break;

    case OP_SELF:
      oplen = 2;
      break;

    default:
      args = 1 + ((int) opcode < (int) BINOP_END);
    }

  /* Copy the final operator itself, from the end of the input
     to the beginning of the output.  */
  inend -= oplen;
  bcopy (&inexpr->elts[inend], &outexpr->elts[outbeg],
	 oplen * sizeof (union exp_element));
  outbeg += oplen;

  /* Find the lengths of the arg subexpressions.  */
  arglens = (int *) alloca (args * sizeof (int));
  for (i = args - 1; i >= 0; i--)
    {
      oplen = length_of_subexp (inexpr, inend);
      arglens[i] = oplen;
      inend -= oplen;
    }

  /* Now copy each subexpression, preserving the order of
     the subexpressions, but prefixifying each one.
     In this loop, inend starts at the beginning of
     the expression this level is working on
     and marches forward over the arguments.
     outbeg does similarly in the output.  */
  for (i = 0; i < args; i++)
    {
      oplen = arglens[i];
      inend += oplen;
      prefixify_subexp (inexpr, outexpr, inend, outbeg);
      outbeg += oplen;
    }
}

/* This page contains the two entry points to this file.  */

/* Read a C expression from the string *STRINGPTR points to,
   parse it, and return a pointer to a  struct expression  that we malloc.
   Use block BLOCK as the lexical context for variable names;
   if BLOCK is zero, use the block of the selected stack frame.
   Meanwhile, advance *STRINGPTR to point after the expression,
   at the first nonwhite character that is not part of the expression
   (possibly a null character).

   If COMMA is nonzero, stop if a comma is reached.  */

struct expression *
parse_c_1 (stringptr, block, comma)
     char **stringptr;
     struct block *block;
     int comma;
{
  struct cleanup *old_chain;

  lexptr = *stringptr;

  paren_depth = 0;
  type_stack_depth = 0;

  comma_terminates = comma;

  if (lexptr == 0 || *lexptr == 0)
    error_no_arg ("expression to compute");

  old_chain = make_cleanup (free_funcalls, 0);
  funcall_chain = 0;

  expression_context_block = block ? block : get_selected_block ();

  namecopy = (char *) alloca (strlen (lexptr) + 1);
  expout_size = 10;
  expout_ptr = 0;
  expout = (struct expression *)
    xmalloc (sizeof (struct expression)
	     + expout_size * sizeof (union exp_element));
  make_cleanup (free_current_contents, &expout);
  if (yyparse ())
    yyerror (NULL);
  discard_cleanups (old_chain);
  expout->nelts = expout_ptr;
  expout = (struct expression *)
    xrealloc (expout,
	      sizeof (struct expression)
	      + expout_ptr * sizeof (union exp_element));
  prefixify_expression (expout);
  *stringptr = lexptr;
  return expout;
}

/* Parse STRING as an expression, and complain if this fails
   to use up all of the contents of STRING.  */

struct expression *
parse_c_expression (string)
     char *string;
{
  register struct expression *exp;
  exp = parse_c_1 (&string, 0, 0);
  if (*string)
    error ("Junk after end of expression.");
  return exp;
}

static void 
push_type (tp)
     enum type_pieces tp;
{
  if (type_stack_depth == type_stack_size)
    {
      type_stack_size *= 2;
      type_stack = (union type_stack_elt *)
	xrealloc (type_stack, type_stack_size * sizeof (*type_stack));
    }
  type_stack[type_stack_depth++].piece = tp;
}

static void
push_type_int (n)
     int n;
{
  if (type_stack_depth == type_stack_size)
    {
      type_stack_size *= 2;
      type_stack = (union type_stack_elt *)
	xrealloc (type_stack, type_stack_size * sizeof (*type_stack));
    }
  type_stack[type_stack_depth++].int_val = n;
}

static enum type_pieces 
pop_type ()
{
  if (type_stack_depth)
    return type_stack[--type_stack_depth].piece;
  return tp_end;
}

static int
pop_type_int ()
{
  if (type_stack_depth)
    return type_stack[--type_stack_depth].int_val;
  /* "Can't happen".  */
  return 0;
}

static void
start_msglist()
{
  register struct selname *new = (struct selname *) xmalloc (sizeof (struct selname));

  new->next = selname_chain;
  new->msglist_len = msglist_len;
  new->msglist_sel = msglist_sel;
  msglist_len = 0;
  msglist_sel = (char *)xmalloc(1);
  *msglist_sel = 0;
  selname_chain = new;
}

static void
add_msglist(str, addcolon)
     struct stoken *str;
     int addcolon;
{
  char *s, *p;
  int len, plen;

  if (str == 0) {
    p = "";
    plen = 0;
  } else {
    p = str->ptr;
    plen = str->length;
  }
  len = plen + strlen(msglist_sel) + 2;
  s = (char *)xmalloc(len);
  strcpy(s, msglist_sel);
  strncat(s, p, plen);
  free(msglist_sel);
  msglist_sel = s;
  if (addcolon) {
    s[len-2] = ':';
    s[len-1] = 0;
    msglist_len++;
  }
}

static int
end_msglist()
{
  register int val = msglist_len;
  register struct selname *sel = selname_chain;
  register char *p = msglist_sel;
  int selid;

  selname_chain = sel->next;
  msglist_len = sel->msglist_len;
  msglist_sel = sel->msglist_sel;
  selid = lookup_child_selector(p);
  if (!selid)
    error("Can't find selector \"%s\"", p);
  write_exp_elt_longcst (selid);
  free(p);
  write_exp_elt_longcst (val);	/* Number of args */
  free(sel);
  return val;
}
void
_initialize_expread ()
{
  type_stack_size = 80;
  type_stack_depth = 0;
  type_stack = (union type_stack_elt *)
    xmalloc (type_stack_size * sizeof (*type_stack));
}
short yyexca[] ={
-1, 1,
	0, -1,
	-2, 0,
-1, 31,
	273, 73,
	-2, 133,
-1, 84,
	273, 101,
	-2, 29,
-1, 159,
	273, 74,
	-2, 75,
	};
# define YYNPROD 134
# define YYLAST 1009
short yyact[]={

   7,  53, 192, 165,  49,   5,  51,  14, 162,   4,
  47,  52,   6,  32,  33, 101, 105,  38,  39,  40,
 105,  37, 188, 201,  41,  35,  36,  34, 112, 113,
 106, 180, 102, 103, 106, 164,  49, 163, 100,   7,
  99,  92,  47,  91,   5, 150,  14, 152,   4, 149,
 153,   6,  94,  95,  96,  48,  97,  98,  12, 150,
  43, 152, 146, 149,  94,  95,  96, 107,  97,  98,
 154, 150, 203, 152, 187, 149, 161, 196, 176,   7,
 185, 197, 212,  43,   5, 213,  80,  48,   4, 193,
  13,   6, 194,   8, 206,  89,   2,  12, 155, 150,
 180, 152, 146, 149, 157, 146, 172,  43, 104, 168,
 146, 156, 155, 207, 202, 161, 187, 208,   7, 169,
 120, 175, 173, 144, 155,  14, 177,  82,  81,  13,
  30, 155,   8,  53,  64,  86,  49,  12,  51,  54,
 151,  55,  47,  52, 119, 147,  27, 145, 210, 118,
 111,  21, 155,   1, 171,   0,  62,  70,  63,  69,
  50,   0,   0,   0,   0,  53,   0,   0,  49,  13,
  51,  54,   8,  55,  47,  52,  12,   0,   0,   0,
   0,   0,   0,   0,   0,   0, 189,  48,   0,   0,
  65,   0,  50,   0,   0,  53,   0,   0,  49,   0,
  51,  54, 183,  55,  47,  52,   0,   0,  13,   0,
   0,   8,  94,  95,  96,   0,  97,  98,   0,  48,
  66,   0, 184,   0,  15,  19,  17,  20,  25,  42,
  31,  32,  33,  16,  18,  38,  39,  40,  11,  37,
  29,   0,  41,  35,  36,  34,  22,  23,  24,  48,
  26,  94,  95,  96, 198,  97,  98,   0,  44,  45,
  46,   9,  10,  15,  19,  17,  20,  25,  42,  31,
  32,  84,  16,  18,  38,  39,  40,  11,  37,  29,
 158,  41,  35,  36,  34,  22,  23,  24,   0,  26,
  44,  45,  46,   0, 148, 181, 182,   0, 186,   0,
   9,  10,   0,  15,  19,  17,  20,  25,  42,  31,
  32,  33,  16,  18,  38,  39,  40,  11,  37,  29,
   0,  41,  35,  36,  34,  22,  23,  24,   0,  26,
  94,  95,  96,   0,  97,  98,   0,   0,   0,   0,
   9,  10,  15,  19,  17,  20,  25,  42,  31,  32,
  33,  16,  18,  38,  39,  40,  11,  37,  29,   0,
  41,  35,  36,  34,  22,  23,  24, 209,  26,  94,
  95,  96,   0,  97,  98,   0,   0,   0,  71,   9,
  10,  68,  67,  58,  59,  60,  61,  56,  57,   0,
  44,  45,  46,  53,  64, 116,  49,   0,  51,  54,
   0,  55,  47,  52,   0,  53,  64,   0,  49,   0,
  51,  54,   0,  55,  47,  52,  62,  70,  63,  69,
  50,   0,  44,  45,  46,   0,   0,   0,  62,   0,
  63,  69,  50,  53,  64,   0,  49,   0,  51,  54,
   0,  55,  47,  52,   0,   0,   0,  48,   0,   0,
  65,   0,  44,  45,  46,   0,  62,   0,  63,  48,
  50,   0,  65,   0,  85,   0,   0,  28,  53,  64,
   0,  49,   0,  51,  54,   0,  55,  47,  52,  88,
  66,  87,  90,   0,   0,   0,   0,  48,   0,   0,
  65,  62,  66,  63,   0,  50,  53,  64,   0,  49,
   0,  51,  54,   0,  55,  47,  52,  53,  64,   0,
  49,   0,  51,  54,   0,  55,  47,  52,   0,  62,
  66,  63,  48,  50,   0,  65,   0,   0,   0,   0,
  62,   0,  63,  53,  50,   0,  49,   0,  51,  54,
   0,  55,  47,  52,   0, 143,   0,   0,  90,   0,
  48,   0,   0,  65,   0,  66,  62,   0,  63,  53,
  50,  48,  49,   0,  51,  54,   0,  55,  47,  52,
  53,   0,   0,  49,   0,  51,  54,   0,  55,  47,
  52,   0,  62,   0,  63,   0,  50,  48,   0,   0,
  32,  33,   0,   0,  38,  39,  40,  50,  37,   0,
   0,  41,  35,  36,  34,   0,   0,   0,   0,   0,
   0,   0,   0,  48, 179,  94,  95,  96,   0,  97,
  98,   0,   0,   0,  48,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,  71,   0,
   0,  68,  67,  58,  59,  60,  61,  56,  57,   0,
  44,  45,  46,  68,  67,  58,  59,  60,  61,  56,
  57,   0,  44,  45,  46,   0,   0,   0,   0,   0,
   0,   0,   0, 211,   0,   0,  87,   0, 214,   0,
   0,  87,  67,  58,  59,  60,  61,  56,  57,   0,
  44,  45,  46,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,  58,  59,
  60,  61,  56,  57,   0,  44,  45,  46,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,  58,  59,  60,  61,
  56,  57,   0,  44,  45,  46,   0,  58,  59,  60,
  61,  56,  57,   0,  44,  45,  46,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,  58,  59,  60,  61,  56,  57,   3,
  44,  45,  46,   0,  72,  73,  74,  75,  76,  77,
  78,  79,  83,   0,   0,   0,   0,   0,   0,   0,
   0,  60,  61,  56,  57,   0,  44,  45,  46,   0,
   0,   0,   0,   0,  56,  57,   0,  44,  45,  46,
   0,   0, 160, 114,   0,   0,   0,   0,   0,   0,
 121, 122, 123, 124, 125, 126, 127, 128, 129, 130,
 131, 132, 133, 134, 135, 136, 137, 138, 139, 140,
 141, 142,  93,   0,   0,   0,   0,   0,   0,   0,
   0, 108, 109, 110,   0,   0,   0,   0,   0, 115,
 117,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0, 166,   0, 167,   0,
 170,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0, 159,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0, 178,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0, 190,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0, 195, 190,   0,   0,   0,   0, 200,   0,   0,
   0,   0,   0,   0,   0,   0,   0, 174,   0,   0,
   0,   0,   0,   0, 204,   0,   0, 205,   0,   0,
   0,   0,   0,   0, 191,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0, 199 };
short yypact[]={

 -33,-1000,  39, 356, -33, -33, -33, -33, -33, -33,
 -33,  46,   6,-251, -33,-1000,-1000,-1000,-1000,-1000,
-1000,-1000,-1000,-1000,-1000,-1000,-1000,-230,-232,-210,
-1000,-1000,-1000,-1000,-1000,-238,-263,-244,-198,-210,
-210,-248,-1000, -33,-1000,-1000, 353, 107, -33,-1000,
 -33, -33, -33, -33, -33, -33, -33, -33, -33, -33,
 -33, -33, -33, -33, -33, -33, -33, -33, -33, -33,
 -33, -33,  -4,  -4,  -4,  -4,  -4,  -4,  -4,  -4,
 -33,-1000,-1000, 356,-1000,  22,-1000,  21,  70,  63,
   7,-210, -50,-1000,-1000,-1000,-1000,-1000,-1000,-1000,
-270,-1000,-241,-275,-1000,-1000,-1000,-1000,-1000,-1000,
-1000,-1000,-1000,-1000, 356,-1000, -33,-1000, -33,  16,
 -33, 158,  -4,  -4,  -4, -36, -36, 128, 128, 522,
 522, 533, 533, 533, 533, 496, 470, 459, 431, 396,
  96, 356, 356,  65,  68, -33,-251,-1000,  58,  61,
  61,  40,  33,-1000,-1000, -71, -33,-1000, -11,-1000,
-1000,-210,-1000,-1000,-276,-1000,  -4,  -4,-1000,  48,
 356, -33,  85, -16,  23,  68,-1000, -33,  -4,-250,
-1000,-1000,-1000,-1000,-1000,  75,  73,-1000,-1000, -21,
  -4,-1000,-1000,-1000, -33, 368,-1000, -33,-1000,  23,
 356,  52,-1000,-1000, 356, 356,  72,  77, 326,-1000,
  41,  62,-1000,-251,  62 };
short yypgo[]={

   0, 789,  95, 153, 151, 464, 467, 148, 146,  70,
 140, 145, 135,  50, 832, 130, 108, 128, 127, 123,
 122, 121,  78, 120, 119 };
short yyr1[]={

   0,   3,   2,   2,   1,   1,   1,   1,   1,   1,
   1,   1,   1,   1,   1,   1,   1,   1,   1,  17,
  17,  19,   1,  20,  20,  21,  21,  22,  22,  18,
  23,   1,  24,  24,  24,   1,   1,   1,   1,   1,
   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
   1,   1,   1,   8,   8,   4,   4,   4,   4,   4,
  12,  12,  11,  11,  11,  11,  11,  10,  10,  10,
  10,  10,  13,  13,   9,   5,   5,   5,   5,   5,
   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,
   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,
   6,  16,  16,  16,  16,   7,   7,  14,  14,  14,
  14,  14,  15,  15 };
short yyr2[]={

   0,   1,   1,   3,   2,   2,   2,   2,   2,   2,
   2,   2,   2,   2,   3,   4,   3,   4,   4,   1,
   1,   0,   5,   1,   1,   1,   2,   3,   2,   1,
   0,   5,   0,   1,   3,   4,   4,   3,   3,   3,
   3,   3,   3,   3,   3,   3,   3,   3,   3,   3,
   3,   3,   3,   3,   3,   3,   3,   5,   3,   3,
   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
   4,   1,   1,   1,   3,   3,   3,   4,   2,   1,
   1,   2,   1,   2,   1,   2,   1,   3,   2,   1,
   2,   1,   2,   3,   2,   1,   3,   6,   8,   9,
   1,   1,   1,   1,   1,   2,   3,   2,   3,   3,
   4,   2,   3,   2,   2,   2,   2,   2,   1,   2,
   1,   1,   1,   1,   1,   1,   3,   1,   1,   1,
   1,   1,   1,   1 };
short yychk[]={

-1000,  -3,  -2,  -1,  42,  38,  45,  33, 126, 294,
 295, 271,  91, 123,  40, 257, 266, 259, 267, 258,
 260,  -4, 279, 280, 281, 261, 283,  -8,  -6, 273,
 -15, 263, 264, 265, 278, 276, 277, 272, 268, 269,
 270, 275, 262,  44, 294, 295, 296,  46,  91,  40,
  64,  42,  47,  37,  43,  45, 291, 292, 287, 288,
 289, 290,  60,  62,  38,  94, 124, 286, 285,  63,
  61, 282,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
  40, -17, -18,  -1, 265,  -5, -12,  -6,  -5,  -2,
  -6, 273, 273, -14, 262, 263, 264, 266, 267, 278,
 276, 278, 276, 277, -16, 264, 278, 265, -14, -14,
 -14, -16, 276, 277,  -1, -14,  42, -14,  42,  -2,
 -23,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
  -1,  -1,  -1,  -5, -19, 125,  40, -11, 273,  42,
  38, -10,  40, -13,  -9,  91,  41,  41, 273, -14,
 -14, 126, 278, 278, 276, 278,  -1,  -1,  93, -24,
  -1,  58,  41, -20, -14, -21, -22,  58,  -1,  -6,
  42, -11, -11, -13,  -9,  40, -11,  41,  93, 257,
  -1, -14, 278,  41,  44,  -1,  93,  58, -22, -14,
  -1, 273,  41,  93,  -1,  -1,  42,  41,  40,  41,
  -7,  -5,  41,  44,  -5 };
short yydef[]={

   0,  -2,   1,   2,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,  60,  61,  62,  63,  64,
  65,  66,  67,  68,  69,  71,  72,   0,   0,   0,
  79,  -2, 100, 101, 102, 103, 104, 118,   0,   0,
   0, 120, 132,   0,  11,  12,   0,   0,   0,  30,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   4,   5,   6,   7,   8,   9,  10,  13,
   0,  21,  19,  20,  -2,   0,  95,  80,   0,   0,
  80,   0,   0,  78, 127, 128, 129, 130, 131, 105,
 107, 111, 123, 124, 117, 121, 122, 113, 114, 115,
 116, 119, 123, 124,   3,  14,   0,  16,   0,   0,
  32,  38,  39,  40,  41,  42,  43,  44,  45,  46,
  47,  48,  49,  50,  51,  52,  53,  54,  55,  56,
   0,  58,  59,   0,   0,   0,   0,  81,   0,  82,
  84,  86,   0,  89,  91,   0,   0,  37,   0,  -2,
  76,   0, 108, 106, 109, 112,  15,  17,  18,   0,
  33,   0,  70,   0,  23,  24,  25,   0,  35,   0,
  96,  83,  85,  88,  90,   0,   0,  94,  92,   0,
  36,  77, 110,  31,   0,  57,  22,   0,  26,   0,
  28,   0,  87,  93,  34,  27,   0,  97,   0,  98,
   0, 125,  99,   0, 126 };
# line 1 "/usr/lib/yaccpar"
#ifndef lint
static char yaccpar_sccsid[] = "@(#)yaccpar	4.1	(Berkeley)	2/11/83";
#endif not lint

# define YYFLAG -1000
# define YYERROR goto yyerrlab
# define YYACCEPT return(0)
# define YYABORT return(1)

/*	parser for yacc output	*/

#ifdef YYDEBUG
int yydebug = 0; /* 1 for debugging */
#endif
YYSTYPE yyv[YYMAXDEPTH]; /* where the values are stored */
int yychar = -1; /* current input token number */
int yynerrs = 0;  /* number of errors */
short yyerrflag = 0;  /* error recovery flag */

yyparse() {

	short yys[YYMAXDEPTH];
	short yyj, yym;
	register YYSTYPE *yypvt;
	register short yystate, *yyps, yyn;
	register YYSTYPE *yypv;
	register short *yyxi;

	yystate = 0;
	yychar = -1;
	yynerrs = 0;
	yyerrflag = 0;
	yyps= &yys[-1];
	yypv= &yyv[-1];

 yystack:    /* put a state and value onto the stack */

#ifdef YYDEBUG
	if( yydebug  ) printf( "state %d, char 0%o\n", yystate, yychar );
#endif
		if( ++yyps>= &yys[YYMAXDEPTH] ) { yyerror( "yacc stack overflow" ); return(1); }
		*yyps = yystate;
		++yypv;
		*yypv = yyval;

 yynewstate:

	yyn = yypact[yystate];

	if( yyn<= YYFLAG ) goto yydefault; /* simple state */

	if( yychar<0 ) if( (yychar=yylex())<0 ) yychar=0;
	if( (yyn += yychar)<0 || yyn >= YYLAST ) goto yydefault;

	if( yychk[ yyn=yyact[ yyn ] ] == yychar ){ /* valid shift */
		yychar = -1;
		yyval = yylval;
		yystate = yyn;
		if( yyerrflag > 0 ) --yyerrflag;
		goto yystack;
		}

 yydefault:
	/* default state action */

	if( (yyn=yydef[yystate]) == -2 ) {
		if( yychar<0 ) if( (yychar=yylex())<0 ) yychar = 0;
		/* look through exception table */

		for( yyxi=yyexca; (*yyxi!= (-1)) || (yyxi[1]!=yystate) ; yyxi += 2 ) ; /* VOID */

		while( *(yyxi+=2) >= 0 ){
			if( *yyxi == yychar ) break;
			}
		if( (yyn = yyxi[1]) < 0 ) return(0);   /* accept */
		}

	if( yyn == 0 ){ /* error */
		/* error ... attempt to resume parsing */

		switch( yyerrflag ){

		case 0:   /* brand new error */

			yyerror( "syntax error" );
		yyerrlab:
			++yynerrs;

		case 1:
		case 2: /* incompletely recovered error ... try again */

			yyerrflag = 3;

			/* find a state where "error" is a legal shift action */

			while ( yyps >= yys ) {
			   yyn = yypact[*yyps] + YYERRCODE;
			   if( yyn>= 0 && yyn < YYLAST && yychk[yyact[yyn]] == YYERRCODE ){
			      yystate = yyact[yyn];  /* simulate a shift of "error" */
			      goto yystack;
			      }
			   yyn = yypact[*yyps];

			   /* the current yyps has no shift onn "error", pop stack */

#ifdef YYDEBUG
			   if( yydebug ) printf( "error recovery pops state %d, uncovers %d\n", *yyps, yyps[-1] );
#endif
			   --yyps;
			   --yypv;
			   }

			/* there is no state on the stack with an error shift ... abort */

	yyabort:
			return(1);


		case 3:  /* no shift yet; clobber input char */

#ifdef YYDEBUG
			if( yydebug ) printf( "error recovery discards char %d\n", yychar );
#endif

			if( yychar == 0 ) goto yyabort; /* don't discard EOF, quit */
			yychar = -1;
			goto yynewstate;   /* try again in the same state */

			}

		}

	/* reduction by production yyn */

#ifdef YYDEBUG
		if( yydebug ) printf("reduce %d\n",yyn);
#endif
		yyps -= yyr2[yyn];
		yypvt = yypv;
		yypv -= yyr2[yyn];
		yyval = yypv[1];
		yym=yyn;
			/* consult goto table to find next state */
		yyn = yyr1[yyn];
		yyj = yypgo[yyn] + *yyps + 1;
		if( yyj>=YYLAST || yychk[ yystate = yyact[yyj] ] != -yyn ) yystate = yyact[yypgo[yyn]];
		switch(yym){
			
case 3:
# line 262 "../expread.y"
{ write_exp_elt_opcode (BINOP_COMMA); } break;
case 4:
# line 267 "../expread.y"
{ write_exp_elt_opcode (UNOP_IND); } break;
case 5:
# line 270 "../expread.y"
{ write_exp_elt_opcode (UNOP_ADDR); } break;
case 6:
# line 273 "../expread.y"
{ write_exp_elt_opcode (UNOP_NEG); } break;
case 7:
# line 277 "../expread.y"
{ write_exp_elt_opcode (UNOP_ZEROP); } break;
case 8:
# line 281 "../expread.y"
{ write_exp_elt_opcode (UNOP_LOGNOT); } break;
case 9:
# line 285 "../expread.y"
{ write_exp_elt_opcode (UNOP_PREINCREMENT); } break;
case 10:
# line 289 "../expread.y"
{ write_exp_elt_opcode (UNOP_PREDECREMENT); } break;
case 11:
# line 293 "../expread.y"
{ write_exp_elt_opcode (UNOP_POSTINCREMENT); } break;
case 12:
# line 297 "../expread.y"
{ write_exp_elt_opcode (UNOP_POSTDECREMENT); } break;
case 13:
# line 301 "../expread.y"
{ write_exp_elt_opcode (UNOP_SIZEOF); } break;
case 14:
# line 305 "../expread.y"
{ write_exp_elt_opcode (STRUCTOP_PTR);
			  write_exp_string (yypvt[-0].sval);
			  write_exp_elt_opcode (STRUCTOP_PTR); } break;
case 15:
# line 311 "../expread.y"
{ write_exp_elt_opcode (STRUCTOP_MPTR); } break;
case 16:
# line 315 "../expread.y"
{ write_exp_elt_opcode (STRUCTOP_STRUCT);
			  write_exp_string (yypvt[-0].sval);
			  write_exp_elt_opcode (STRUCTOP_STRUCT); } break;
case 17:
# line 321 "../expread.y"
{ write_exp_elt_opcode (STRUCTOP_MEMBER); } break;
case 18:
# line 325 "../expread.y"
{ write_exp_elt_opcode (BINOP_SUBSCRIPT); } break;
case 21:
# line 333 "../expread.y"
{ start_msglist(); } break;
case 22:
# line 335 "../expread.y"
{ write_exp_elt_opcode (OP_MSGCALL);
			  end_msglist();
			  write_exp_elt_opcode (OP_MSGCALL); } break;
case 23:
# line 341 "../expread.y"
{ add_msglist(&yypvt[-0].sval, 0); } break;
case 27:
# line 350 "../expread.y"
{ add_msglist(&yypvt[-2].sval, 1); } break;
case 28:
# line 352 "../expread.y"
{ add_msglist(0, 1); } break;
case 29:
# line 356 "../expread.y"
{ 
			    write_exp_elt_opcode (OP_LONG);
			    write_exp_elt_type (builtin_type_int);
			    write_exp_elt_longcst (yypvt[-0].sclass.class);
			    write_exp_elt_opcode (OP_LONG);
			} break;
case 30:
# line 367 "../expread.y"
{ start_arglist (); } break;
case 31:
# line 369 "../expread.y"
{ write_exp_elt_opcode (OP_FUNCALL);
			  write_exp_elt_longcst ((LONGEST) end_arglist ());
			  write_exp_elt_opcode (OP_FUNCALL); } break;
case 33:
# line 378 "../expread.y"
{ arglist_len = 1; } break;
case 34:
# line 382 "../expread.y"
{ arglist_len++; } break;
case 35:
# line 386 "../expread.y"
{ write_exp_elt_opcode (UNOP_MEMVAL);
			  write_exp_elt_type (yypvt[-2].tval);
			  write_exp_elt_opcode (UNOP_MEMVAL); } break;
case 36:
# line 392 "../expread.y"
{ write_exp_elt_opcode (UNOP_CAST);
			  write_exp_elt_type (yypvt[-2].tval);
			  write_exp_elt_opcode (UNOP_CAST); } break;
case 37:
# line 398 "../expread.y"
{ } break;
case 38:
# line 404 "../expread.y"
{ write_exp_elt_opcode (BINOP_REPEAT); } break;
case 39:
# line 408 "../expread.y"
{ write_exp_elt_opcode (BINOP_MUL); } break;
case 40:
# line 412 "../expread.y"
{ write_exp_elt_opcode (BINOP_DIV); } break;
case 41:
# line 416 "../expread.y"
{ write_exp_elt_opcode (BINOP_REM); } break;
case 42:
# line 420 "../expread.y"
{ write_exp_elt_opcode (BINOP_ADD); } break;
case 43:
# line 424 "../expread.y"
{ write_exp_elt_opcode (BINOP_SUB); } break;
case 44:
# line 428 "../expread.y"
{ write_exp_elt_opcode (BINOP_LSH); } break;
case 45:
# line 432 "../expread.y"
{ write_exp_elt_opcode (BINOP_RSH); } break;
case 46:
# line 436 "../expread.y"
{ write_exp_elt_opcode (BINOP_EQUAL); } break;
case 47:
# line 440 "../expread.y"
{ write_exp_elt_opcode (BINOP_NOTEQUAL); } break;
case 48:
# line 444 "../expread.y"
{ write_exp_elt_opcode (BINOP_LEQ); } break;
case 49:
# line 448 "../expread.y"
{ write_exp_elt_opcode (BINOP_GEQ); } break;
case 50:
# line 452 "../expread.y"
{ write_exp_elt_opcode (BINOP_LESS); } break;
case 51:
# line 456 "../expread.y"
{ write_exp_elt_opcode (BINOP_GTR); } break;
case 52:
# line 460 "../expread.y"
{ write_exp_elt_opcode (BINOP_LOGAND); } break;
case 53:
# line 464 "../expread.y"
{ write_exp_elt_opcode (BINOP_LOGXOR); } break;
case 54:
# line 468 "../expread.y"
{ write_exp_elt_opcode (BINOP_LOGIOR); } break;
case 55:
# line 472 "../expread.y"
{ write_exp_elt_opcode (BINOP_AND); } break;
case 56:
# line 476 "../expread.y"
{ write_exp_elt_opcode (BINOP_OR); } break;
case 57:
# line 480 "../expread.y"
{ write_exp_elt_opcode (TERNOP_COND); } break;
case 58:
# line 484 "../expread.y"
{ write_exp_elt_opcode (BINOP_ASSIGN); } break;
case 59:
# line 488 "../expread.y"
{ write_exp_elt_opcode (BINOP_ASSIGN_MODIFY);
			  write_exp_elt_opcode (yypvt[-1].opcode);
			  write_exp_elt_opcode (BINOP_ASSIGN_MODIFY); } break;
case 60:
# line 494 "../expread.y"
{ write_exp_elt_opcode (OP_LONG);
			  if (yypvt[-0].lval == (int) yypvt[-0].lval || yypvt[-0].lval == (unsigned int) yypvt[-0].lval)
			    write_exp_elt_type (builtin_type_int);
			  else
			    write_exp_elt_type (BUILTIN_TYPE_LONGEST);
			  write_exp_elt_longcst ((LONGEST) yypvt[-0].lval);
			  write_exp_elt_opcode (OP_LONG); } break;
case 61:
# line 504 "../expread.y"
{ YYSTYPE val;
			  parse_number (yypvt[-0].ssym.stoken.ptr, yypvt[-0].ssym.stoken.length, 0, &val);
			  write_exp_elt_opcode (OP_LONG);
			  if (val.lval == (int) val.lval ||
			      val.lval == (unsigned int) val.lval)
			    write_exp_elt_type (builtin_type_int);
			  else
			    write_exp_elt_type (BUILTIN_TYPE_LONGEST);
			  write_exp_elt_longcst (val.lval);
			  write_exp_elt_opcode (OP_LONG); } break;
case 62:
# line 517 "../expread.y"
{
			  write_exp_elt_opcode (OP_LONG);
			  if (yypvt[-0].ulval == (unsigned int) yypvt[-0].ulval)
			    write_exp_elt_type (builtin_type_unsigned_int);
			  else
			    write_exp_elt_type (BUILTIN_TYPE_UNSIGNED_LONGEST);
			  write_exp_elt_longcst ((LONGEST) yypvt[-0].ulval);
			  write_exp_elt_opcode (OP_LONG);
			} break;
case 63:
# line 529 "../expread.y"
{ YYSTYPE val;
			  parse_number (yypvt[-0].ssym.stoken.ptr, yypvt[-0].ssym.stoken.length, 0, &val);
			  write_exp_elt_opcode (OP_LONG);
			  if (val.ulval == (unsigned int) val.ulval)
			    write_exp_elt_type (builtin_type_unsigned_int);
			  else
			    write_exp_elt_type (BUILTIN_TYPE_UNSIGNED_LONGEST);
			  write_exp_elt_longcst ((LONGEST)val.ulval);
			  write_exp_elt_opcode (OP_LONG);
			} break;
case 64:
# line 542 "../expread.y"
{ write_exp_elt_opcode (OP_LONG);
			  write_exp_elt_type (builtin_type_char);
			  write_exp_elt_longcst ((LONGEST) yypvt[-0].lval);
			  write_exp_elt_opcode (OP_LONG); } break;
case 65:
# line 549 "../expread.y"
{ write_exp_elt_opcode (OP_DOUBLE);
			  write_exp_elt_type (builtin_type_double);
			  write_exp_elt_dblcst (yypvt[-0].dval);
			  write_exp_elt_opcode (OP_DOUBLE); } break;
case 67:
# line 559 "../expread.y"
{ write_exp_elt_opcode (OP_LAST);
			  write_exp_elt_longcst ((LONGEST) yypvt[-0].lval);
			  write_exp_elt_opcode (OP_LAST); } break;
case 68:
# line 565 "../expread.y"
{ write_exp_elt_opcode (OP_REGISTER);
			  write_exp_elt_longcst ((LONGEST) yypvt[-0].lval);
			  write_exp_elt_opcode (OP_REGISTER); } break;
case 69:
# line 571 "../expread.y"
{ write_exp_elt_opcode (OP_INTERNALVAR);
			  write_exp_elt_intern (yypvt[-0].ivar);
			  write_exp_elt_opcode (OP_INTERNALVAR); } break;
case 70:
# line 577 "../expread.y"
{ write_exp_elt_opcode (OP_LONG);
			  write_exp_elt_type (builtin_type_int);
			  write_exp_elt_longcst ((LONGEST) TYPE_LENGTH (yypvt[-1].tval));
			  write_exp_elt_opcode (OP_LONG); } break;
case 71:
# line 584 "../expread.y"
{ write_exp_elt_opcode (OP_STRING);
			  write_exp_string (yypvt[-0].sval);
			  write_exp_elt_opcode (OP_STRING); } break;
case 72:
# line 591 "../expread.y"
{ write_exp_elt_opcode (OP_THIS);
			  write_exp_elt_opcode (OP_THIS); } break;
case 73:
# line 598 "../expread.y"
{
			  if (yypvt[-0].ssym.sym != 0)
			      yyval.bval = SYMBOL_BLOCK_VALUE (yypvt[-0].ssym.sym);
			  else
			    {
			      struct symtab *tem =
				  lookup_symtab (copy_name (yypvt[-0].ssym.stoken));
			      if (tem)
				yyval.bval = BLOCKVECTOR_BLOCK
					 (BLOCKVECTOR (tem), STATIC_BLOCK);
			      else
				error ("No file or function \"%s\".",
				       copy_name (yypvt[-0].ssym.stoken));
			    }
			} break;
case 74:
# line 616 "../expread.y"
{ struct symbol *tem
			    = lookup_symbol (copy_name (yypvt[-0].sval), yypvt[-2].bval,
					     VAR_NAMESPACE, 0, NULL);
			  if (!tem || SYMBOL_CLASS (tem) != LOC_BLOCK)
			    error ("No function \"%s\" in specified context.",
				   copy_name (yypvt[-0].sval));
			  yyval.bval = SYMBOL_BLOCK_VALUE (tem); } break;
case 75:
# line 626 "../expread.y"
{ struct symbol *sym;
			  sym = lookup_symbol (copy_name (yypvt[-0].sval), yypvt[-2].bval,
					       VAR_NAMESPACE, 0, NULL);
			  if (sym == 0)
			    error ("No symbol \"%s\" in specified context.",
				   copy_name (yypvt[-0].sval));
			  write_exp_elt_opcode (OP_VAR_VALUE);
			  write_exp_elt_sym (sym);
			  write_exp_elt_opcode (OP_VAR_VALUE); } break;
case 76:
# line 638 "../expread.y"
{
			  struct type *type = yypvt[-2].tval;
			  if (TYPE_CODE (type) != TYPE_CODE_STRUCT
			      && TYPE_CODE (type) != TYPE_CODE_UNION)
			    error ("`%s' is not defined as an aggregate type.",
				   TYPE_NAME (type));

			  write_exp_elt_opcode (OP_SCOPE);
			  write_exp_elt_type (type);
			  write_exp_string (yypvt[-0].sval);
			  write_exp_elt_opcode (OP_SCOPE);
			} break;
case 77:
# line 651 "../expread.y"
{
			  struct type *type = yypvt[-3].tval;
			  if (TYPE_CODE (type) != TYPE_CODE_STRUCT
			      && TYPE_CODE (type) != TYPE_CODE_UNION)
			    error ("`%s' is not defined as an aggregate type.",
				   TYPE_NAME (type));

			  if (strcmp (type_name_no_tag (type), yypvt[-0].sval.ptr))
			    error ("invalid destructor `%s::~%s'",
				   type_name_no_tag (type), yypvt[-0].sval.ptr);

			  write_exp_elt_opcode (OP_SCOPE);
			  write_exp_elt_type (type);
			  write_exp_string (yypvt[-0].sval);
			  write_exp_elt_opcode (OP_SCOPE);
			  write_exp_elt_opcode (UNOP_LOGNOT);
			} break;
case 78:
# line 669 "../expread.y"
{
			  char *name = copy_name (yypvt[-0].sval);
			  struct symbol *sym;
			  int i;

			  sym =
			    lookup_symbol (name, 0, VAR_NAMESPACE, 0, NULL);
			  if (sym)
			    {
			      write_exp_elt_opcode (OP_VAR_VALUE);
			      write_exp_elt_sym (sym);
			      write_exp_elt_opcode (OP_VAR_VALUE);
			      break;
			    }
			  for (i = 0; i < misc_function_count; i++)
			    if (!strcmp (misc_function_vector[i].name, name))
			      break;

			  if (i < misc_function_count)
			    {
			      enum misc_function_type mft =
				  misc_function_vector[i].type;
			      
			      write_exp_elt_opcode (OP_LONG);
			      write_exp_elt_type (builtin_type_int);
			      write_exp_elt_longcst ((LONGEST) misc_function_vector[i].address);
			      write_exp_elt_opcode (OP_LONG);
			      write_exp_elt_opcode (UNOP_MEMVAL);
			      if (mft == mf_data || mft == mf_bss)
				write_exp_elt_type (builtin_type_int);
			      else if (mft == mf_text)
				write_exp_elt_type (lookup_function_type (builtin_type_int));
			      else
				write_exp_elt_type (builtin_type_char);
			      write_exp_elt_opcode (UNOP_MEMVAL);
			    }
			  else
			    if (symtab_list == 0
				&& partial_symtab_list == 0)
			      error ("No symbol table is loaded.  Use the \"file\" command.");
			    else
			      error ("No symbol \"%s\" in current context.", name);
			} break;
case 79:
# line 715 "../expread.y"
{ struct symbol *sym = yypvt[-0].ssym.sym;

			  if (sym)
			    {
			      switch (sym->class)
				{
				case LOC_REGISTER:
				case LOC_ARG:
				case LOC_REF_ARG:
				case LOC_REGPARM:
				case LOC_LOCAL:
				case LOC_LOCAL_ARG:
				  if (innermost_block == 0 ||
				      contained_in (block_found, 
						    innermost_block))
				    innermost_block = block_found;
				case LOC_UNDEF:
				case LOC_CONST:
				case LOC_STATIC:
				case LOC_TYPEDEF:
				case LOC_LABEL:
				case LOC_BLOCK:
				case LOC_EXTERNAL:
				case LOC_CONST_BYTES:

				  /* In this case the expression can
				     be evaluated regardless of what
				     frame we are in, so there is no
				     need to check for the
				     innermost_block.  These cases are
				     listed so that gcc -Wall will
				     report types that may not have
				     been considered.  */

				  break;
				}
			      write_exp_elt_opcode (OP_VAR_VALUE);
			      write_exp_elt_sym (sym);
			      write_exp_elt_opcode (OP_VAR_VALUE);
			    }
			  else if (yypvt[-0].ssym.is_a_field_of_this)
			    {
			      /* C++: it hangs off of `this'.  Must
			         not inadvertently convert from a method call
				 to data ref.  */
			      if (innermost_block == 0 || 
				  contained_in (block_found, innermost_block))
				innermost_block = block_found;
			      write_exp_elt_opcode (yypvt[-0].ssym.is_a_field_of_this);
			      write_exp_elt_opcode (yypvt[-0].ssym.is_a_field_of_this);
			      write_exp_elt_opcode (STRUCTOP_PTR);
			      write_exp_string (yypvt[-0].ssym.stoken);
			      write_exp_elt_opcode (STRUCTOP_PTR);
			    }
			  else
			    {
			      register int i;
			      register char *arg = copy_name (yypvt[-0].ssym.stoken);

				/* FIXME, this search is linear!  At least
				   optimize the strcmp with a 1-char cmp... */
			      for (i = 0; i < misc_function_count; i++)
				if (!strcmp (misc_function_vector[i].name, arg))
				  break;

			      if (i < misc_function_count)
				{
				  enum misc_function_type mft =
				      misc_function_vector[i].type;
				  
				  write_exp_elt_opcode (OP_LONG);
				  write_exp_elt_type (builtin_type_int);
				  write_exp_elt_longcst ((LONGEST) misc_function_vector[i].address);
				  write_exp_elt_opcode (OP_LONG);
				  write_exp_elt_opcode (UNOP_MEMVAL);
				  if (mft == mf_data || mft == mf_bss)
				    write_exp_elt_type (builtin_type_int);
				  else if (mft == mf_text)
				    write_exp_elt_type (lookup_function_type (builtin_type_int));
				  else
				    write_exp_elt_type (builtin_type_char);
				  write_exp_elt_opcode (UNOP_MEMVAL);
				}
			      else if (symtab_list == 0
				       && partial_symtab_list == 0)
				error ("No symbol table is loaded.  Use the \"file\" command.");
			      else
				error ("No symbol \"%s\" in current context.",
				       copy_name (yypvt[-0].ssym.stoken));
			    }
			} break;
case 81:
# line 811 "../expread.y"
{
		  /* This is where the interesting stuff happens.  */
		  int done = 0;
		  int array_size;
		  struct type *follow_type = yypvt[-1].tval;
		  
		  while (!done)
		    switch (pop_type ())
		      {
		      case tp_end:
			done = 1;
			break;
		      case tp_pointer:
			follow_type = lookup_pointer_type (follow_type);
			break;
		      case tp_reference:
			follow_type = lookup_reference_type (follow_type);
			break;
		      case tp_array:
			array_size = pop_type_int ();
			if (array_size != -1)
			  follow_type = create_array_type (follow_type,
							   array_size);
			else
			  follow_type = lookup_pointer_type (follow_type);
			break;
		      case tp_function:
			follow_type = lookup_function_type (follow_type);
			break;
		      }
		  yyval.tval = follow_type;
		} break;
case 82:
# line 846 "../expread.y"
{ push_type (tp_pointer); yyval.voidval = 0; } break;
case 83:
# line 848 "../expread.y"
{ push_type (tp_pointer); yyval.voidval = yypvt[-0].voidval; } break;
case 84:
# line 850 "../expread.y"
{ push_type (tp_reference); yyval.voidval = 0; } break;
case 85:
# line 852 "../expread.y"
{ push_type (tp_reference); yyval.voidval = yypvt[-0].voidval; } break;
case 87:
# line 857 "../expread.y"
{ yyval.voidval = yypvt[-1].voidval; } break;
case 88:
# line 859 "../expread.y"
{
			  push_type_int (yypvt[-0].lval);
			  push_type (tp_array);
			} break;
case 89:
# line 864 "../expread.y"
{
			  push_type_int (yypvt[-0].lval);
			  push_type (tp_array);
			  yyval.voidval = 0;
			} break;
case 90:
# line 870 "../expread.y"
{ push_type (tp_function); } break;
case 91:
# line 872 "../expread.y"
{ push_type (tp_function); } break;
case 92:
# line 876 "../expread.y"
{ yyval.lval = -1; } break;
case 93:
# line 878 "../expread.y"
{ yyval.lval = yypvt[-1].lval; } break;
case 94:
# line 882 "../expread.y"
{ yyval.voidval = 0; } break;
case 96:
# line 887 "../expread.y"
{ yyval.tval = lookup_member_type (builtin_type_int, yypvt[-2].tval); } break;
case 97:
# line 889 "../expread.y"
{ yyval.tval = lookup_member_type (yypvt[-5].tval, yypvt[-3].tval); } break;
case 98:
# line 891 "../expread.y"
{ yyval.tval = lookup_member_type
			    (lookup_function_type (yypvt[-7].tval), yypvt[-5].tval); } break;
case 99:
# line 894 "../expread.y"
{ yyval.tval = lookup_member_type
			    (lookup_function_type (yypvt[-8].tval), yypvt[-6].tval);
			  free (yypvt[-1].tvec); } break;
case 100:
# line 901 "../expread.y"
{ yyval.tval = yypvt[-0].tsym.type; } break;
case 101:
# line 903 "../expread.y"
{ yyval.tval = yypvt[-0].sclass.type; } break;
case 102:
# line 905 "../expread.y"
{ yyval.tval = builtin_type_int; } break;
case 103:
# line 907 "../expread.y"
{ yyval.tval = builtin_type_long; } break;
case 104:
# line 909 "../expread.y"
{ yyval.tval = builtin_type_short; } break;
case 105:
# line 911 "../expread.y"
{ yyval.tval = builtin_type_long; } break;
case 106:
# line 913 "../expread.y"
{ yyval.tval = builtin_type_unsigned_long; } break;
case 107:
# line 915 "../expread.y"
{ yyval.tval = builtin_type_long_long; } break;
case 108:
# line 917 "../expread.y"
{ yyval.tval = builtin_type_long_long; } break;
case 109:
# line 919 "../expread.y"
{ yyval.tval = builtin_type_unsigned_long_long; } break;
case 110:
# line 921 "../expread.y"
{ yyval.tval = builtin_type_unsigned_long_long; } break;
case 111:
# line 923 "../expread.y"
{ yyval.tval = builtin_type_short; } break;
case 112:
# line 925 "../expread.y"
{ yyval.tval = builtin_type_unsigned_short; } break;
case 113:
# line 927 "../expread.y"
{ yyval.tval = yypvt[-0].sclass.type; } break;
case 114:
# line 929 "../expread.y"
{ yyval.tval = lookup_struct (copy_name (yypvt[-0].sval),
					      expression_context_block); } break;
case 115:
# line 932 "../expread.y"
{ yyval.tval = lookup_union (copy_name (yypvt[-0].sval),
					     expression_context_block); } break;
case 116:
# line 935 "../expread.y"
{ yyval.tval = lookup_enum (copy_name (yypvt[-0].sval),
					    expression_context_block); } break;
case 117:
# line 938 "../expread.y"
{ yyval.tval = lookup_unsigned_typename (TYPE_NAME(yypvt[-0].tsym.type)); } break;
case 118:
# line 940 "../expread.y"
{ yyval.tval = builtin_type_unsigned_int; } break;
case 119:
# line 942 "../expread.y"
{ yyval.tval = yypvt[-0].tsym.type; } break;
case 120:
# line 944 "../expread.y"
{ yyval.tval = builtin_type_int; } break;
case 122:
# line 949 "../expread.y"
{
		  yyval.tsym.stoken.ptr = "int";
		  yyval.tsym.stoken.length = 3;
		  yyval.tsym.type = builtin_type_int;
		} break;
case 123:
# line 955 "../expread.y"
{
		  yyval.tsym.stoken.ptr = "long";
		  yyval.tsym.stoken.length = 4;
		  yyval.tsym.type = builtin_type_long;
		} break;
case 124:
# line 961 "../expread.y"
{
		  yyval.tsym.stoken.ptr = "short";
		  yyval.tsym.stoken.length = 5;
		  yyval.tsym.type = builtin_type_short;
		} break;
case 125:
# line 970 "../expread.y"
{ yyval.tvec = (struct type **)xmalloc (sizeof (struct type *) * 2);
		  yyval.tvec[0] = (struct type *)0;
		  yyval.tvec[1] = yypvt[-0].tval;
		} break;
case 126:
# line 975 "../expread.y"
{ int len = sizeof (struct type *) * ++(yypvt[-2].ivec[0]);
		  yyval.tvec = (struct type **)xrealloc (yypvt[-2].tvec, len);
		  yyval.tvec[yyval.ivec[0]] = yypvt[-0].tval;
		} break;
case 127:
# line 981 "../expread.y"
{ yyval.sval = yypvt[-0].ssym.stoken; } break;
case 128:
# line 982 "../expread.y"
{ yyval.sval = yypvt[-0].ssym.stoken; } break;
case 129:
# line 983 "../expread.y"
{ yyval.sval = yypvt[-0].tsym.stoken; } break;
case 130:
# line 984 "../expread.y"
{ yyval.sval = yypvt[-0].ssym.stoken; } break;
case 131:
# line 985 "../expread.y"
{ yyval.sval = yypvt[-0].ssym.stoken; } break;
# line 148 "/usr/lib/yaccpar"

		}
		goto yystack;  /* stack new state and value */

	}
